//	
//	BestCrypt Database functions.
//	They work with BCStorage.
//
#ifndef __BCDB_Functions__
#define __BCDB_Functions__

#include <windows.h>
#include "..\Includes\UnicodeFuncs.h"
#include "..\CertificateLib\SMemory\cPassword.h"
#include "..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "BCKeyMan.h"

//
// Simple functions:
//
bool BCDB_GenerateLocalKeyID( cPTB *ptbLocalKeyID, HPKA ska, bool secret = true );
int BCDB_bytes2string( char *str, int strLen, BYTE *bf, int bfSize );
int BCDB_bytes2string( wchar_t *str, int strLen, BYTE *bf, int bfSize );

//
//	BCDB_GetCrtSerialNumber
//		Every secret key in data base has corresponded number of issued certificates.
//		Gets serial number for new sertificate issued by the secret key with specified issuerLocalKeyID.
//	 1. Finds secret key in BCDB by local key ID.
//	 2. Gets stored number of previously signed certificates
//	 3. Increments the number if autoIncrement is true and stores it back 
//
bool BCDB_GetCrtSerialNumber( int *serialNumber, BYTE *issuerLocalKeyID, int size, bool autoIncrementSN = false );

//
//	BCDB_GetDataBasePath
//		Gets assigned path for BCStorage 
//
bool BCDB_GetDataBasePath( wchar_t *path, int size, bool silent = true );
bool BCDB_SetDataBasePath( wchar_t *path );
//
//	BCDB_SecKeyFileName
//		Generates name of secret key from local key ID, it simplifies find process
//		in BCStorage folder.
//
bool BCDB_SecKeyFileName( wchar_t *nameBuff, int buffLen, BYTE *lkid, int lkidSize, bool addBCStoragePath = true );
bool BCDB_SecKeyFileName( wchar_t *nameBuff, int buffLen, HPKA ska, bool addBCStoragePath = true );

//
//	BCDB_GetP12SecretKeyInfo - high level function works in user interface mode
//	1. Parses P12 shrouded key bag buffer.
//	2. Asks password and tries to decrypt excrypted secret key.
//	3. Create secret key handler ( use ska_free functions to free the handle ).
//
bool BCDB_GetP12SecretKeyInfo(
		HINSTANCE hInst, HWND hWnd,
		BYTE *buff, int buffSize,
		HPKA *pSka = NULL,
		cUniversalString *pUSFriendlyName = NULL,
		cPTB *ptbLocalKeyID = NULL,
		wchar_t *errStr = NULL, int errStrSize = 0
	);

//
//	BCDB_GetCerPublicKeyInfo - high level function works in user interface mode
//	1. Parses certificate public key.
//	2. Create public key handler ( use pka_free functions to free the handle ).
//
bool BCDB_GetCerPublicKeyInfo(
		HINSTANCE hInst, HWND hWnd,
		BYTE *buff, int buffSize,
		HPKA *pPka = NULL,
		cUniversalString *pUSFriendlyName = NULL,
		cPTB *ptbLocalKeyID = NULL,	// Public key local key ID generated from publick key packet
		cPTB *ptbThumbprint = NULL	// Certificate thumbprint is SHA1 digest of entire certificate. It is used as local key id of PFX PDU secret key package. 
									// See certificate lib VersionInfo.txt v. 1.04.8 point 8.
	);

//
//	BCDB_GetPGPSecretKeyInfo - high level function works in user interface mode
//	1. Parses PGP secret key packet.
//	2. Asks password and tries to decrypt excrypted secret key.
//	3. Create secret key handler ( use ska_free functions to free the handle ).
//
bool BCDB_GetPGPSecretKeyInfo(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *buffer, int bufferSize,
		HPKA *pHPKA,
		const wchar_t *friendlyName
	);

//
//	BCDB_GetPGPPublicKeyInfo - high level function works in user interface mode
//	1. Parses PGP public key packet.
//	2. Create public key handler ( use pka_free functions to free the handle ).
//
bool BCDB_GetPGPPublicKeyInfo( HINSTANCE hInst, HWND hWnd, BYTE *buffer, int bufferSize, HPKA *pHPKA );


//
//	BCDB_GetX509KeyByID
//		Finds public certificate or secret p12_pfx_pdu in BCStorage specified by localKeyID
//
//bool BCDB_GetX509KeyByID( cPTB *keyPtb, BYTE *localKeyIDBuffer, int bufferSize, bool secret, cPTB *ptbUserID = NULL, wchar_t *fileName = NULL );
bool BCDB_GetX509KeyByID( cPTB *keyPtb, BYTE *localKeyIDBuffer, int bufferSize, bool secret, cWCStr *wcsUserID = NULL, wchar_t *fileName = NULL );

///////////////////////////////
//
//	Local Key Ring stores public keys what are used for:
//		1. automatic sygnature verification.
//		2. for quick access to recipient public key.
//
//	BCDB_GetLocalKeyRingPath
//		Gets path to local PGP key ring, returns true if the files exists. 
//
bool BCDB_GetLocalKeyRingPath( wchar_t *path, int size, bool secret );

//
//	BCDB_AddPGPKeyToLocalKeyRing
//		Adds PGP key to local secret or public key ring
//		If error returns IDS of error string, otherwise returns 0 
//
int BCDB_AddPGPKeyToLocalKeyRing( HWND hWnd, BYTE *keyBuff, int keyBuffSize, bool secret = false );

//
//	BCDB_AddCertificateToStorage
//		1. Generates certificate name from RDN common name, 
//			if common name attribute does not present the certificate file name generated from local key ID
//		2. Checks existing the certificate file in BCDB, overwrites or skips the overwritting.
//		3. Fills pbtCharInfo information about the process.
//
//	returns true if no error or file writting is skipped.
//

bool BCDB_AddCertificateToStorage( BYTE *crtPkt, int crtPktSize, cWCStr *ptbCharInfo = NULL, HPKA *pPKA = NULL, wchar_t *path = NULL);

//
//	BCDB_AddP12SecretKeyToStorage
//		1. Parses secret key packet and gets local key ID
//		2. Generates file name by the local key ID and save BCDB
//
//	returns true if no error
//
bool BCDB_AddP12SecretKeyToStorage( BYTE *pkt, int pktSize, cWCStr *ptbCharInfo = NULL, HPKA *pSKA = NULL );

//
//	Key file format.
//
// enum type_kff { kffNone = 0, kffX509, kffPGP }; // defined in BCKeyMan.h
type_kff getKeyFileFormat( const wchar_t *fileName );

//
//	Export or import BCDB to/form file
//
bool BCDB_Port( HWND hWnd, bool bExport );

int BCDB_GetPGPKeyUserIDT(
		//cPTB *ptbUserID,		// [OUT]
		cWCStr *wcsUserID,
		BYTE *krBuff, int krBuffSize,	// [IN ] Key buffer or entire key ring
		BYTE *keyID,					// [IN ] The key identified to be found
		bool secret );


//
//	BCDB_GetCrtInfo
//		Reads used ID and key ID for parsed certificate.
//	Sample:
//		cFileView cfv( certificatePath );
//		x509_certificate crt;
//		if ( crt.parse( cfv.m_ptr, cfv.m_size ) != x509e_ok )
//			return false;
//		if ( !BCDB_GetCrtInfo( &crt, ptbUserID, ptbKeyID ) )
//			return false;
//
//		wchar_t lkIdStr[ MAX_PATH ];
//		BCDB_bytes2string( lkIdStr, maxLen( lkIdStr ), ptbKeyID.buffer, ptbKeyID.bufferSize );
//
bool BCDB_GetCrtInfo( x509_certificate *crt, cWCStr *wcsUserID = NULL, cPTB *ptbKeyID = NULL, cPTB *ptbThumbprint= NULL );

#endif // __BCDB_Functions__