#include "cWizardDlgs.h"
#include <windowsx.h>
#include <io.h>
#include <stdio.h>
#include "..\BCZip\Frame\OpenFiles.h"
#include "..\BCZip\Frame\stddlgs.h"
#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\BCZip\FileView.cpp"
#include "BCKeyMan.h"

enum { algRSA, algRSAWithSafePrime };

bool cWizardDlg_GenKey::OnInitDialog()
{
	HWND hWnd = GetDlgItem( m_hDlg, IDC_cmbAlgorithm );
	ComboBox_InsertString( hWnd, algRSA, "RSA" );
	ComboBox_InsertString( hWnd, algRSAWithSafePrime, "RSA ( generate safe prime )" );
	ComboBox_SetCurSel( hWnd, m_pBlock->algPos );

	SetDlgItemInt( m_hDlg, IDC_edtLength, m_pBlock->keyBitSize, false );

	Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ), m_pBlock->generateNew ? BST_CHECKED : BST_UNCHECKED );
	Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbUseExistingKey ), !m_pBlock->generateNew ? BST_CHECKED : BST_UNCHECKED );

	SetDlgItemText( m_hDlg, IDC_edtPath, m_pBlock->path );

	return true;
}

bool cWizardDlg_GenKey::OnCommand( WPARAM wParam, LPARAM lParam )
{
	if ( OnStdButton( wParam ) )
	{
		BOOL success = FALSE;
		UINT size = GetDlgItemInt( m_hDlg, IDC_edtLength, &success, 0 );
		if ( success )
			m_pBlock->keyBitSize = size;

		m_pBlock->algPos = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_cmbAlgorithm ) );
		m_pBlock->generateNew = Button_GetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ) ) == BST_CHECKED;

		return true;
	}
	else
	{
		switch ( LOWORD(wParam) )
		{
			case IDC_rbUseExistingKey:
				if ( ( m_pBlock->path[0] == '\0' ) ||
					 ( access( m_pBlock->path, 0 ) != 0 ) // does not exist
					)
					OnBrowseKeyFile();

			case IDC_rbGenerateNewKey:
				UpdateControls();
				break;
			case IDC_btBrowse:
				OnBrowseKeyFile();
				break;
			default:
				return false;
		}
		return true;
	}
}

const char *szFilters = 
	"PKCS 12 ( *.p12 )\0*.p12\0"
	"PGP secret keys ( *.skr,*.pgp )\0*.skr;*.pgp\0"
	"All files ( *.* )\0*.*\0";

//
//	Key file format.
//
enum type_kff { kffNone = 0, kffP12, kffPGP };

static type_kff getKeyFileFormat( const char *fileName )
{
	if ( fileName == NULL )
		return kffNone;

	int len = strlen( fileName );

	if ( len <= 4 )
		return kffNone;

	const char *extPos = fileName + len - 4;

	if ( stricmp( extPos, ".p12" ) == 0 ) return kffP12;
	if ( stricmp( extPos, ".skr" ) == 0 ) return kffPGP;
	if ( stricmp( extPos, ".pgp" ) == 0 ) return kffPGP;

	return kffNone;
}

bool cWizardDlg_GenKey::ReadSecretKeyFile()
{
	cPlainTextBlock ptbKey;
	cPlainTextBlock friendlyName;
	int res;

	const char *pwdDlgCaption = "Enter password for secret key";

	type_kff kff = getKeyFileFormat( m_pBlock->path );

	if ( kff == kffNone )
	{
		DlgMessageBox( IDS_UndefinedKeyFileFormat, MB_ICONSTOP );
		return false;
	}

	if ( kff == kffP12 )
	{
		cFileView cfv( m_pBlock->path );

		if ( cfv.m_err )
			return false;

		cPlainTextBlock ptbAttrSet;

		res = P12_PFX_PDU_parse( cfv.m_ptr, cfv.m_size, &ptbKey, NULL, &ptbAttrSet );

		if ( res != x509e_emptyParameters )
		{
			DlgMessageBox( IDS_SecretKeyPacketIsInvalid, MB_ICONSTOP );
			return false;
		}

		cUniversalString us;

		if ( ptbAttrSet.bufferSize == 0 )
		{
			us.Set( LoadStringOnce( IDS_Empty ) );
		}
		else
		{
			P12_attrSet attrSet;
			res = attrSet.parse( ptbAttrSet.buffer, ptbAttrSet.bufferSize, NULL, NULL, asn1tag_CONSTRUCTED_SET );
			if ( res != x509e_ok )
			{
				DlgMessageBox( IDS_SecretKeyPacketIsInvalid, MB_ICONSTOP );
				return false;
			}

			x509_struct sUnicodeString;
			res = attrSet.getFriendlyName( &sUnicodeString );

			if ( res == x509e_ok )
			{
				us.SetUnicode( sUnicodeString.B, sUnicodeString.L );
				us.MakeFromUnicodeString();
			}
			else
				us.Set( LoadStringOnce( IDS_Empty ) );
		}
	
		char text[ 1024 ];
		sprintf( text, "User name: %s", us.buffer );

		cPassword pwd;
		cPasswordDlg pwDlg( text, IDC_Text, IDC_Password, pwdDlgCaption, &pwd );

		while( true )
		{
			if ( pwDlg.DoModal( m_hInst, IDD_DlgPassword, m_hDlg ) != IDOK )
				return false;
		
			res = P12_PFX_PDU_parse( cfv.m_ptr, cfv.m_size, &ptbKey, &pwd );

			if ( res == ppe_ok )
				break;

			if ( res == ppe_invalidPassword )
			{
				DlgMessageBox( IDS_InvalidPassword, MB_ICONINFORMATION  );
				continue;
			}
			else
			{
				DlgMessageBox( IDS_ErrorInReadingKeyPacket, MB_ICONSTOP  );
				return false;
			}
		}

		return res == ppe_ok;
	}

	res = BrowseKeyRingForUser( m_hInst, m_hDlg, &ptbKey, m_pBlock->path, m_pBlock->path, true, &friendlyName );
	if ( res != IDOK )
		return false;

	//
	//	Checks if valid PGP secret key packet.
	//	The procedure parses all but key and cannot decrypt secret key becase pwd was passed NULL
	//
	res = PKT_CreateSecretKeyAlgFromPacket( &m_pBlock->ska, ( sSecretKey * )ptbKey.buffer, ptbKey.bufferSize, NULL );
	if ( res == ppe_invalidPassword )
	{
		char text[ 1024 ];
		friendlyName.PutByte( '\0' ); // Adds EOS
		sprintf( text, "User ID %s", friendlyName.buffer );

		cPassword pwd;
		cPasswordDlg pwDlg( text, IDC_Text, IDC_Password, pwdDlgCaption, &pwd );

		while( true )
		{
			if ( pwDlg.DoModal( m_hInst, IDD_DlgPassword, m_hDlg ) != IDOK )
				return false;

			res = PKT_CreateSecretKeyAlgFromPacket( &m_pBlock->ska, ( sSecretKey * )ptbKey.buffer, ptbKey.bufferSize, pwd.buffer );

			if ( res == ppe_ok )
				break;

			if ( res == ppe_invalidPassword )
			{
				DlgMessageBox( IDS_InvalidPassword, MB_ICONINFORMATION  );
				continue;
			}
			else
			{
				DlgMessageBox( IDS_ErrorInReadingKeyPacket, MB_ICONSTOP  );
				return false;
			}
		}
	}

	return res == ppe_ok;
}

bool cWizardDlg_GenKey::UpdateControls()
{
	m_pBlock->generateNew = Button_GetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ) ) == BST_CHECKED;

	bool enable = m_pBlock->generateNew;

	EnableWindow( GetDlgItem( m_hDlg, IDC_stAlgorithm ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_stSize ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_cmbAlgorithm ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_edtLength ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_spinSize ), enable );

	enable = !enable;

	EnableWindow( GetDlgItem( m_hDlg, IDC_edtPath ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btBrowse ), enable );
	ShowWindow( GetDlgItem( m_hDlg, IDC_stUsedAlgorithm ), enable? SW_SHOW : SW_HIDE );
	ShowWindow( GetDlgItem( m_hDlg, IDC_stUsedKeyID ), enable ? SW_SHOW : SW_HIDE );
	ShowWindow( GetDlgItem( m_hDlg, IDC_usedAlgorithm ), enable ? SW_SHOW : SW_HIDE );
	ShowWindow( GetDlgItem( m_hDlg, IDC_usedKeyID ), enable ? SW_SHOW : SW_HIDE );

	return false;
}

static int bytes2string( char *str, int strSize, BYTE *bf, int bfSize )
{
	int len = 0;
	int pos = 0;

	for(; pos < bfSize; pos++ )
	{
		if ( len + 3 > strSize )
			break;

		sprintf( str + len, "%02X ", (DWORD)bf[pos] );
		len += 3;
	}

	return pos;
}


bool cWizardDlg_GenKey::OnBrowseKeyFile()
{
	bool res = BrowseFiles( m_hDlg, m_pBlock->path, sizeof( m_pBlock->path ), szFilters, LoadStringOnce( IDS_ChooseSecretKey ), false ) == TRUE;

	if ( res )
		res = ReadSecretKeyFile();

	if ( res )
	{
		SetDlgItemText( m_hDlg, IDC_edtPath, m_pBlock->path );
		if ( m_pBlock->ska != NULL )
		{
			int id = ska_getAlgID( m_pBlock->ska );
			int length = ska_getModuleSize( m_pBlock->ska );
			char buff[ 1024 ];
			sprintf( buff, "%s ( %d bits )", PKT_GetPublicKeyAlgName( id ), length * 8 );
			SetDlgItemText( m_hDlg, IDC_usedAlgorithm, buff );

			int size = 0;
			cPlainTextBlock ptbAuthKeyID;
			int ires = ska_generateLocalKeyID( m_pBlock->ska, NULL, &size );
			if ( ires != x509e_bufferTooSmall ) { assert( false ); }
			else
			{
				if ( ptbAuthKeyID.Allocate( size ) == NULL ) return false;
			    ires = ska_generateLocalKeyID( m_pBlock->ska, ptbAuthKeyID.buffer, &size );

		        if ( ires != x509e_ok ) return false;
				bytes2string( buff, sizeof( buff ), ptbAuthKeyID.buffer, size );

				SetDlgItemText( m_hDlg, IDC_usedKeyID, buff );
			}
		}
	}

	return res;
}

bool cWizardDlg_GenKey::OnNotify( NMHDR *hdr )
{
	switch( hdr->code )
	{
		case UDN_DELTAPOS:
			if ( hdr->idFrom == IDC_spinSize )
				return UDNDeltapos( (NMUPDOWN *)hdr, IDC_edtLength, 512, 8 * 1024, 512 );
		default : break;
	}

	return FALSE;
}
