#include "secKeySN.h"
#include "BCDB_Functions.h"
#include "..\Includes\FileView.h"
#include <stdio.h>
#include <io.h>

//////////////////////////////
//
//	BCStorage serial number functions
//


//
//	SNDB file
//	every records is string with two fields
//		1. local key ID hex byte string  
//		2. corresponded number of certificates generated by the secret key with the ID.
//
const wchar_t *szSNDB_FileName = L"SNDB.txt";

//
//	sItem
//	Internal item of cSecKeySN database.
//
struct sItem
{
	BYTE *m_keyID;
	int m_keyIDSize;
	int m_num;
	sItem *m_next;

	sItem()
	{ memset( this, 0, sizeof( *this ) ); }

	static sItem *create( BYTE *keyID, int keyIDSize, int num )
	{
		sItem *t = new sItem;
		if ( t == NULL )
			return NULL;

		t->m_keyID = new BYTE[ keyIDSize ];
		if ( t->m_keyID == NULL )
		{
			delete t;
			return NULL;
		}

		memcpy( t->m_keyID, keyID, keyIDSize );

		t->m_keyIDSize = keyIDSize;
		t->m_num = num;

		return t;
	}

	~sItem()
	{
		if ( m_keyID != NULL )
			delete m_keyID;
	}

	bool writeToPtb( cPlainTextBlock *ptb )
	{
		char buff[ 1024 ];
		int len = BCDB_bytes2string( buff, sizeof( buff ), m_keyID, m_keyIDSize );
		if ( len > 1000 )
			return false;

		sprintf( buff + len, " %d\n", m_num );

		len += strlen( buff + len );

		ptb->WriteBytes( (BYTE*)buff, len );

		return true;
	}
};

//
//	class cSecKeySN
//	1. Reads file with serial numbers data base see 'SNDB file'
//	2. Finds corresponded local key ID and returns it's serial number
//	3. Adds new local key ID to DB
//	4. Writes the SNDB file.
//

cSecKeySN::~cSecKeySN()
{
	while ( m_head )
	{
		sItem *t = m_head;
		m_head = m_head->m_next;
		delete t;
	}
}

cSecKeySN::cSecKeySN()
{
	m_head = NULL;
	m_err = false;

	if ( !BCDB_GetDataBasePath( m_path, maxLen( m_path ) ) )
	{
		m_err = true;
		return;
	}

	wcscat_s( m_path, szSNDB_FileName );

	if ( !FileExistsT( m_path ) ) // does not exist
		return;

	cFileView cfv;
	if ( !cfv.Open( m_path ) )
	{
		m_err = true;
		return;
	}

	if ( !readBuffer( ( char *)cfv.m_ptr, (int)cfv.m_size.QuadPart ) )
	{
		m_err = true;
		return;
	}
}

bool cSecKeySN::addItem( BYTE *keyID, int keyIDSize, int num )
{
	sItem *t = sItem::create( keyID, keyIDSize, num );
	if ( t == NULL )
		return false;

	t->m_next = m_head;
	m_head = t;
	return true;
}

static bool hex2byte( BYTE *pByte, const char *hex )
{
	*pByte = 0;
	for( int i = 0; i<2; i++ )
	{
		BYTE bt;
		char h = hex[i];

		if ( ( h >= '0' ) && ( h <= '9' ) )
			bt = h - '0';
		else
		if ( ( h >= 'A' ) && ( h <= 'F' ) )
			bt = h - 'A' + 10;
		else
		if ( ( h >= 'a' ) && ( h <= 'f' ) )
			bt = h - 'a' + 10;
		else
			return false;

		*pByte = ( *pByte << 4 ) | bt ;
	}

	return true;
}

static bool parseHex( BYTE *keyID, int *pkeyIDSize, const char *buff, int *pPos )
{
	int size = 0;
	while( buff[ *pPos ] != ' ' )
	{
		if ( size > *pkeyIDSize )
			return false;

		if ( !hex2byte( &(keyID[size]), &(buff[*pPos]) ) )
			return false;

		size++;
		*pPos += 2;
	}

	*pkeyIDSize = size;
	*pPos += 1;

	return true;
}

static bool parseInt( int *pInt, const char *buff, int *pPos )
{
	*pInt = atoi( &(buff[ *pPos ]) );
	return true;
}

static bool getNextString( const char *buff, int *pPos, int buffSize )
{
	// Finds new line
	while( ( buff[ *pPos ] != 0x0D ) && ( buff[ *pPos ] != 0x0A ))
	{
		if ( ++(*pPos) >= buffSize )
			return false;
	}

	// Skips new line
	while( ( buff[ *pPos ] == 0x0D ) || ( buff[ *pPos ] == 0x0A ))
	{
		if ( ++(*pPos) >= buffSize )
			return false;
	}

	return true;
}

bool cSecKeySN::readBuffer( const char *buff, int buffSize )
{
	int pos = 0;
	do
	{
		BYTE keyID[ 1024 ];
		int keyIDSize = sizeof( keyID );
		if ( !parseHex( keyID, &keyIDSize, buff, &pos ) )
			return false;

		int sn;
		if ( !parseInt( &sn, buff, &pos ) )
			return false;

		if ( !addItem( keyID, keyIDSize, sn ) )
			return false;
	}
	while( getNextString( buff, &pos, buffSize ) );

	return true;
}

int cSecKeySN::findKeyID( BYTE *keyID, int size, bool increment )
{
	for ( sItem *t = m_head; t; t = t->m_next )
	{
		if ( ( t->m_keyIDSize == size ) 
			&& ( memcmp( t->m_keyID, keyID, size ) == 0 )
			)
		{
			if ( increment )
				(t->m_num)++;

			return t->m_num;
		}
	}

	if ( increment )
		addItem( keyID, size, 1 );
	
	return 0;
}

bool cSecKeySN::writeToFile()
{
	cPlainTextBlock ptb;

	for( sItem *t = m_head; t; t = t->m_next )
	{
		if ( !t->writeToPtb( &ptb ) )
			return false;
	}

	return WriteToFile( m_path, ptb.buffer, ptb.bufferSize ) != 0;
}
