/////////////////////////////////////////////////////////////////
//
//      Copyright (c) 2017 Jetico, Inc., Finland
//      All rights reserved.
//
//      File: camellia.c
//      Revision:
//      Created:	22.07.2017
//      Description:
//			RFC 3713             Camellia Encryption Algorithm            April 2004
//
/////////////////////////////////////////////////////////////////

#include "camellia.h"

#include <memory.h>

#include "../../SMemory/smemory.h"

typedef struct _sKey
{
	// For 128-bit keys, 64-bit subkeys kw1, ..., kw4, k1, ..., k18, ke1, ..., ke4 are generated as follows.
	// For 192- and 256-bit keys, 64-bit subkeys kw1, ..., kw4, k1, ..., k24, ke1, ..., ke6 are generated as follows.
	bc_u64 kw1;
	bc_u64 kw2;
	bc_u64 kw3;
	bc_u64 kw4;

	bc_u64 k1;
	bc_u64 k2;
	bc_u64 k3;
	bc_u64 k4;
	bc_u64 k5;
	bc_u64 k6;
	bc_u64 k7;
	bc_u64 k8;
	bc_u64 k9;
	bc_u64 k10;
	bc_u64 k11;
	bc_u64 k12;
	bc_u64 k13;
	bc_u64 k14;
	bc_u64 k15;
	bc_u64 k16;
	bc_u64 k17;
	bc_u64 k18;

	bc_u64 k19;
	bc_u64 k20;
	bc_u64 k21;
	bc_u64 k22;
	bc_u64 k23;
	bc_u64 k24;

	bc_u64 ke1;
	bc_u64 ke2;
	bc_u64 ke3;
	bc_u64 ke4;

	bc_u64 ke5;
	bc_u64 ke6;

} sKey;

#pragma pack (1)

#ifdef _BIG_ENDIAN_DEFINED

typedef struct _S8
{
	bc_u8 u1;
	bc_u8 u2;
	bc_u8 u3;
	bc_u8 u4;
	bc_u8 u5;
	bc_u8 u6;
	bc_u8 u7;
	bc_u8 u8;
} S8;

typedef struct _S16
{
	bc_u16 u1;
	bc_u16 u2;
	bc_u16 u3;
	bc_u16 u4;
} S16;

typedef struct _S32
{
	bc_u32 u1;
	bc_u32 u2;
} S32;

#else

typedef struct _S8
{
	bc_u8 u8;
	bc_u8 u7;
	bc_u8 u6;
	bc_u8 u5;
	bc_u8 u4;
	bc_u8 u3;
	bc_u8 u2;
	bc_u8 u1;
} S8;

typedef struct _S16
{
	bc_u16 u4;
	bc_u16 u3;
	bc_u16 u2;
	bc_u16 u1;
} S16;

typedef struct _S32
{
	bc_u32 u2;
	bc_u32 u1;
} S32;

#endif

#ifdef _BIG_ENDIAN_DEFINED
#else
#endif


typedef union _UU64
{
	bc_u64 u64;
	S32    s32;
	S16    s16;
	S8     s8;

} UU64;


#pragma pack ()


typedef int type_encrypt( const sKey *key, bc_u8 *dst, const bc_u8 *src );
static int encrypt_128( const sKey *key, bc_u8 *dst, const bc_u8 *src );
static int encrypt_256( const sKey *key, bc_u8 *dst, const bc_u8 *src );

struct _sKeyCtx
{
	type_encrypt *ecnrypt;

	//bc_u32 keyBitSize;

	// 128-bit variables of KL and KR
	bc_u64 KL[2]; // T[0] is alder T[1] is little, so rotation left bites from T[1] moves to T[0]
	bc_u64 KR[2];
	bc_u64 KA[2];
	bc_u64 KB[2];

	// Encryption and decryption operations are same but uses different reversed keys.
	sKey keyEnc;
	sKey keyDec;

};

#define useMacros
#ifdef useMacros

#define rotLeft(D0, D1, ST, SHIFT)	\
{															\
	int shift = SHIFT;										\
	if ( shift < 64 )										\
	{														\
		(D0) = (ST)[0] << shift ^ (ST)[1] >> (64 - shift);	\
		(D1) = (ST)[1] << shift ^ (ST)[0] >> (64 - shift);	\
	}														\
	else													\
	{														\
		shift -= 64;										\
		(D1) = (ST)[0] << shift ^ (ST)[1] >> (64 - shift);	\
		(D0) = (ST)[1] << shift ^ (ST)[0] >> (64 - shift);	\
	}														\
}

#else

static void rotLeft( bc_u64 &D0, bc_u64 &D1, bc_u64 *ST, int SHIFT)
{
	int shift = SHIFT;
	if ( shift < 64 )
	{				
		(D0) = (ST)[0] << shift ^ (ST)[1] >> (64 - shift);
		(D1) = (ST)[1] << shift ^ (ST)[0] >> (64 - shift);
	}
	else
	{
		shift -= 64;
		(D1) = (ST)[0] << shift ^ (ST)[1] >> (64 - shift);
		(D0) = (ST)[1] << shift ^ (ST)[0] >> (64 - shift);
	}
}

#endif

#define rotLeft32( S, SHIFT )	( (S) << (SHIFT) ^ (S) >> ( 32 - SHIFT ) )

static int makeKey_256( sKeyCtx *key )
{
	//For 192- and 256-bit keys, 64-bit subkeys kw1, ..., kw4, k1, ..., k24, ke1, ..., ke6 are generated as follows.
	sKey *subKey = &(key->keyEnc);
	key->ecnrypt = encrypt_256;

	//kw1 = (KL <<<   0) >> 64;
	//kw2 = (KL <<<   0) & MASK64;
	rotLeft( subKey->kw1, subKey->kw2, key->KL, 0 );
	//k1  = (KB <<<   0) >> 64;
	//k2  = (KB <<<   0) & MASK64;
	rotLeft( subKey->k1, subKey->k2, key->KB, 0 );
	//k3  = (KR <<<  15) >> 64;
	//k4  = (KR <<<  15) & MASK64;
	rotLeft( subKey->k3, subKey->k4, key->KR, 15 );
	//k5  = (KA <<<  15) >> 64;
	//k6  = (KA <<<  15) & MASK64;
	rotLeft( subKey->k5, subKey->k6, key->KA, 15 );
	//ke1 = (KR <<<  30) >> 64;
	//ke2 = (KR <<<  30) & MASK64;
	rotLeft( subKey->ke1, subKey->ke2, key->KR, 30 );
	//k7  = (KB <<<  30) >> 64;
	//k8  = (KB <<<  30) & MASK64;
	rotLeft( subKey->k7, subKey->k8, key->KB, 30 );
	//k9  = (KL <<<  45) >> 64;
	//k10 = (KL <<<  45) & MASK64;
	rotLeft( subKey->k9, subKey->k10, key->KL, 45 );
	//k11 = (KA <<<  45) >> 64;
	//k12 = (KA <<<  45) & MASK64;
	rotLeft( subKey->k11, subKey->k12, key->KA, 45 );
	//ke3 = (KL <<<  60) >> 64;
	//ke4 = (KL <<<  60) & MASK64;
	rotLeft( subKey->ke3, subKey->ke4, key->KL, 60 );
	//k13 = (KR <<<  60) >> 64;
	//k14 = (KR <<<  60) & MASK64;
	rotLeft( subKey->k13, subKey->k14, key->KR, 60 );
	//k15 = (KB <<<  60) >> 64;
	//k16 = (KB <<<  60) & MASK64;
	rotLeft( subKey->k15, subKey->k16, key->KB, 60 );
	//k17 = (KL <<<  77) >> 64;
	//k18 = (KL <<<  77) & MASK64;
	rotLeft( subKey->k17, subKey->k18, key->KL, 77 );
	//ke5 = (KA <<<  77) >> 64;
	//ke6 = (KA <<<  77) & MASK64;
	rotLeft( subKey->ke5, subKey->ke6, key->KA, 77 );
	//k19 = (KR <<<  94) >> 64;
	//k20 = (KR <<<  94) & MASK64;
	rotLeft( subKey->k19, subKey->k20, key->KR, 94 );
	//k21 = (KA <<<  94) >> 64;
	//k22 = (KA <<<  94) & MASK64;
	rotLeft( subKey->k21, subKey->k22, key->KA, 94 );
	//k23 = (KL <<< 111) >> 64;
	//k24 = (KL <<< 111) & MASK64;
	rotLeft( subKey->k23, subKey->k24, key->KL, 111 );
	//kw3 = (KB <<< 111) >> 64;
	//kw4 = (KB <<< 111) & MASK64;
	rotLeft( subKey->kw3, subKey->kw4, key->KB, 111 );

	return eCamelliaError_ok;
}

static int makeKey_128( sKeyCtx *key )
{
	// For 128-bit keys, 64-bit subkeys kw1, ..., kw4, k1, ..., k18, ke1, ..., ke4 are generated as follows.
	bc_u64 tmp;
	sKey *subKey = &(key->keyEnc);
	key->ecnrypt = encrypt_128;

	//kw1 = (KL <<<   0) >> 64;
	//kw2 = (KL <<<   0) & MASK64;
	rotLeft( subKey->kw1, subKey->kw2, key->KL, 0 );
	//k1  = (KA <<<   0) >> 64;
	//k2  = (KA <<<   0) & MASK64;
	rotLeft( subKey->k1, subKey->k2, key->KA, 0 );
	//k3  = (KL <<<  15) >> 64;
	//k4  = (KL <<<  15) & MASK64;
	rotLeft( subKey->k3, subKey->k4, key->KL, 15 );
	//k5  = (KA <<<  15) >> 64;
	//k6  = (KA <<<  15) & MASK64;
	rotLeft( subKey->k5, subKey->k6, key->KA, 15 );
	//ke1 = (KA <<<  30) >> 64;
	//ke2 = (KA <<<  30) & MASK64;
	rotLeft( subKey->ke1, subKey->ke2, key->KA, 30 );
	//k7  = (KL <<<  45) >> 64;
	//k8  = (KL <<<  45) & MASK64;
	rotLeft( subKey->k7, subKey->k8, key->KL, 45 );

	//k9  = (KA <<<  45) >> 64;
	rotLeft( subKey->k9, tmp, key->KA, 45 );

	//k10 = (KL <<<  60) & MASK64;
	rotLeft( tmp, subKey->k10, key->KL, 60 );

	//k11 = (KA <<<  60) >> 64;
	//k12 = (KA <<<  60) & MASK64;
	rotLeft( subKey->k11, subKey->k12, key->KA, 60 );
	//ke3 = (KL <<<  77) >> 64;
	//ke4 = (KL <<<  77) & MASK64;
	rotLeft( subKey->ke3, subKey->ke4, key->KL, 77 );
	//k13 = (KL <<<  94) >> 64;
	//k14 = (KL <<<  94) & MASK64;
	rotLeft( subKey->k13, subKey->k14, key->KL, 94 );
	//k15 = (KA <<<  94) >> 64;
	//k16 = (KA <<<  94) & MASK64;
	rotLeft( subKey->k15, subKey->k16, key->KA, 94 );
	//k17 = (KL <<< 111) >> 64;
	//k18 = (KL <<< 111) & MASK64;
	rotLeft( subKey->k17, subKey->k18, key->KL, 111 );
	//kw3 = (KA <<< 111) >> 64;
	//kw4 = (KA <<< 111) & MASK64;
	rotLeft( subKey->kw3, subKey->kw4, key->KA, 111 );

	return eCamelliaError_ok;
}


#ifdef _BIG_ENDIAN_DEFINED

# define getU64( x, pt) x = *((bc_u64*)(pt))
# define putU64( pt, x ) *((bc_u64*)pt) = x

#else // _BIG_ENDIAN_DEFINED

//
//	Functions with 64-bit shift in WIN32 environment calls function @ILT and uses 5 instructions per byte element
//	witout shift do not uses functions and just 4 instruction.
//
# define _noShift

# ifdef _noShift

/*
004F91A8  mov         eax,dword ptr [key] 
004F91AB  mov         ecx,dword ptr [keyBuff] 
004F91AE  mov         dl,byte ptr [ecx] 
004F91B0  mov         byte ptr [eax+0Fh],dl 
004F91B3  mov         eax,dword ptr [key] 
004F91B6  mov         ecx,dword ptr [keyBuff] 
004F91B9  mov         dl,byte ptr [ecx+1] 
004F91BC  mov         byte ptr [eax+0Eh],dl 
004F91BF  mov         eax,dword ptr [key] 
004F91C2  mov         ecx,dword ptr [keyBuff] 
004F91C5  mov         dl,byte ptr [ecx+2] 
004F91C8  mov         byte ptr [eax+0Dh],dl 
004F91CB  mov         eax,dword ptr [key] 
004F91CE  mov         ecx,dword ptr [keyBuff] 
004F91D1  mov         dl,byte ptr [ecx+3] 
004F91D4  mov         byte ptr [eax+0Ch],dl 
004F91D7  mov         eax,dword ptr [key] 
004F91DA  mov         ecx,dword ptr [keyBuff] 
004F91DD  mov         dl,byte ptr [ecx+4] 
004F91E0  mov         byte ptr [eax+0Bh],dl 
*/

#  define getU64( x, pt) \
{	((UU64*)(&(x)))->s8.u1 = (pt)[0]; \
	((UU64*)(&(x)))->s8.u2 = (pt)[1]; \
	((UU64*)(&(x)))->s8.u3 = (pt)[2]; \
	((UU64*)(&(x)))->s8.u4 = (pt)[3]; \
	((UU64*)(&(x)))->s8.u5 = (pt)[4]; \
	((UU64*)(&(x)))->s8.u6 = (pt)[5]; \
	((UU64*)(&(x)))->s8.u7 = (pt)[6]; \
	((UU64*)(&(x)))->s8.u8 = (pt)[7]; }

/*
	putU64( dst, D2 );
00504DA2  mov         eax,dword ptr [D2] 
00504DA5  mov         dword ptr [X],eax 
00504DAB  mov         ecx,dword ptr [ebp-18h] 
00504DAE  mov         dword ptr [ebp-328h],ecx 
00504DB4  mov         eax,dword ptr [dst] 
00504DB7  mov         cl,byte ptr [ebp-325h] 
00504DBD  mov         byte ptr [eax],cl 
00504DBF  mov         eax,dword ptr [dst] 
00504DC2  mov         cl,byte ptr [ebp-326h] 
00504DC8  mov         byte ptr [eax+1],cl 
00504DCB  mov         eax,dword ptr [dst] 
00504DCE  mov         cl,byte ptr [ebp-327h] 
00504DD4  mov         byte ptr [eax+2],cl 
00504DD7  mov         eax,dword ptr [dst] 
00504DDA  mov         cl,byte ptr [ebp-328h] 
00504DE0  mov         byte ptr [eax+3],cl 
00504DE3  mov         eax,dword ptr [dst] 
00504DE6  mov         cl,byte ptr [ebp-329h] 
*/

#  define putU64( pt, x )\
{	\
	UU64 X;			\
	X.u64 = (x);		\
	(pt)[0] = X.s8.u1;	\
	(pt)[1] = X.s8.u2;	\
	(pt)[2] = X.s8.u3;	\
	(pt)[3] = X.s8.u4;	\
	(pt)[4] = X.s8.u5;	\
	(pt)[5] = X.s8.u6;	\
	(pt)[6] = X.s8.u7;	\
	(pt)[7] = X.s8.u8;	\
}

# else // _noShift

/*
004F92CA  mov         eax,dword ptr [keyBuff] 
004F92CD  movzx       eax,byte ptr [eax] 
004F92D0  cdq              
004F92D1  mov         cl,38h 
004F92D3  call        @ILT+1720(__allshl) (49C6BDh) 
004F92D8  mov         esi,eax 
004F92DA  mov         edi,edx 
004F92DC  mov         ecx,dword ptr [keyBuff] 
004F92DF  movzx       eax,byte ptr [ecx+1] 
004F92E3  cdq              
004F92E4  mov         cl,30h 
004F92E6  call        @ILT+1720(__allshl) (49C6BDh) 
004F92EB  xor         esi,eax 
004F92ED  xor         edi,edx 
004F92EF  mov         edx,dword ptr [keyBuff] 
004F92F2  movzx       eax,byte ptr [edx+2] 
004F92F6  cdq              
004F92F7  mov         cl,28h 
004F92F9  call        @ILT+1720(__allshl) (49C6BDh) 
*/

#  define getU64( x, pt) \
	x=(((bc_u64)(pt)[0] << 56) \
	 ^ ((bc_u64)(pt)[1] << 48) \
	 ^ ((bc_u64)(pt)[2] << 40) \
	 ^ ((bc_u64)(pt)[3] << 32) \
	 ^ ((bc_u64)(pt)[4] << 24) \
	 ^ ((bc_u64)(pt)[5] << 16) \
	 ^ ((bc_u64)(pt)[6] <<  8) \
	 ^ ((bc_u64)(pt)[7]))

/*
	putU64( dst, D2 );
005051C0  mov         eax,dword ptr [D2] 
005051C3  mov         edx,dword ptr [ebp-18h] 
005051C6  mov         cl,38h 
005051C8  call        @ILT+7580(__aullshr) (49DDA1h) 
005051CD  mov         ecx,dword ptr [dst] 
005051D0  mov         byte ptr [ecx],al 
005051D2  mov         eax,dword ptr [D2] 
005051D5  mov         edx,dword ptr [ebp-18h] 
005051D8  mov         cl,30h 
005051DA  call        @ILT+7580(__aullshr) (49DDA1h) 
005051DF  mov         ecx,dword ptr [dst] 
005051E2  mov         byte ptr [ecx+1],al 
005051E5  mov         eax,dword ptr [D2] 
005051E8  mov         edx,dword ptr [ebp-18h] 
005051EB  mov         cl,28h 
005051ED  call        @ILT+7580(__aullshr) (49DDA1h) 
005051F2  mov         ecx,dword ptr [dst] 
005051F5  mov         byte ptr [ecx+2],al 
005051F8  mov         eax,dword ptr [D2] 
005051FB  mov         edx,dword ptr [ebp-18h] 
005051FE  mov         cl,20h 
00505200  call        @ILT+7580(__aullshr) (49DDA1h) 
00505205  mov         ecx,dword ptr [dst] 
00505208  mov         byte ptr [ecx+3],al 
0050520B  mov         eax,dword ptr [D2] 
0050520E  mov         edx,dword ptr [ebp-18h] 
00505211  mov         cl,18h 
00505213  call        @ILT+7580(__aullshr) (49DDA1h) */

#  define putU64( pt, x ) {	\
	(pt)[0] = (bc_u8)((x) >> 56);	\
	(pt)[1] = (bc_u8)((x) >> 48);	\
	(pt)[2] = (bc_u8)((x) >> 40);	\
	(pt)[3] = (bc_u8)((x) >> 32);	\
	(pt)[4] = (bc_u8)((x) >> 24);	\
	(pt)[5] = (bc_u8)((x) >> 16);	\
	(pt)[6] = (bc_u8)((x) >>  8);	\
	(pt)[7] = (bc_u8)( x ); }

# endif // _noShift

#endif // _BIG_ENDIAN_DEFINED

#define getU128( rT, pt ) { getU64( rT[0], pt ); getU64( rT[1], (bc_u8*)pt + 8 ); }
#define putU128( pt, rT ) { putU64( pt, rT[0] ); putU64( (bc_u8*)pt + 8, rT[1] ); }

#define MASK8 0xFF

#include "sboxTabs.h"

//
//	F_XOR( F_OUT, F_IN, KE ) <==> F_OUT = F_OUT ^ F( F_IN, KE )
//
#ifdef useMacros
#define F_XOR( F_OUT, F_IN, KE ) \
{	\
	UU64 t,y,tx;	\
	bc_u8 a;		\
	t.u64 = F_IN ^ KE;			\
	t.s8.u1 = SBOX1[t.s8.u1];	\
	t.s8.u2 = SBOX2[t.s8.u2];	\
	t.s8.u3 = SBOX3[t.s8.u3];	\
	t.s8.u4 = SBOX4[t.s8.u4];	\
	t.s8.u5 = SBOX2[t.s8.u5];	\
	t.s8.u6 = SBOX3[t.s8.u6];	\
	t.s8.u7 = SBOX4[t.s8.u7];	\
	t.s8.u8 = SBOX1[t.s8.u8];	\
	tx.u64 = t.u64;				\
	tx.s32.u1 ^= tx.s32.u2;		\
	tx.s16.u1 ^= tx.s16.u2;		\
	tx.s8 .u1 ^= tx.s8 .u2;		\
	a = tx.s8 .u1;				\
	y.s8.u1 = a         ^ t.s8.u2                       ^ t.s8.u5;								\
	y.s8.u2 = a                   ^ t.s8.u3                     ^ t.s8.u6;						\
	y.s8.u3 = a                             ^ t.s8.u4 ^                     t.s8.u7;			\
	y.s8.u4 = a^t.s8.u1                                                             ^ t.s8.u8;	\
	y.s8.u5 = a^                    t.s8.u3 ^ t.s8.u4 ^ t.s8.u5;								\
	y.s8.u6 = a^t.s8.u1                     ^ t.s8.u4           ^ t.s8.u6;						\
	y.s8.u7 = a^t.s8.u1 ^ t.s8.u2                                         ^ t.s8.u7;			\
	y.s8.u8 = a^          t.s8.u2 ^ t.s8.u3                                          ^ t.s8.u8;	\
	*F_OUT ^= y.u64;	\
}

#define FL_D( X, KE )	\
{						\
	((UU64*)(X))->s32.u2 ^= rotLeft32( ((UU64*)(X))->s32.u1 & ((UU64*)(&(KE)))->s32.u1, 1 );	\
	((UU64*)(X))->s32.u1 ^= ((UU64*)(X))->s32.u2 | ((UU64*)(&(KE)))->s32.u2;	\
}

#define FLINV_D( Y, KE )	\
{							\
	((UU64*)(Y))->s32.u1 ^= ((UU64*)(Y))->s32.u2 | ((UU64*)(&KE))->s32.u2;	\
	((UU64*)(Y))->s32.u2 ^= rotLeft32( ((UU64*)(Y))->s32.u1 & ((UU64*)(&KE))->s32.u1, 1 );	\
}

#else
static void F_XOR( bc_u64 *F_OUT, const bc_u64 F_IN, const bc_u64 KE ) 
{
	//x  = F_IN ^ KE;
	//t1 =  x >> 56;
	//t2 = (x >> 48) & MASK8;
	//t3 = (x >> 40) & MASK8;
	//t4 = (x >> 32) & MASK8;
	//t5 = (x >> 24) & MASK8;
	//t6 = (x >> 16) & MASK8;
	//t7 = (x >>  8) & MASK8;
	//t8 =  x        & MASK8;
	//t1 = SBOX1[t1];
	//t2 = SBOX2[t2];
	//t3 = SBOX3[t3];
	//t4 = SBOX4[t4];
	//t5 = SBOX2[t5];
	//t6 = SBOX3[t6];
	//t7 = SBOX4[t7];
	//t8 = SBOX1[t8];
	//y1 = t1 ^ t3 ^ t4 ^ t6 ^ t7 ^ t8;
	//y2 = t1 ^ t2 ^ t4 ^ t5 ^ t7 ^ t8;
	//y3 = t1 ^ t2 ^ t3 ^ t5 ^ t6 ^ t8;
	//y4 = t2 ^ t3 ^ t4 ^ t5 ^ t6 ^ t7;
	//y5 = t1 ^ t2 ^ t6 ^ t7 ^ t8;
	//y6 = t2 ^ t3 ^ t5 ^ t7 ^ t8;
	//y7 = t3 ^ t4 ^ t5 ^ t6 ^ t8;
	//y8 = t1 ^ t4 ^ t5 ^ t6 ^ t7;
	//F_OUT = (y1 << 56) | (y2 << 48) | (y3 << 40) | (y4 << 32) | (y5 << 24) | (y6 << 16) | (y7 <<  8) | y8;

	UU64 t,y,tx;
	bc_u8 a;

	t.u64 = F_IN ^ KE;

	t.s8.u1 = SBOX1[t.s8.u1];
	t.s8.u2 = SBOX2[t.s8.u2];
	t.s8.u3 = SBOX3[t.s8.u3];
	t.s8.u4 = SBOX4[t.s8.u4];
	t.s8.u5 = SBOX2[t.s8.u5];
	t.s8.u6 = SBOX3[t.s8.u6];
	t.s8.u7 = SBOX4[t.s8.u7];
	t.s8.u8 = SBOX1[t.s8.u8];

	tx.u64 = t.u64;
	tx.s32.u1 ^= tx.s32.u2;
	tx.s16.u1 ^= tx.s16.u2;
	tx.s8 .u1 ^= tx.s8 .u2;

	// Xor all bytes.
	a = tx.s8 .u1;

	//y.s8.u1 = t.s8.u1 ^	       t.s8.u3 ^ t.s8.u4 ^           t.s8.u6 ^ t.s8.u7 ^ t.s8.u8;
	y.s8.u1 = a         ^ t.s8.u2                       ^ t.s8.u5;

	//y.s8.u2 = t.s8.u1 ^ t.s8.u2 ^           t.s8.u4 ^ t.s8.u5 ^           t.s8.u7 ^ t.s8.u8;
	y.s8.u2 = a                   ^ t.s8.u3                     ^ t.s8.u6;

	//y.s8.u3 = t.s8.u1 ^ t.s8.u2 ^ t.s8.u3 ^           t.s8.u5 ^ t.s8.u6 ^           t.s8.u8;
	y.s8.u3 = a                             ^ t.s8.u4 ^                     t.s8.u7;

	//y.s8.u4 =           t.s8.u2 ^ t.s8.u3 ^ t.s8.u4 ^ t.s8.u5 ^ t.s8.u6 ^ t.s8.u7;
	y.s8.u4 = a^t.s8.u1                                                             ^ t.s8.u8;

	//y.s8.u5 = t.s8.u1 ^ t.s8.u2 ^                               t.s8.u6 ^ t.s8.u7 ^ t.s8.u8;
	y.s8.u5 = a^                    t.s8.u3 ^ t.s8.u4 ^ t.s8.u5;

	//y.s8.u6 =           t.s8.u2 ^ t.s8.u3 ^           t.s8.u5 ^           t.s8.u7 ^ t.s8.u8;
	y.s8.u6 = a^t.s8.u1                     ^ t.s8.u4           ^ t.s8.u6;

	//y.s8.u7 =                     t.s8.u3 ^ t.s8.u4 ^ t.s8.u5 ^ t.s8.u6 ^           t.s8.u8;
	y.s8.u7 = a^t.s8.u1 ^ t.s8.u2                                         ^ t.s8.u7;

	//y.s8.u8 = t.s8.u1 ^                     t.s8.u4 ^ t.s8.u5 ^ t.s8.u6 ^ t.s8.u7;
	y.s8.u8 = a^          t.s8.u2 ^ t.s8.u3                                          ^ t.s8.u8;

	//y.s8.u1 = t.s8.u1 ^	          t.s8.u3 ^ t.s8.u4 ^           t.s8.u6 ^ t.s8.u7 ^ t.s8.u8;
	//y.s8.u2 = t.s8.u1 ^ t.s8.u2 ^           t.s8.u4 ^ t.s8.u5 ^           t.s8.u7 ^ t.s8.u8;
	//y.s8.u3 = t.s8.u1 ^ t.s8.u2 ^ t.s8.u3 ^           t.s8.u5 ^ t.s8.u6 ^           t.s8.u8;
	//y.s8.u4 =           t.s8.u2 ^ t.s8.u3 ^ t.s8.u4 ^ t.s8.u5 ^ t.s8.u6 ^ t.s8.u7;
	//y.s8.u5 = t.s8.u1 ^ t.s8.u2 ^                               t.s8.u6 ^ t.s8.u7 ^ t.s8.u8;
	//y.s8.u6 =           t.s8.u2 ^ t.s8.u3 ^           t.s8.u5 ^           t.s8.u7 ^ t.s8.u8;
	//y.s8.u7 =                     t.s8.u3 ^ t.s8.u4 ^ t.s8.u5 ^ t.s8.u6 ^           t.s8.u8;
	//y.s8.u8 = t.s8.u1 ^                     t.s8.u4 ^ t.s8.u5 ^ t.s8.u6 ^ t.s8.u7;

	*F_OUT ^= y.u64;
}

static void FL_D( bc_u64 *X, const bc_u64 KE )
{
	((UU64*)(X))->s32.u2 ^= rotLeft32( ((UU64*)(X))->s32.u1 & ((UU64*)(&KE))->s32.u1, 1 );
	((UU64*)(X))->s32.u1 ^= ((UU64*)(X))->s32.u2 | ((UU64*)(&KE))->s32.u2;
}

static void FLINV_D( bc_u64 *Y, const bc_u64 KE )
{
	((UU64*)(Y))->s32.u1 ^= ((UU64*)(Y))->s32.u2 | ((UU64*)(&KE))->s32.u2;
	((UU64*)(Y))->s32.u2 ^= rotLeft32( ((UU64*)(Y))->s32.u1 & ((UU64*)(&KE))->s32.u1, 1 );
}

#endif

/*
	F_XOR is used instead
static bc_u64 F( const bc_u64 F_IN, const bc_u64 KE )
{
	bc_u8 t[8];
	bc_u8 y[8];

	//x  = F_IN ^ KE;
	//t1 =  x >> 56;
	//t2 = (x >> 48) & MASK8;
	//t3 = (x >> 40) & MASK8;
	//t4 = (x >> 32) & MASK8;
	//t5 = (x >> 24) & MASK8;
	//t6 = (x >> 16) & MASK8;
	//t7 = (x >>  8) & MASK8;
	//t8 =  x        & MASK8;
	//t1 = SBOX1[t1];
	//t2 = SBOX2[t2];
	//t3 = SBOX3[t3];
	//t4 = SBOX4[t4];
	//t5 = SBOX2[t5];
	//t6 = SBOX3[t6];
	//t7 = SBOX4[t7];
	//t8 = SBOX1[t8];
	//y1 = t1 ^ t3 ^ t4 ^ t6 ^ t7 ^ t8;
	//y2 = t1 ^ t2 ^ t4 ^ t5 ^ t7 ^ t8;
	//y3 = t1 ^ t2 ^ t3 ^ t5 ^ t6 ^ t8;
	//y4 = t2 ^ t3 ^ t4 ^ t5 ^ t6 ^ t7;
	//y5 = t1 ^ t2 ^ t6 ^ t7 ^ t8;
	//y6 = t2 ^ t3 ^ t5 ^ t7 ^ t8;
	//y7 = t3 ^ t4 ^ t5 ^ t6 ^ t8;
	//y8 = t1 ^ t4 ^ t5 ^ t6 ^ t7;
	//F_OUT = (y1 << 56) | (y2 << 48) | (y3 << 40) | (y4 << 32) | (y5 << 24) | (y6 << 16) | (y7 <<  8) | y8;

	putU64( t, F_IN ^ KE );
	t[0] = SBOX1[t[0]];
	t[1] = SBOX2[t[1]];
	t[2] = SBOX3[t[2]];
	t[3] = SBOX4[t[3]];
	t[4] = SBOX2[t[4]];
	t[5] = SBOX3[t[5]];
	t[6] = SBOX4[t[6]];
	t[7] = SBOX1[t[7]];
	y[0] = t[0] ^ t[2] ^ t[3] ^ t[5] ^ t[6] ^ t[7];
	y[1] = t[0] ^ t[1] ^ t[3] ^ t[4] ^ t[6] ^ t[7];
	y[2] = t[0] ^ t[1] ^ t[2] ^ t[4] ^ t[5] ^ t[7];
	y[3] = t[1] ^ t[2] ^ t[3] ^ t[4] ^ t[5] ^ t[6];
	y[4] = t[0] ^ t[1] ^ t[5] ^ t[6] ^ t[7];
	y[5] = t[1] ^ t[2] ^ t[4] ^ t[6] ^ t[7];
	y[6] = t[2] ^ t[3] ^ t[4] ^ t[5] ^ t[7];
	y[7] = t[0] ^ t[3] ^ t[4] ^ t[5] ^ t[6];

	bc_u64 Y;
	getU64( Y, y );
	return Y;
}

static bc_u64 FL( const bc_u64 FL_IN, const bc_u64 KE )
{
	bc_u32 x1, x2;
	bc_u32 k1, k2;
	x1 = (bc_u32)( FL_IN >> 32 );
	x2 = (bc_u32)( FL_IN );
	k1 = (bc_u32)( KE >> 32 );
	k2 = (bc_u32)( KE );
	x2 = x2 ^ rotLeft32( x1 & k1, 1 );
	x1 = x1 ^ (x2 | k2);

	//FL_OUT = (x1 << 32) | x2;
	return ((bc_u64)x1 << 32 ) | x2;
}

static bc_u64 FLINV( const bc_u64 FLINV_IN, const bc_u64 KE )
{
	bc_u32 y1, y2;
	bc_u32 k1, k2;
	y1 = (bc_u32)( FLINV_IN >> 32 );
	y2 = (bc_u32)( FLINV_IN );
	k1 = (bc_u32)( KE >> 32 );
	k2 = (bc_u32)( KE );
	y1 = y1 ^ (y2 | k2);
	y2 = y2 ^ rotLeft32( y1 & k1, 1);
	//FLINV_OUT = (y1 << 32) | y2;
	return ((bc_u64)y1 << 32 ) | y2;
}

*/

#define setU64T2( D, S ) { (D)[0]=(S)[0]; (D)[1]=(S)[1]; }

const bc_u64 Sigma1 = 0xA09E667F3BCC908B;
const bc_u64 Sigma2 = 0xB67AE8584CAA73B2;
const bc_u64 Sigma3 = 0xC6EF372FE94F82BE;
const bc_u64 Sigma4 = 0x54FF53A5F1D36F1C;
const bc_u64 Sigma5 = 0x10E527FADE682D1D;
const bc_u64 Sigma6 = 0xB05688C2B3E6C1FD;

static int initKey( sKeyCtx *key, const bc_u8 *keyBuff, const bc_u32 keyBitSize )
{
	bc_u64 D[2];

	memset( key, 0, sizeof( sKey ) );

	if ( keyBitSize == 128 )
	{
		//   KL = K;    KR = 0;
		getU128( key->KL, keyBuff );
		key->KR[0] = 0;
		key->KR[1] = 0;
	}
	else
	if ( keyBitSize == 192 )
	{
		//   KL = K >> 64;
		//   KR = ((K & MASK64) << 64) | (~(K & MASK64));
		getU128( key->KL, keyBuff );
		getU64( key->KR[0], keyBuff + 16 );
		key->KR[1] = ~(key->KR[0]);
	}
	else
	if ( keyBitSize == 256 )
	{
		//   KL = K >> 128;
		//   KR = K & MASK128;
		getU128( key->KL, keyBuff );
		getU128( key->KR, keyBuff + 16 );
	}
	else
		return eCamelliaError_invalidKeySize;

	// 'D[0] = D[0] ^ F(' <-> 'F_XOR(&D[0], '
	// 'D[1] = D[1] ^ F(' <-> 'F_XOR(&D[1], '

	D[0] = (key->KL[0] ^ key->KR[0]);
	D[1] = (key->KL[1] ^ key->KR[1]);
	F_XOR(&D[1], D[0], Sigma1);
	F_XOR(&D[0], D[1], Sigma2);
	D[0] = D[0] ^ key->KL[0];
	D[1] = D[1] ^ key->KL[1];
	F_XOR(&D[1], D[0], Sigma3);
	F_XOR(&D[0], D[1], Sigma4);
	//KA = (D[0] << 64) | D[1];
	setU64T2( key->KA, D );
	D[0] = (key->KA[0] ^ key->KR[0]);
	D[1] = (key->KA[1] ^ key->KR[1]);
	F_XOR(&D[1], D[0], Sigma5);
	F_XOR(&D[0], D[1], Sigma6);
	//KB = (D[0] << 64) | D[1];
	setU64T2( key->KB, D );

	if ( keyBitSize == 128 )
		return makeKey_128( key );
	else
		return makeKey_256( key );
}

//#define SWAP( X, Y ) { X ^=Y; Y^=X; X^=Y; }

static int reverseDecryptionKey( sKey *keyDec, const sKey *keyEnc, bc_u32 keyBitSize )
{
	//The decryption procedure of Camellia can be done in the same way as
	//the encryption procedure by reversing the order of the subkeys.

	memset( keyDec, 0xCC, sizeof( sKey ) );

	keyDec->kw1 = keyEnc->kw3;
	keyDec->kw2 = keyEnc->kw4;
	keyDec->kw3 = keyEnc->kw1;
	keyDec->kw4 = keyEnc->kw2;

	if ( keyBitSize == 128 )
	{
		keyDec->k1  = keyEnc->k18;
		keyDec->k2  = keyEnc->k17;
		keyDec->k3  = keyEnc->k16;
		keyDec->k4  = keyEnc->k15;
		keyDec->k5  = keyEnc->k14;
		keyDec->k6  = keyEnc->k13;
		keyDec->k7  = keyEnc->k12;
		keyDec->k8  = keyEnc->k11;
		keyDec->k9  = keyEnc->k10;
		keyDec->k10 = keyEnc->k9;
		keyDec->k11 = keyEnc->k8;
		keyDec->k12 = keyEnc->k7;
		keyDec->k13 = keyEnc->k6;
		keyDec->k14 = keyEnc->k5;
		keyDec->k15 = keyEnc->k4;
		keyDec->k16 = keyEnc->k3;
		keyDec->k17 = keyEnc->k2;
		keyDec->k18 = keyEnc->k1;
		keyDec->ke1 = keyEnc->ke4;
		keyDec->ke2 = keyEnc->ke3;
		keyDec->ke3 = keyEnc->ke2;
		keyDec->ke4 = keyEnc->ke1;
	}
	else //192- or 256-bit key:
	{
		keyDec->k1  = keyEnc->k24;
		keyDec->k2  = keyEnc->k23;
		keyDec->k3  = keyEnc->k22;
		keyDec->k4  = keyEnc->k21;
		keyDec->k5  = keyEnc->k20;
		keyDec->k6  = keyEnc->k19;
		keyDec->k7  = keyEnc->k18;
		keyDec->k8  = keyEnc->k17;
		keyDec->k9  = keyEnc->k16;
		keyDec->k10 = keyEnc->k15;
		keyDec->k11 = keyEnc->k14;
		keyDec->k12 = keyEnc->k13;

		keyDec->k13 = keyEnc->k12;
		keyDec->k14 = keyEnc->k11;
		keyDec->k15 = keyEnc->k10;
		keyDec->k16 = keyEnc->k9;
		keyDec->k17 = keyEnc->k8;
		keyDec->k18 = keyEnc->k7;
		keyDec->k19 = keyEnc->k6;
		keyDec->k20 = keyEnc->k5;
		keyDec->k21 = keyEnc->k4;
		keyDec->k22 = keyEnc->k3;
		keyDec->k23 = keyEnc->k2;
		keyDec->k24 = keyEnc->k1;

		keyDec->ke1 = keyEnc->ke6;
		keyDec->ke2 = keyEnc->ke5;
		keyDec->ke3 = keyEnc->ke4;
		keyDec->ke4 = keyEnc->ke3;
		keyDec->ke5 = keyEnc->ke2;
		keyDec->ke6 = keyEnc->ke1;
	}

	return eCamelliaError_ok;
}

//
//	Camellia_GenerateKey
//		Allocates and fills key context.
//
int Camellia_GenerateKey( sKeyCtx **pCtx, const bc_u8 *key, const bc_u32 keyBitSize )
{
	int res = -1;

	if ( ( keyBitSize != 128 ) && ( keyBitSize != 192 ) && ( keyBitSize != 256 ) )
		return eCamelliaError_invalidKeySize;
	
	if ( pCtx == NULL )
		return eCamelliaError_internal;

	*pCtx = (sKeyCtx *)SMalloc( sizeof( sKeyCtx ) );
	if ( *pCtx == NULL )
		return eCamelliaError_memoryAllocation;

	//(*pCtx)->keyBitSize = keyBitSize;

	res = initKey( *pCtx, key, keyBitSize );

	if ( res != eCamelliaError_ok )
	{
		Camellia_FreeKey( pCtx );
		return res;
	}

	return reverseDecryptionKey( &(*pCtx)->keyDec, &(*pCtx)->keyEnc, keyBitSize );
}

int Camellia_FreeKey( sKeyCtx **pCtx )
{
	if ( pCtx == NULL )
		return eCamelliaError_internal;

	memset( *pCtx, 0, sizeof( sKeyCtx ) );

	SFree( *pCtx );

	*pCtx = NULL;

	return eCamelliaError_ok;
}

static int encrypt_128( const sKey *subKey, bc_u8 *dst, const bc_u8 *src )
{
	//D1 = M >> 64;
	//D2 = M & MASK64;
	bc_u64 D1,D2;

	getU64( D1, src );
	getU64( D2, src + 8 );

	// 'F_XOR( &D2, ' <-> 'D2 = D2 ^ F('
	// 'F_XOR( &D1, ' <-> 'D1 = D1 ^ F('
	// 'FL_D( &D1' <-> 'D1 = FL(D1'
	// 'FLINV_D( &D2' <-> 'D2 = FLINV(D2'

	D1 = D1 ^ subKey->kw1;           // Prewhitening
	D2 = D2 ^ subKey->kw2;
	F_XOR( &D2, D1, subKey->k1);     // Round 1
	F_XOR( &D1, D2, subKey->k2);     // Round 2
	F_XOR( &D2, D1, subKey->k3);     // Round 3
	F_XOR( &D1, D2, subKey->k4);     // Round 4
	F_XOR( &D2, D1, subKey->k5);     // Round 5
	F_XOR( &D1, D2, subKey->k6);     // Round 6
	FL_D( &D1, subKey->ke1);        // FL
	FLINV_D( &D2, subKey->ke2);     // FLINV
	F_XOR( &D2, D1, subKey->k7);     // Round 7
	F_XOR( &D1, D2, subKey->k8);     // Round 8
	F_XOR( &D2, D1, subKey->k9);     // Round 9
	F_XOR( &D1, D2, subKey->k10);    // Round 10
	F_XOR( &D2, D1, subKey->k11);    // Round 11
	F_XOR( &D1, D2, subKey->k12);    // Round 12
	FL_D( &D1, subKey->ke3);        // FL
	FLINV_D( &D2, subKey->ke4);     // FLINV
	F_XOR( &D2, D1, subKey->k13);    // Round 13
	F_XOR( &D1, D2, subKey->k14);    // Round 14
	F_XOR( &D2, D1, subKey->k15);    // Round 15
	F_XOR( &D1, D2, subKey->k16);    // Round 16
	F_XOR( &D2, D1, subKey->k17);    // Round 17
	F_XOR( &D1, D2, subKey->k18);    // Round 18
	D2 = D2 ^ subKey->kw3;           // Postwhitening
	D1 = D1 ^ subKey->kw4;

	// C = (D2 << 64) | D1;
	// Note, limbs are reversed! 
	putU64( dst, D2 );
	putU64( dst+8, D1 );
	return eCamelliaError_ok;
}

static int encrypt_256( const sKey *subKey, bc_u8 *dst, const bc_u8 *src )
{
	//D1 = M >> 64;
	//D2 = M & MASK64;
	bc_u64 D1,D2;

	getU64( D1, src );
	getU64( D2, src + 8 );

	// 'F_XOR( &D2, ' <-> 'D2 = D2 ^ F('
	// 'F_XOR( &D1, ' <-> 'D1 = D1 ^ F('
	// 'FL_D( &D1' <-> 'D1 = FL(D1'
	// 'FLINV_D( &D2' <-> 'D2 = FLINV(D2'

	D1 = D1 ^ subKey->kw1;           // Prewhitening
	D2 = D2 ^ subKey->kw2;
	F_XOR( &D2, D1, subKey->k1);     // Round 1
	F_XOR( &D1, D2, subKey->k2);     // Round 2
	F_XOR( &D2, D1, subKey->k3);     // Round 3
	F_XOR( &D1, D2, subKey->k4);     // Round 4
	F_XOR( &D2, D1, subKey->k5);     // Round 5
	F_XOR( &D1, D2, subKey->k6);     // Round 6
	FL_D( &D1, subKey->ke1);     // FL
	FLINV_D( &D2, subKey->ke2);     // FLINV
	F_XOR( &D2, D1, subKey->k7);     // Round 7
	F_XOR( &D1, D2, subKey->k8);     // Round 8
	F_XOR( &D2, D1, subKey->k9);     // Round 9
	F_XOR( &D1, D2, subKey->k10);    // Round 10
	F_XOR( &D2, D1, subKey->k11);    // Round 11
	F_XOR( &D1, D2, subKey->k12);    // Round 12
	FL_D( &D1, subKey->ke3);     // FL
	FLINV_D( &D2, subKey->ke4);     // FLINV
	F_XOR( &D2, D1, subKey->k13);    // Round 13
	F_XOR( &D1, D2, subKey->k14);    // Round 14
	F_XOR( &D2, D1, subKey->k15);    // Round 15
	F_XOR( &D1, D2, subKey->k16);    // Round 16
	F_XOR( &D2, D1, subKey->k17);    // Round 17
	F_XOR( &D1, D2, subKey->k18);    // Round 18
	FL_D( &D1, subKey->ke5);     // FL
	FLINV_D( &D2, subKey->ke6);     // FLINV
	F_XOR( &D2, D1, subKey->k19);    // Round 19
	F_XOR( &D1, D2, subKey->k20);    // Round 20
	F_XOR( &D2, D1, subKey->k21);    // Round 21
	F_XOR( &D1, D2, subKey->k22);    // Round 22
	F_XOR( &D2, D1, subKey->k23);    // Round 23
	F_XOR( &D1, D2, subKey->k24);    // Round 24
	D2 = D2 ^ subKey->kw3;           // Postwhitening
	D1 = D1 ^ subKey->kw4;

	// C = (D2 << 64) | D1;
	// Note, limbs are reversed! 
	putU64( dst, D2 );
	putU64( dst+8, D1 );

	return eCamelliaError_ok;
}

void Camellia_Encrypt_ECB( const sKeyCtx *ctx, bc_u8 *dst, const bc_u8 *src )
{
	ctx->ecnrypt( &(ctx->keyEnc), dst, src );
}

void Camellia_Decrypt_ECB( const sKeyCtx *ctx, bc_u8 *dst, const bc_u8 *src )
{
	ctx->ecnrypt( &(ctx->keyDec), dst, src );
}

static const struct {
  int keylen;
  unsigned char key[32], pt[16], ct[16];
} tests[] = {
	{  /* CAMELLIA-256-ECB:key=603DEB1015CA71BE2B73AEF0857D77811F352C073B6108D72D9810A30914DFF4:plaintext=F69F2445DF4F9B17AD2B417BE66C3710:ciphertext=7960109FB6DC42947FCFE59EA3C5EB6B */
	   32,
	   { 0x60, 0x3D, 0xEB, 0x10, 0x15, 0xCA, 0x71, 0xBE, 0x2B, 0x73, 0xAE, 0xF0, 0x85, 0x7D, 0x77, 0x81, 0x1F, 0x35, 0x2C, 0x07, 0x3B, 0x61, 0x08, 0xD7, 0x2D, 0x98, 0x10, 0xA3, 0x09, 0x14, 0xDF, 0xF4 },
	   { 0xF6, 0x9F, 0x24, 0x45, 0xDF, 0x4F, 0x9B, 0x17, 0xAD, 0x2B, 0x41, 0x7B, 0xE6, 0x6C, 0x37, 0x10 },
	   { 0x79, 0x60, 0x10, 0x9F, 0xB6, 0xDC, 0x42, 0x94, 0x7F, 0xCF, 0xE5, 0x9E, 0xA3, 0xC5, 0xEB, 0x6B  },
	},
	{
		16,
		{ 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{ 0x6C, 0x22, 0x7F, 0x74, 0x93, 0x19, 0xA3, 0xAA, 0x7D, 0xA2, 0x35, 0xA9, 0xBB, 0xA0, 0x5A, 0x2C },
	},
	// Camellia tests from RFC3713
	{ //CAMELLIA-128-ECB:0123456789abcdeffedcba9876543210::0123456789abcdeffedcba9876543210:67673138549669730857065648eabe43
		128/8,
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0x67, 0x67, 0x31, 0x38, 0x54, 0x96, 0x69, 0x73, 0x08, 0x57, 0x06, 0x56, 0x48, 0xea, 0xbe, 0x43 }
	},
	{ // CAMELLIA-192-ECB:0123456789abcdeffedcba98765432100011223344556677::0123456789abcdeffedcba9876543210:b4993401b3e996f84ee5cee7d79b09b9
		192 / 8,
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77 },
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0xb4, 0x99, 0x34, 0x01, 0xb3, 0xe9, 0x96, 0xf8, 0x4e, 0xe5, 0xce, 0xe7, 0xd7, 0x9b, 0x09, 0xb9 },
	},
	{ // CAMELLIA-256-ECB:0123456789abcdeffedcba987654321000112233445566778899aabbccddeeff::0123456789abcdeffedcba9876543210:9acc237dff16d76c20ef7c919e3a7509
		256/8,
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10, 0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff },
		{ 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef, 0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10 },
		{ 0x9a, 0xcc, 0x23, 0x7d, 0xff, 0x16, 0xd7, 0x6c, 0x20, 0xef, 0x7c, 0x91, 0x9e, 0x3a, 0x75, 0x09 }
	},
};

int Camellia_SelfTest()
{
	bc_u8 enc[16] = {0};
	bc_u8 dec[16] = {0};

	sKeyCtx *ctx = NULL;

	int res = -1;

	int i;

	for( i = sizeof( tests )/sizeof(tests[0]); i--; )
	{
		res = Camellia_GenerateKey( &ctx, tests[i].key, tests[i].keylen * 8 );

		if ( res != eCamelliaError_ok )
			return res;

		Camellia_Encrypt_ECB( ctx, enc, tests[i].pt );

		if ( memcmp( enc, tests[i].ct, sizeof( enc ) ) != 0 )
			res = eCamelliaError_test;

		if ( res == eCamelliaError_ok )
		{
			Camellia_Decrypt_ECB( ctx, dec, enc );

			if ( memcmp( dec, tests[i].pt, sizeof( dec ) ) != 0 )
				res = eCamelliaError_test;
		}

		Camellia_FreeKey( &ctx );

		if ( res != eCamelliaError_ok )
			break;
	}

	return res;
}
