#include "textEncode.h"
#include "../symmetricAlgs/symmetricAlgs.h"
#include "../PGPPkts/string2key.h"
#include "../PGPPkts/pktdefs.h"
#include "../PGPPkts/PGPPkts.h"
#include "../PGPPkts/cPacket.h"
#include "../PGPPkts/Packets.h"
#include "../crtLibErrs.h"
#include "../SMemory/SBlocks.h"
#include <memory.h>
#include "../symmetricAlgs/symmetricAlgs.h"
#include "../publicKeyAlgs/publicKeyAlgs.h"
#include "../RNG/cRNG.h"
#include "../CompressAlgs/compressAlgs.h"
#include "../x509pkts/base64funcs.h"
#include "sessionKey.h"

#ifdef _ANDROID_LOG
#	include "../../../fs-lib/log.h"
#else
#	define __android_log_print(...) void(0)
#	define LIBLOG_INFO(...) void(0)
#	define LIBLOG_ERROR(...) void(0)
#	define LIBLOG_WARN(...) void(0)
#	define LIBLOG_FUNC(...) void(0)
#	define LIBLOG_FUNC_RES(...) void(0)
#	define __odsPrint(...) void(0)
#endif


static const char *szPGP_B64_Begin = "-----BEGIN PGP MESSAGE-----";
static const char *szPGP_B64_End   = "-----END PGP MESSAGE-----";

static const char *sz_B64_Begin = "-----BEGIN ENCODED MESSAGE-----";
static const char *sz_B64_End   = "-----END ENCODED MESSAGE-----";

//
//	1. Creates public key encrypted session key packet( tag 1 ), see RFC 2440
//		1.1. Generates session key for encryption
//		1.2. Encrypts it by public key
//		1.3. Forms the packet
//			[0-0] Version number ( v.3 )
//			[1-8] Public key ID
//			[9-9] Public key algorithm ID ( RSA 1 )
//			[10-end] Encrypted vlue ( one mpi for RSA )
//			Original dat packet:
//				[0-0] Sym key algorithm ID
//				[1-(end-2)] Sym key
//				[end-1-end] Check summ of byte [1-(end-2)] mod 0x10000
//
//	2. Creates Sym Encrypted and Integrity Protected Data paket ( tag 18 )
//		2.1. Creates Literal Data Packet ( tag 11 )
//			Forms the packet body as binary data:
//				[ 0 - 0 ] one octet binary data type 0x62 ('b')
//				[ 1 - 4 ] four bytes time stamp
//				[ 5 - end ] UTF8 encoded input widee char data
//
//		2.2. Creates compressed packet ( tag 8 )
//			[ 0 - 0 ] one octet compression algotrithm identifier ( ZLIB = 2 )
//			[ 1 - end ] compressed data
//
//		2.3. Creates Sym Encrypted and Integrity Protected Data paket ( tag 18 )
//			[0-0] one octet version number ( 1 )
//			[1-end] encrypted packet
//			Sym Encrypted packed original data
//				[ 0 - blockSize - 1 ]			blockSize octets random syncro
//				[ blockSize - blockSize + 1 ]	two octets are copy of two octet before
//				[ blockSize + 2 - end ]			CFB encrypted data
//

//
//	generateSK_createPKESK
//		1. Generates session key and create sym key algorithm (pSA)
//		2. Encrypts session key by public key and writes it to ptbPkt
//			as Public Key Encrypted Session Key Packet.
//
static int generateSK_createPKESK( BYTE saAlgID, HSA *pSA, cPlainTextBlock *ptbPkt, HPKA pka, BYTE *keyID, const unsigned char *seed, int *pSeedSize )
{
	// BYTE saAlgID = pgpsaAES_256;

	int keySize = sa_getKeySize( saAlgID );

	if ( *pSeedSize < keySize )
		return pkae_randomBufferRequired;

	const unsigned char *key = seed;

	seed += keySize;
	*pSeedSize -= keySize;

	int msgSize = pka_getBlockSize( pka );

	cPTB adMsg( msgSize );
	if ( adMsg.buffer == NULL )
		return ppe_memoryAllocation;

	int res = pgp_SessionKeyBlock_Format( adMsg.buffer, msgSize, saAlgID, key, keySize );
	if ( res != pkae_ok ) return res;

	// 1.2.
	int encBuffSize = 0;
	res = pka_encrypt( pka, NULL, &encBuffSize, adMsg.buffer, msgSize );
	if ( res != pkae_bufferSizeTooSmall ) return res;

	cPTB adEncBuff( encBuffSize );
	if ( adEncBuff.buffer == NULL )
		return ppe_memoryAllocation;

	res = pka_encrypt( pka, adEncBuff.buffer, &encBuffSize, adMsg.buffer, msgSize );

	//
	//	29.12.2010
	//	ElGamal encryption required random buffer for koeffs size of the random buffer has to same as prime size.
	//
	if ( res == pkae_randomBufferRequired )
	{
		int blockSize = cRNG::m_lastRequiredSize;

		if ( *pSeedSize < blockSize )
			return pkae_randomBufferRequired;

		*pSeedSize -= blockSize;

		cRNG::setPool( (BYTE*)seed, blockSize );

		res = pka_encrypt( pka, adEncBuff.buffer, &encBuffSize, adMsg.buffer, msgSize );
	}

	if ( res != pkae_ok ) return res;

	int pktSize = 0;

	// 1.3.
	res = PKT_PublicKeyEncryptedSessionKey_Format( NULL, &pktSize, keyID, (BYTE)pka_getAlgID( pka ), adEncBuff.buffer, encBuffSize );
	if ( res != pkae_bufferSizeTooSmall ) return res;

	cPGPPacket pgpPkt;

	pgpPkt.Init( pkt_PublicKeyEncryptedSessionKey );

	BYTE *space = pgpPkt.GetSpace( pktSize );

	if ( space == NULL )
		return ppe_memoryAllocation;

	res = PKT_PublicKeyEncryptedSessionKey_Format( (sPublicKeyEncryptedSessionKey*)space, &pktSize, keyID, (BYTE)pka_getAlgID( pka ), adEncBuff.buffer, encBuffSize );
	if ( res != pkae_ok ) return res;

	if ( !pgpPkt.Final( ptbPkt ) )
		return ppe_internalError;

	res = sa_createAlgorithm( pSA, saAlgID, (BYTE*)key, keySize );
	if ( res != pkae_ok ) return res;

	return ppe_ok;
}

//
//	createSKE
//		See above 2. Creates Sym Encrypted and Integrity Protected Data packet ( tag 18 )
//
static int createSKE( cPlainTextBlock *ptbPkt, HSA sa, const char *szData, int size, const unsigned char *seed, int *pSeedSize )
{
	cPlainTextBlock ptbReady;

	// 2.1

	cPGPPacket pgpPkt;

	if ( !pgpPkt.Init( pkt_LiteralData ) )
		return ppe_internalError;

	int spaceSize = 0;

	int res = pgp_LiteralDataPacket_Format( NULL, &spaceSize, -1, NULL, 0, (BYTE*)szData, size );

	if ( res != ppe_bufferTooSmall )
		return res;

	BYTE *space = pgpPkt.GetSpace( spaceSize );

	if ( *pSeedSize < sizeof(DWORD) )
		return pkae_randomBufferRequired;

	DWORD dwSalt = *(DWORD*)seed;
	*pSeedSize -= sizeof(DWORD);

	res = pgp_LiteralDataPacket_Format( space, &spaceSize, -1, NULL, dwSalt, (BYTE*)szData, size );

	if ( res != ppe_ok )
		return res;

	ptbReady.Free();
	if ( !pgpPkt.Final( &ptbReady ) ) return ppe_internalError;

	// 2.2.

	if ( !pgpPkt.Init( pkt_CompressedData ) ) return ppe_internalError;

	BYTE algId = caID_ZLIB;

	spaceSize = ca_getMaxCompressedSize( algId, ptbReady.bufferSize );

	// See pgp_CompressedPacket_Parse

	cPTB adCompressed( spaceSize + 1 );
	if ( adCompressed.buffer == NULL ) return ppe_memoryAllocation;

	HCA ca;

	res = ca_createAlgorithm( &ca, algId );
	if ( res != cae_ok ) return res;

	adCompressed.buffer[0] = algId;

	bool ok = ( ca_compress( ca, adCompressed.buffer + 1, &spaceSize, ptbReady.buffer, ptbReady.bufferSize ) == cae_ok );
	ca_free( ca );

	//
	//	It does not compress if compressed size is more then original size.
	//
	//	12.10.2015 OpenPGP 'Kleaopatra' opens but show error if literal packed was not compressed.
	//
	//if ( ok && ( ptbReady.bufferSize > spaceSize ) )
	if ( ok )
	{
		if ( !pgpPkt.Add( adCompressed.buffer, spaceSize + 1 ) ) return ppe_memoryAllocation;

		ptbReady.Free();

		//
		//	12.10.2015 OpenPGP 'Kleaopatra' does not compressed packed other then with undefined size.
		//
		bool undefinedSize = true;
		if ( !pgpPkt.Final( &ptbReady, undefinedSize ) ) return ppe_memoryAllocation;
	}

	// 2.3.

	if ( !pgpPkt.Init( pkt_SymEncryptedAndIntegrityProtectedData ) ) return ppe_memoryAllocation;

	if ( !pgpPkt.Add( (BYTE)1 ) ) return ppe_memoryAllocation;

	int blockSize = sa_getBlockSize( sa );

	// Prepares random buffer.
	if ( *pSeedSize < blockSize )
		return pkae_randomBufferRequired;

	*pSeedSize -= blockSize;

	cRNG::setPool( (unsigned char *)seed, blockSize );

	int pktSize = 0;

	res = pgp_SymEncryptedAndIntegrityProtectedDataPacket_Format(
						NULL, &pktSize, blockSize, ptbReady.buffer, ptbReady.bufferSize );

	if ( res != ppe_bufferTooSmall )
		return res;

	cPTB adbf( pktSize );
	if ( adbf.buffer == NULL ) return ppe_memoryAllocation;

	res = pgp_SymEncryptedAndIntegrityProtectedDataPacket_Format(
				adbf.buffer, &pktSize, blockSize, ptbReady.buffer, ptbReady.bufferSize );

	// Uses zero initial vector.
	res = sa_initCFB( sa, NULL );
	if ( res != sae_ok ) return res;

	space = pgpPkt.GetSpace( pktSize );
	if ( space == NULL ) return ppe_memoryAllocation;

	res = sa_encryptCFB( sa, space, adbf.buffer, pktSize );
	if ( res != sae_ok ) return res;

	if ( !pgpPkt.Final( ptbPkt ) )
		return ppe_memoryAllocation;

	return ppe_ok;
}

// bool EncryptDataPGP( HWND hWnd, cWCStr *encoded, const wchar_t *szData, int size, HPKA pka, BYTE *keyID )
int encodeTextByPublicKey(
		char *encodedBuffer, int &encodedBufferSize,
		const char *plainTextBuffer, int plainTextBufferSize,
		int saAlgID,
		HPKA pka, unsigned char *keyID,
		const unsigned char *seed, int seedSize,
		const char *szVersionInfo
	)
{
	cPlainTextBlock ptbPkt;
	HSA sa = NULL;

	if ( saAlgID == -1 )
		saAlgID = pgpsaAES_256;

	// 1.
	int res = generateSK_createPKESK( (BYTE)saAlgID, &sa, &ptbPkt, pka, keyID, seed, &seedSize );
	if ( res != ppe_ok )
		return res;

	//
	//	2. Creates Sym Encrypted and Integrity Protected Data packet ( tag 18 )
	//
	res = createSKE( &ptbPkt, sa, plainTextBuffer, plainTextBufferSize, seed, &seedSize );

	sa_free( sa );

	if ( res != ppe_ok )
		return res;

	bool keyIDIsEmpty = true;
	for( int i = pgp_keyIDSize; i--; )
	{
		if ( keyID[i] != 0 )
		{
			keyIDIsEmpty = false;
			break;
		}
	}

	const char *szBegin = sz_B64_Begin;
	const char *szEnd = sz_B64_End;

	if ( !keyIDIsEmpty )
	{
		// If key id was from PGP key ring so recipient is PGP user and can uses own PGP encode utility.
		// So we conclude armored block in PGP compatible strings.
		szBegin = szPGP_B64_Begin;
		szEnd = szPGP_B64_End;
	}

	cPlainTextBlock ptbB64Encoded;

	if ( !createArmoredBlock( &ptbB64Encoded, ptbPkt.buffer, ptbPkt.bufferSize, szBegin, szEnd, szVersionInfo ) )
		return ppe_internalError;

	if ( encodedBufferSize < ptbB64Encoded.bufferSize )
	{
		encodedBufferSize = ptbB64Encoded.bufferSize;
		return ppe_bufferTooSmall;
	}

	encodedBufferSize = ptbB64Encoded.bufferSize;
	memcpy( encodedBuffer, ptbB64Encoded.buffer, ptbB64Encoded.bufferSize );

	return ppe_ok;
}

static int makePasswordEncryptedSessionKeyPacket(
		cPlainTextBlock *ptEncBlock,
		BYTE *sessKeyBuffer, int sessKeyBufferSize,
		int keyAlgID,
		char *pwd,
		BYTE hashAlgID	= pgphaSHA1,
		BYTE *salt		= NULL,
		int iterations	=  -1
	)
{
	//
	//	Check size of required packet.
	//
	int pktSize = 0;
	BYTE algID = (BYTE)keyAlgID;

	int res = pgp_FormatSymmetricKeyEncrytpedSessionKeyPacket(
				NULL, &pktSize,
				algID,
				NULL, sessKeyBufferSize,
				keyAlgID,
				pwd,
				hashAlgID,
				salt,
				iterations
			);

	if ( res != pkae_bufferSizeTooSmall ) return res;

	int pktWithHeaderSize = 0;

	res = pgp_PacketHeader_Format(
		NULL, &pktWithHeaderSize,
		pkt_SymmetricKeyEncryptedSessionKey,
		NULL, pktSize );

	if ( res != pkae_bufferSizeTooSmall ) return res;

	int hdrSize = pktWithHeaderSize - pktSize;

	if ( ptEncBlock->Allocate( pktWithHeaderSize ) == NULL )
	{
		return pkae_memoryAllocation;
	}

	res = pgp_FormatSymmetricKeyEncrytpedSessionKeyPacket(
				ptEncBlock->buffer + hdrSize, &pktSize,
				algID,
				sessKeyBuffer, sessKeyBufferSize,
				keyAlgID,
				pwd,
				hashAlgID,
				salt,
				iterations
			);

	if ( res != ppe_ok ) return res;

	res = pgp_PacketHeader_Format(
		ptEncBlock->buffer, &pktWithHeaderSize,
		pkt_SymmetricKeyEncryptedSessionKey,
		ptEncBlock->buffer + hdrSize, pktSize );

	return res;
}

static bool checkSesKeyPacket( BYTE *pkt, int pktSize, BYTE *key, int keySize, BYTE saAlgID, const char *pwd )
{
	//LIBLOG_FUNC( "%p, %d, %d, %d, %d", pkt, pktSize, keySize, saAlgID, strlen(pwd) );

	BYTE tag;
	BYTE *data;
	int dataSize;
	int hdrSize;


	bool partial = false;

	int res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt, pktSize );
	if ( res != ppe_ok )
	{
		LIBLOG_INFO( "Error %d", res );
		return false;
	}

	data = pkt + hdrSize;

#ifdef _DEBUG
	const char *tagName;
	res = PKT_GetTagName( tag, &tagName );
#endif

	if ( dataSize == -1 ) // undefined size
	{
		dataSize = pktSize - hdrSize;
	}

	BYTE chk_sesKeyAlgID;
	int chk_sesKeySize = keySize;
	cPTB adChk( chk_sesKeySize );
	BYTE *chk_sesKeyBuff = adChk.buffer;

	if ( chk_sesKeyBuff == NULL )
	{
		LIBLOG_INFO( "sesKeyBuff == NULL" );
		return false;
	}

	res = pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket( &chk_sesKeyAlgID, chk_sesKeyBuff, &chk_sesKeySize, data, dataSize, pwd );

	if (
		( res != ppe_ok ) ||
		( chk_sesKeySize != keySize ) ||
		( chk_sesKeyAlgID != saAlgID ) ||
		( memcmp( chk_sesKeyBuff, key, chk_sesKeySize ) != 0 )
		)
	{
		LIBLOG_INFO( "Error res = %d, chk_sesKeySize = %d, chk_sesKeyAlgID %d, memcmp %d", res, chk_sesKeySize, chk_sesKeyAlgID, memcmp( chk_sesKeyBuff, key, chk_sesKeySize ) );
		LIBLOG_INFO( bytes2string( chk_sesKeyBuff,	keySize, keySize, true ) );
		LIBLOG_INFO( bytes2string( key, 			keySize, keySize, true ) );
		return false;
	}

	return true;
}


int encodeTextByPassword(
		cPTB &ptbB64Encoded,
		const char *plainTextBuffer, int plainTextBufferSize,
		int saAlgID,
		const char *utf8pwd,
		const unsigned char *seedBuffer, int seedBufferSize,
		const char *szVersionInfo )
{

	LIBLOG_FUNC( "'%s', %d, '%s', %d, %s", plainTextBuffer, saAlgID, utf8pwd, seedBufferSize, szVersionInfo );

	bool bIsZero = true;
	for( int i = 0; i < seedBufferSize; i++ )
	{
		if ( seedBuffer[i] != 0 )
		{
			bIsZero = false;
			break;
		}
	}

#ifdef _ANDROID_LOG

	if ( bIsZero )
		LIBLOG_INFO( "Random buffer is zero!" );
	static cPTB chk;
	if ( chk.bufferSize >= seedBufferSize )
	{
		if ( memcmp( chk.buffer, seedBuffer, seedBufferSize ) == 0 )
			LIBLOG_ERROR( "Assertion! Random is same as before!" );
	}

	chk.WriteBytes( seedBuffer, seedBufferSize, true );

#else
	if ( bIsZero )
		return ppe_randomBufferRequired;
#endif


	if ( saAlgID == -1 )
		saAlgID = pgpsaAES_256;

	int res = 0;

	int keySize = sa_getKeySize( saAlgID );
	int saltSize = sizeof( ((S2K*)(0))->salt );

	if ( keySize + saltSize > seedBufferSize )
	{
		LIBLOG_INFO( "Random buffer is required!" );
		return ppe_randomBufferRequired;
	}

	const unsigned char *salt = seedBuffer + keySize;

	seedBuffer += keySize + saltSize;
	seedBufferSize -= keySize + saltSize;

	cPlainTextBlock ptbEncBlock;

	// 2. Prepares pwd encrypted session key packet.
	res = makePasswordEncryptedSessionKeyPacket( &ptbEncBlock, (BYTE *)seedBuffer, keySize, saAlgID, (char *)utf8pwd, pgphaSHA1, (BYTE*)salt );
	if ( res != ppe_ok )
	{
		LIBLOG_INFO( "Error %d", res );
		return res;
	}

	// 3. Check prepared packet and extract session key back.
	if ( !checkSesKeyPacket( ptbEncBlock.buffer, ptbEncBlock.bufferSize, (BYTE *)seedBuffer, keySize, (BYTE)saAlgID, utf8pwd ) )
	{
		LIBLOG_INFO( "Check session key packet error!" );
		return ppe_internalError;
	}

	HSA sa;
	res = sa_createAlgorithm( &sa, saAlgID, (unsigned char *)seedBuffer, keySize );
	if ( res != pkae_ok )
	{
		LIBLOG_INFO( "Error %d", res );
		return res;
	}

	//
	//	2. Creates Sym Encrypted and Integrity Protected Data packet ( tag 18 )
	//
	res = createSKE( &ptbEncBlock, sa, plainTextBuffer, plainTextBufferSize, seedBuffer, &seedBufferSize );

	sa_free( sa );

	if ( res != ppe_ok )
	{
		LIBLOG_INFO( "Error %d", res );
		return res;
	}

	if ( !createArmoredBlock( &ptbB64Encoded, ptbEncBlock.buffer, ptbEncBlock.bufferSize, sz_B64_Begin, sz_B64_End, szVersionInfo ) )
	{
		LIBLOG_INFO( "Error in createArmoredBlock" );
		return ppe_internalError;
	}

	LIBLOG_FUNC_RES( "Ok!");
	return ppe_ok;
}

//--------------------------------------- Decoding --------------------------------------------------------//

// 
//	Stolen from int cDecryptDlg::ParsePacketBody( BYTE tag, BYTE *pkt, int pktSize, int recCount, bool partial, cPTB *ptb )
//
static int parsePacketBody( BYTE tag, BYTE *pkt, int pktSize, bool partial, cPTB &ptb, cSessionKey &sesKey, cSecretKey &secretKey )
{
	int res = sae_ok;

	switch ( tag )
	{
		case pkt_LiteralData:
		{
			//  - A one-octet field that describes how the data is formatted.

			//If it is a 'b' (0x62), then the literal packet contains binary data.
			//If it is a 't' (0x74), then it contains text data, and thus may need
			//line ends converted to local form, or other text-mode changes. The
			//tag 'u' (0x75) means the same as 't', but also indicates that
			//implementation believes that the literal data contains UTF-8 text.

			if ( ptb.WriteBytes( pkt, pktSize, !partial ) == 0 )
				return RSA_MALLOC_ERROR;

			return ppe_ok;
		}

		case pkt_CompressedData:
		{
			if ( ptb.WriteBytes( pkt, pktSize, !partial ) == 0 )
				return RSA_MALLOC_ERROR;

			return ppe_ok;
		}

		case pkt_SymEncryptedAndIntegrityProtectedData :
		{
			// see rfc2440bis-22 

			if ( sesKey.hsa == NULL )
				return ppe_invalidBlockType;

			if ( !partial )
			{
				if ( pkt[0] != 1 )
					return ppe_invalidBlockType;

				res = sa_initCFB( sesKey.hsa, NULL );

				pkt++;
				pktSize--;
			}

			cPTB ptbData;
			ptbData.Allocate( pktSize );

			if ( res == sae_ok )
				res = sa_decryptCFB( sesKey.hsa, ptbData.buffer, pkt, pktSize );

			if ( !partial )
			{
				int blockSize = sa_getBlockSize( sesKey.hsa );
				if ( ptbData.bufferSize < blockSize + 2 )
					return ppe_invalidBlockType;

				if ( memcmp( ptbData.buffer + blockSize - 2, ptbData.buffer + blockSize, 2 ) != 0 )
					return ppe_invalidSignature; // ppe_invalidBlockType; v.1.02.1.1 - 03.10.2016
			}

			if ( res == sae_ok )
				ptb.WriteBytes( ptbData.buffer, ptbData.bufferSize );
			
			return res;
		}
		case pkt_SymmetricKeyEncryptedSessionKey:
		{
			int sesKeySize = 0;
			BYTE sesKeyAlgID = 0;

			res = pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket( &sesKeyAlgID, NULL, &sesKeySize, pkt, pktSize, NULL );

			if ( res != ppe_bufferTooSmall )
				return res;

			cPTB ptb;
			BYTE *sesKeyBuff = ptb.Allocate( sesKeySize );

			if ( sesKeyBuff == NULL )
			{
				assert( false );
				return pkae_memoryAllocation;
			}

			res = pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket( &sesKeyAlgID, sesKeyBuff, &sesKeySize, pkt, pktSize, sesKey.pwd.buffer );

			if ( ( res == ppe_ok ) && !sa_isValidID( sesKeyAlgID ) )
				res = ppe_invalidPassword;

			if ( res != ppe_ok )
				return res;

			res = sesKey.createAlgorithm( sesKeyAlgID, sesKeyBuff, sesKeySize );

			if ( res != ppe_ok )
				return res;
			
			//m_decodedBy.Set( _LoadIDS( IDS_szDecodedByPwd ) );

			return res;
		}

		case pkt_PublicKeyEncryptedSessionKey :
		{
//			return x509e_featureIsNotImplemented;

			sPublicKeyEncryptedSessionKey *pkesk = ( sPublicKeyEncryptedSessionKey * )pkt;

			int keyMpiSize;

			res = PKT_PublicKeyEncryptedSessionKey_Parse( pkesk, pktSize, &keyMpiSize );

			if ( res != ppe_ok )
				return res;

			cPTB ptb;
			BYTE *sesKeyBuff = NULL;
			int sesKeySize = 0;
			BYTE sesKeyAlgID = 0;

			// Public key ID may be zero. In this case, the receiving implementation 
			//	would try all available private keys.
			bool isZero = true;

			for( int i = pgp_keyIDSize; i--; )
			{
				if ( pkesk->keyID[i] != 0 )
				{
					isZero = false;
					break;
				}
			}

			//
			//	Corresponded keys has to be added before by cSecretKey::AddKey
			//
			if ( !secretKey.openSecretKey( pkesk->keyID ) )
				return RSA_INVALID_KEY_BLOCK;

			int keyN = -1;

			if ( !isZero && !secretKey.findKey( pkesk->keyID, &keyN ) )
			{
				//if ( IDYES != IH.ErrorMessage( IDS_PKID_IsIncompatible_Q, IDS_AppName, MB_YESNO ) )
				//	continue;
				return ppe_keyNotFound;
			}

			for( int kn = secretKey.skaQ; kn--; )
			{
				res = RSA_INVALID_KEY_BLOCK;

				if ( keyN != -1 ) // To check the found key only. See below.
				{
					if ( keyN > secretKey.skaQ )
					{
						assert( false );
						break;
					}

					kn = keyN;
				}

				int decSize = 0;

				HPKA ska = secretKey.skaT[kn]->m_ska; // ska;

				res = ska_decrypt( ska, NULL, &decSize, pkesk->mpi, keyMpiSize );
				if ( res != pkae_bufferSizeTooSmall )
					return res;
				
				if ( ptb.Allocate( decSize ) == NULL )
					return RSA_MALLOC_ERROR;

				res = ska_decrypt( ska, ptb.buffer, &decSize, pkesk->mpi, keyMpiSize );
				if ( res != pkae_ok )
					continue; //return res;

				res = RSA_SessionKeyBlock_Parse( &sesKeyBuff, &sesKeySize, &sesKeyAlgID, ptb.buffer, decSize );

				if ( res == pkae_ok )
				{
					// lastUsedName = secretKey.skaT[ kn ]->m_name;
					break;
				}

				if ( keyN != -1 ) // To check the found key only. See above.
					break;
			}

			if ( res == pkae_ok )
				return res;				

			res = sesKey.createAlgorithm( sesKeyAlgID, sesKeyBuff, sesKeySize );

			return res;
		}
	}

	return RSA_INVALID_KEY_BLOCK;
}

//
//	Stolen from int cDecryptDlg::ParsePacket( BYTE *pkt, int totalPktSize, int *parsed, cPTB *ptb )
//
static int parsePacket( BYTE *pkt, int totalPktSize, int *parsed, cPTB &ptb, cSessionKey &sesKey, cSecretKey &secretKey )
{
	BYTE tag;
	BYTE *data;
	int dataSize;
	int hdrSize;
	bool partial = false;

	*parsed = 0;

	int res;

	cPTB ptbData;
	
	do
	{
		res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt + *parsed, totalPktSize );

		if ( res != ppe_ok ) return res;

		data = pkt + *parsed + hdrSize;

#ifdef _DEBUG
		const char *tagName;
		res = PKT_GetTagName( tag, &tagName );
#endif

		if ( dataSize == -1 ) // undefined size
		{
			dataSize = totalPktSize - hdrSize;
		}
		
		//
		//	v. 1.00.2 27.01.10
		//	Not all package allows processing packe partially f.e. pkt_PublicKeyEncryptedSessionKey can be parsed as entire.
		//	So we collect the partial packets to one common here and parse entire the packet below.
		//
		ptbData.WriteBytes( data, dataSize );

		*parsed += dataSize + hdrSize;
	}
	while ( partial );

	//
	// See v. 1.00.2 27.01.10 notes above.
	//
	res = parsePacketBody( tag, ptbData.buffer, ptbData.bufferSize, partial, ptb, sesKey, secretKey );

	//if ( res != ppe_ok ) return res;

	return res;
}


static int decompressData( cPTB &ptb, BYTE *pkt, int pktSize )
{
	BYTE algID;
	BYTE *body;
	int bodySize;
	int res = pgp_CompressedPacket_Parse( &algID, &body, &bodySize, pkt, pktSize );

	if ( res != ppe_ok ) return res;

	int size = 0;

	cPTB ptbHdr;

	//
	// We'll try increment partial packet size
	//
	int partialPtrSize = 0x1000;	

	for( int count = 1;; count *= 2 )
	{
		if ( ptbHdr.Allocate( count * partialPtrSize ) == NULL )
			return RSA_MALLOC_ERROR;

		size = ptbHdr.bufferSize;

		res = ca_uncompress( algID, ptbHdr.buffer, &size, body, bodySize );

		if ( res == cae_bufferTooSmall )
		{
			if ( count > 2000 )
			{
				return ppe_messageTooLong;
				//if ( IDYES != MessageBoxT(
				//		0, 
				//		_FormatT( L"Decompressed package exceeds %s\nDo you want to continue decompressing?", cWCStr().GetSizeText( count * partialPtrSize ) ),
				//		L"Decompress data",
				//		MB_ICONQUESTION | MB_YESNO | MB_SYSTEMMODAL ) )
				//			return res;
			}

			continue;
		}

		if ( res != cae_ok )
			res = pgp_getValidSize( ptbHdr.buffer, &size );

		if ( res == cae_ok )
		{
			ptb.WriteBytes( ptbHdr.buffer, size );
			return cae_ok;
		}

		break;
	}

	return res;
}


//
//	Stolen from int cDecryptDlg::ParsePackets( BYTE *ptr, int size, cPTB *ptb )
//
static int parsePackets( BYTE *ptr, int size, cPTB &ptb, cSessionKey &sesKey, cSecretKey &secretKey )
{
	int res = 0;

	int cnt = 0;
	int parsed;

	cPTB ptbDecrypted;
	cPTB ptbDecompressed;

	//
	//	Parses all packets
	//	1. First packts are public key encrypted session keys.
	//	If same packet was encountered it opens secret key by secretKey.openSecretKey()
	//	and try to implement the key to the packet to get session key.
	//	2. If session key was decoded and SymEncryptedAndIntegrityProtectedDataPacket was 
	//	encountered the data is decrypted to ptbDecrypted.
	//

	int totalParsed = 0;
	while ( size > totalParsed )
	{
		res = parsePacket( ptr + totalParsed, size - totalParsed, &parsed, ptbDecrypted, sesKey, secretKey );

		if ( res == ppe_invalidHeader )
		{
			res = ppe_ok;
			break;
		}

		if ( res == ppe_invalidSignature )
		{
			//
			//	v.1.03 checks base64 integrity, so we may definitely say here thar pasword is incorrect.
			//	The 'incorrect password' message appiares if symmetric key encryption algorithm was not recognize at pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket
			//
			//	If ParsePacket allows to ask password so we may continue here.
			//		totalParsed = 0;
			//		continue;
			//
			return ppe_invalidPassword;
		}

		if ( res != 0 )
			return res;

		cnt++;

		totalParsed += parsed;
	}

	if ( res != pkae_ok )
		return res;

	if ( ptbDecrypted.bufferSize == 0 )
		return ppe_invalidPacketSize;

	BYTE *pkt;
	int pktSize;

	res = pgp_SymEncryptedAndIntegrityProtectedDataPacket_Parse( (const BYTE**)&pkt, &pktSize, sa_getBlockSize( sesKey.hsa ), ptbDecrypted.buffer, ptbDecrypted.bufferSize );
	if ( res != pkae_ok ) return res;

	//
	//	The decrypted packet is Literal or Compressed data
	//	Let's check this.
	//

	BYTE tag;
	int dataSize;
	int hdrSize;
	bool partial = false;

	res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt, pktSize );
	if ( res != ppe_ok ) return res;

	if ( tag == pkt_CompressedData )
	{
		//
		//	The packet may be partial, so we have to accumulate all parts to common packet first.
		//
		//// Decompress
		//res = decompressData( &ptbDecompressed, pkt + hdrSize, dataSize );
		//if ( res != cae_ok ) return res;

		cPTB ptbCompressed;

		res = parsePacket( pkt, pktSize, &parsed, ptbCompressed, sesKey, secretKey );
		if ( res != ppe_ok ) return res;

		// Decompress
		res = decompressData( ptbDecompressed, ptbCompressed.buffer, ptbCompressed.bufferSize );
		if ( res != cae_ok ) return res;

		//
		//	Prepares to read Literal Data.
		//
		pkt = ptbDecompressed.buffer;
		pktSize = ptbDecompressed.bufferSize;

		//
		//	Read header of decompressed again.
		//
		partial = false;
		res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, pkt,  pktSize);
		if ( res != ppe_ok ) return res;
	}

	if ( tag != pkt_LiteralData )
		return ppe_invalidBlockType;
	
	return parsePacket( pkt, pktSize, &parsed, ptb, sesKey, secretKey );
}

//
//	Stolen from bool cDecryptDlg::decrypt()
//
static int decryptData( cPTB &plainText, const BYTE *buffer, int bufferSize, const char *utf8pwd, cSecretKey &secretKey )
{
	cSessionKey sesKey( utf8pwd );

	cPTB ptbOut;
	
	int res = parsePackets( (BYTE*)buffer, bufferSize, ptbOut, sesKey, secretKey );

	if ( res != ppe_ok )
		return res;

	if ( ptbOut.bufferSize == 0 )
		return ppe_inputDataIsCorrupted;

	const BYTE *body;
	int bodySize;

	int codePage = 0;

	res = pgp_LiteralDataPacket_Parse( &body, &bodySize, &codePage, NULL, NULL, NULL, ptbOut.buffer, ptbOut.bufferSize );

	if ( res != ppe_ok )
		return res;

	plainText.WriteBytes( body, bodySize );

	sesKey.closeKey();

	return ppe_ok;
}

int decodeText(
		cPTB &ptbPlainText,
		const char *base64encodedBuffer, int base64encodedBufferSize,
		const char *utf8pwd,
		cSecretKey &secretKey
	)
{
	const BYTE *pos;
	int size;

	int res = findEncodedBlock( &pos, &size, (const BYTE *)base64encodedBuffer, base64encodedBufferSize, sz_B64_Begin, sz_B64_End  );
	if ( res != pkae_ok )
		return res;

	cPTB ptbDecoded;
	if ( !readArmoredBlock( &ptbDecoded, base64encodedBuffer, base64encodedBufferSize, sz_B64_Begin, sz_B64_End ) )
		return ppe_inputDataIsCorrupted;

	res = decryptData( ptbPlainText, ptbDecoded.buffer, ptbDecoded.bufferSize, utf8pwd, secretKey );

	return res;
}

int encodeTextTest()
{
	cPTB b64;
	cPTB seed;
	seed.WriteBytes( NULL, 1024 );

	const char *in = "qwerty";
	const char *pwd = "qqqqqqqq";
	int len = strlen(in);

	int res = encodeTextByPassword( b64, in, len, pgpsaTripleDES, pwd, seed.buffer, seed.bufferSize, "Test" );

	if ( res != ppe_ok )
		return res;

	cSecretKey secretKey;

	cPTB out;

	res = decodeText( out, (char *)b64.buffer, b64.bufferSize, pwd, secretKey );

	if ( res != ppe_ok )
		return res;

	if ( len != out.bufferSize )
		return -1;

	return memcmp( out.buffer, in, len );
}

