#include "CDll.h"
#include "ODS.h"

#ifdef _ODS
#include "UnicodeFuncs.h"
#endif

#ifndef maxLen
#define maxLen( buff ) ( sizeof( buff ) / sizeof( *buff ) )
#endif

static int _MessageBoxA( HWND hWnd, LPCSTR lpText, LPCSTR lpCaption, UINT uType )
{
	static CDll dll( false, L"user32.dll", "MessageBoxA", false );

	if ( dll.error == CDll::errNo )
		return -1;

	typedef int WINAPI typeMessageBoxA( HWND hWnd, LPCSTR lpText, LPCSTR lpCaption, UINT uType );

	return ((typeMessageBoxA*)dll.pProc)(hWnd, lpText, lpCaption, uType);
}


//static 
int CDll::ErrorOut( const char *err, const char *name, const char *caption , int id )
{
	char msg[ 1024 ];
	size_t size = sizeof( msg );
	strcpy_s( msg, size, err );
	strcat_s( msg, size, " '" );
	strcat_s( msg, size, name );
	strcat_s( msg, size, "'" );
	return _MessageBoxA( 0, msg, caption, id );
}

static 
HMODULE
_LoadLibraryExT(
	LPCWSTR lpLibFileName,
	HANDLE hFile,
	DWORD dwFlags,
	int *pError
	)
{
	HMODULE hInst = NULL;
	int lastErr = 0;

	bool isNT = ( GetVersion() < 0x80000000 );

	if ( isNT )
	{
		hInst = LoadLibraryExW( lpLibFileName, hFile, dwFlags );
		lastErr = GetLastError();
	}
	else
	{

		char libFileNameA[MAX_PATH];
		size_t num;
		wcstombs_s( &num, libFileNameA, sizeof( libFileNameA ), lpLibFileName, _TRUNCATE );

		hInst = LoadLibraryExA( libFileNameA, hFile, dwFlags );

		if ( hInst == NULL )
		{
			lastErr = GetLastError();

			if ( ( hFile == NULL ) && ( dwFlags == 0 ) )
			{
				hInst = LoadLibraryA( libFileNameA );

				if ( hInst == NULL )
					lastErr = GetLastError();
			}

		}
	}

	if ( pError != NULL )
		*pError = lastErr;

	return hInst;
}


bool CDll::Load( BOOL usePath , const wchar_t *libraryName, const char *procName , BOOL showMessage )
{
	Unload();

	int err;

	m_showMessage = ( showMessage == TRUE );

	char libraryNameA[MAX_PATH];
	size_t num;
	wcstombs_s( &num, libraryNameA, sizeof( libraryNameA ), libraryName, _TRUNCATE );

	hInst = _LoadLibraryExT( libraryName , NULL , usePath ? LOAD_WITH_ALTERED_SEARCH_PATH : 0, &err );

	error = (hInst == NULL ? errInit : errNo );

	if ( error != errNo )
	{
		//ODS_fls( cWCSF().FormatT( L"%s '%s', err - %d, usePath %d, show %d", L"Can not load module", libraryName, err, usePath, showMessage ) );

		if ( showMessage )
		{
			char msg[ 1024 ];
			sprintf_s( msg, "%s '%s'\nerr - %d", "Can not load module", libraryNameA, err );
			_MessageBoxA( 0, msg, "Error", MB_ICONSTOP );
		}

		return false;
	}

	//
	//	2.08.2012
	//	wcscpy_s debug version uses invalid offset for m_path
	//	It rather not correcly estimate the shift with default pack() of CDll class.
	//	it does not fill first byte but fills next byte after m_path by 0xfe
	//
	//wcscpy_s( m_path, libraryName );
	int len = (int)wcslen( libraryName );
	if ( len > MAX_PATH )
		len = MAX_PATH - 1;

	memcpy( m_path, libraryName, len * sizeof( wchar_t ) );

	if ( procName == NULL )
		return true;

	return GetProcAddress( procName, showMessage ) == TRUE;
}

bool CDll::Load( BOOL usePath , const char *lpLibFileName, const char *procName, BOOL showMessage )
{
	wchar_t libFileNameW[MAX_PATH];

	m_showMessage = ( showMessage == TRUE );

	size_t num;

	mbstowcs_s( &num, libFileNameW, sizeof( libFileNameW )/ sizeof( *libFileNameW ), lpLibFileName, _TRUNCATE );

	bool res = Load( usePath, libFileNameW, procName, showMessage );
	
	wcscpy_s( m_path, libFileNameW );

	return res;
}

void *CDll::GetProcAddressEx( const char *procName )
{
	if ( hInst == NULL )
	{
		ODS_fls( L"Assertion! Dll is not initilized" );
		return NULL;
	}

	void *address = (void *)::GetProcAddress( hInst, procName );

	if ( address == NULL )
		ODSF( L"Function '%s' is not found", procName );

	return address;
}

BOOL CDll::GetProcAddress( const char *procName, BOOL showMessage, void **ppProc)
{
	//ODSFunc;

	if ( ppProc != NULL )
		*ppProc = NULL;

	if ( hInst == NULL )
		return FALSE;

	pProc = (void *)::GetProcAddress( hInst, procName );

	error = ( pProc == NULL ? errGetProcAddress : errNo );

	if ( error  )
	{
		//ODS_fls( _FormatT( L"Assertion! Error %d in get procedure %s address", GetLastError(), _getWcs( procName ) ) );

		if ( showMessage )
			ErrorOut( "Can not get procedure address -", procName  );
	}

	if ( ppProc != NULL )
		*ppProc = pProc;

	return error == errNo;
}

bool CDll::UpdateModuleTranslation()
{
	typedef bool typeUpdateModuleTranslation();

	typeUpdateModuleTranslation *proc = (typeUpdateModuleTranslation *)::GetProcAddress( hInst, "UpdateModuleTranslation" );

	if ( proc == NULL )
		return false;

	return proc();
}

bool CDll::LoadOnce( BOOL usePath , const char *libraryName, BOOL showMessage )
{
	if ( hInst != NULL )
		return true;

	return Load( usePath , libraryName, NULL, showMessage );
}

bool CDll::LoadProcOnce( BOOL usePath , const wchar_t *libraryName, const char *procName, BOOL showMessage )
{
	if ( hInst == NULL )
		return Load( usePath, libraryName, procName, showMessage );
	else
		return GetProcAddress( procName, showMessage ) != FALSE;
}

bool CDll::LoadProcOnce( BOOL usePath , const char *libraryName, const char *procName, BOOL showMessage )
{
	if ( hInst == NULL )
		return Load( usePath, libraryName, procName, showMessage );
	else
		return GetProcAddress( procName, showMessage ) != FALSE;
}

CDll::CDll( BOOL usePath , const wchar_t *libraryName, const char *procName, BOOL showMessage )
{
	//hInst = NULL;
	memset( this, 0, sizeof( *this ) );

	Load( usePath , libraryName, procName, showMessage );
}

CDll::CDll()
{
	memset( this, 0, sizeof( *this ) );
	error = true;
	//hInst = NULL;
	m_showMessage = false;
}

CDll::CDll( BOOL usePath , const char *libraryName, const char *procName, BOOL showMessage )
{
	//hInst = NULL;
	memset( this, 0, sizeof( *this ) );
	Load( usePath , libraryName, procName, showMessage );
}

void CDll::Unload()
{
	if (hInst)
	{
		error = true;
		FreeLibrary(hInst);
		hInst = NULL;
	}
}

CDll::~CDll()
{
	Unload();
}
