#ifndef __FileView__
#define __FileView__

#include "windefs.h"
#include "UnicodeFuncs.h"
#include "cStringBuffer.h"
#include "FileViewA.h"
#include "ODS.h"
#include "cMBStr.h"
#include "cStringBlock.h"

extern const BYTE unicodeLEPrefix[ 2 ]; 
extern const BYTE unicodeBEPrefix[ 2 ]; 
extern const BYTE utf8prefix[ 3 ];

//
//	Helper class for viewing mapped files
//
//	Sample code:
//		cFileView cfv( fileName );
//
//		if ( cfv.m_err )
//			return cfv.m_err;
//
//		UseTheMemory( cfv.m_ptr, cfv.m_size );
//

#ifdef _UseFileViewA_
class cFileView : public cFileViewA
{
public:
	cFileView( const char *fileName, bool readOnly = true ) :cFileViewA( fileName, readOnly ) {}
	cFileView()	: cFileViewA() {};
	BOOL Open( const char *fileName, bool readOnly = true )
	{
		return cFileViewA::Open( fileName, readOnly );
	}
#else
class cFileView
{
public:
#ifdef _WIN32
	HANDLE m_hFile;
	HANDLE m_hMap;
#endif // #ifdef _WIN32

	enum { cfv_ok = 0, cfv_fileDoesNotExist, cfv_errorInOpeningFile, cfv_errorInMapingFile };
	int m_err;
	DWORD m_sysErr;
	ULARGE_INTEGER m_size;
	BYTE *m_ptr;

	BOOL OpenMapping( bool readOnly = true );
	void CloseMapping();
	void Close();

#endif
	BOOL Open( const wchar_t *fileName, bool readOnly = true );
	BOOL Open( const char *fileName, bool readOnly = true )
	{
		return Open( getWCStr( fileName ), readOnly );
	}

	void commomConstructor()
	{
#ifdef _WIN32
		m_hFile = NULL;
		m_hMap = NULL;
#endif // #ifdef _WIN32

		m_err = cfv_fileDoesNotExist;
		m_sysErr = 0;
		m_size.QuadPart = 0;
		m_ptr = NULL;

		m_curPosition.QuadPart = 0;
		m_nextPosition = 0;
		m_stepSize = 0;
		m_windowSize = 0;
		m_mapAccess = 0;
		m_bSkipInitAtOpening = false;
	}

	cFileView( const char *fileName, bool readOnly = true )
	{
		//memset( this, 0, sizeof( *this ) );
		commomConstructor();

		Open( fileName, readOnly );
	}

	cFileView()
	{
		//memset( this, 0, sizeof( *this ) );
		commomConstructor();
	}

	cFileView( const wchar_t *fileName, bool readOnly = true, SIZE_T stepSize = 0 )
	{
		//memset( this, 0, sizeof( *this ) );
		commomConstructor();
		m_stepSize = stepSize;
		Open( fileName, readOnly );
	}
	
	~cFileView()
	{
		Close();
	}

	ULARGE_INTEGER m_curPosition;
	ULONGLONG m_nextPosition;
	SIZE_T m_stepSize;
	SIZE_T m_windowSize;
	DWORD m_mapAccess;
	bool m_bSkipInitAtOpening;

	//
	//	MappingNextWindow
	//		1. Closes old mapping and increment current position to one step forward.
	//		2. Maps window in current position.
	//
	//	The function is used for first mapping at cFileView::OpenMapping and 
	//	for mapping next window at cStringFile::UpdateMapping() ( see comment there )
	//
	//	Note, for real next window mapping the m_stepSize has to assigned as non zero before 
	//	cFileView::Open function call.
	//
	bool MappingNextWindow();

	virtual bool Init();
};

//
//	WriteToFile
//		All functions return number of byte are written to file,
//		return zero if error.
//
int WriteToFile( const char *fileName, const BYTE *ptr, DWORD size, bool append = false, DWORD *pError = NULL, DWORD shareMode = FILE_SHARE_READ );
int WriteToFile( const wchar_t *fileName, const BYTE *ptr, DWORD size, bool append = false, DWORD *pError = NULL, DWORD shareMode = FILE_SHARE_READ );
int WriteToFile( const wchar_t *fileName, const BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError = NULL );
int WriteToFile( HANDLE hFile, const BYTE *ptr, DWORD size, bool append = false, DWORD *pError = NULL );
int WriteToFile( HANDLE hFile, const BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError = NULL );
int WriteToFile( const wchar_t *fileName, const wchar_t *str, int len = -1, bool append = false, DWORD *pError = NULL );

int ReadFromFile( HANDLE hFile, BYTE *ptr, DWORD size, DWORD *pError = NULL );
int ReadFromFile( HANDLE hFile, BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError = NULL );
int ReadFromFile( const wchar_t *fileName, BYTE *ptr, DWORD size, ULONGLONG pos, DWORD *pError = NULL );


//
//	Copies opened file ( hFileSrc ) to new file ( fileNameDst )
//		1. Saves current hFile position
//		2. Read portions from source and write to destination
//		3. Returns current position of source file back
//	Returns number of bytes copied.
//
DWORDLONG CopyToFile( HANDLE hFileSrc, const wchar_t *fileNameDst );

//
//	Unicode string file functions.
//		Init - writes prefix at the file start
//		Add  - adds string with additional \r\n at current position.
//
//	 codePage is  CP_ACP, CP_UTF8, -1 = means unicode little endian
//
bool InitUnicodeFile( const wchar_t *fileName, HANDLE *pHFile = NULL, int codePage = -1  );
bool InitUnicodeFile( HANDLE hFile, int codePage = -1 );
bool AddStrToFile( HANDLE hFile, const wchar_t *str, int len = -1 );

class cTempFile : public cStringBlock
{
public:
	cWCStr m_path;
	HANDLE m_hFile;
	cWCS m_prefix;
	cWCS m_postfix;
	int m_codePage;
	bool m_deleteOnExit;
	DWORD m_dwErr;

	virtual bool createFile( const wchar_t *prefix = NULL )
	{
		closeFile();

		m_error = true;

		if ( prefix == NULL )
			prefix = m_prefix.m_szBody;

		if ( m_path.isEmpty() )
		{
			
			if ( m_path.SetLength( MAX_PATH ) == 0 )							{ ODS_fls( L"Assertion!" ); return false; }
			if ( GetTempPathT( m_path.m_len, m_path.m_szBody, false ) == 0 )	{ ODS_fls( L"Assertion!" ); return false; }
			if ( CreateTempFileName( m_path.m_szBody, prefix, 0, m_path.m_szBody, &m_hFile, FILE_SHARE_READ, FILE_ATTRIBUTE_NORMAL, &m_dwErr, m_postfix.m_szBody ) == 0 )
			{ ODS_fls( _FormatT( L"Assertion! Error '%s' in creating temp file name, path '%s'", _GetSystemError( m_dwErr ), m_path.m_szBody ) ); return false; }
		}

		if ( !InitUnicodeFile( m_hFile, m_codePage ) )
			return false;
		
		m_error = false;

		return true;
	}

	cTempFile( bool create = true, const wchar_t *prefix = L"BC", const wchar_t *postfix = L".tmp", int codePage = -1, bool deleteOnExit = false  )
	{
		m_dwErr = 0;
		m_hFile = NULL;
		m_error = true;
		m_prefix.Set( prefix );
		m_postfix.Set( postfix );
		m_codePage = codePage;
		m_deleteOnExit = deleteOnExit;

		if ( create )
			createFile( m_prefix.m_szBody );
	}

	bool AddStringToBlock( const wchar_t *str, int len = -1 )
	{
		if ( m_hFile == NULL )
			createFile();

		return ::AddStrToFile( m_hFile, str, len );
	}

	void closeFile()
	{
		if ( ( m_hFile != NULL ) && ( m_hFile != INVALID_HANDLE_VALUE ) )
			CloseHandle( m_hFile );
		
		m_hFile = NULL;
	}
	
	virtual void deleteFile()
	{
		//ODSFuncS( m_path );

		closeFile();

		if ( !FileExistsT( m_path.m_szBody ) )
		{
			//ODS_fls( _FormatT( L"Assertion! File '%s'", m_path.m_szBody ) );
		}
		else
			DeleteFileT( m_path.m_szBody );
	}

	~cTempFile()
	{
		closeFile();
		if ( m_deleteOnExit )
			deleteFile();
	}
};

class cStringFile : public cFileView, public cStringBuffer
{
public:

	bool commonConstructor( const wchar_t *fileName );

	//
	//	cStringFile
	//		if blockSize is assigned so file is not mapped to all size
	//		it uses stepped mapping window. See comments at cFileView::MappingNextWindow.
	//
	cStringFile( const wchar_t *fileName, SIZE_T blockSize = 0 )
	{
		m_stepSize = blockSize;
		commonConstructor( fileName );
	}

	cStringFile( const char *fileName = NULL )	 { commonConstructor( getWCStr( fileName ) ); }

	bool Open( const wchar_t *fileName, bool readOnly = true );
	bool Open( const char *fileName, bool readOnly = true ) {	return Open( getWCStr( fileName ), readOnly ); }

	virtual bool Init();

	bool getNextFile( cMBStr *nextFileName );
	bool getNextFileW( cWCStr *nextFileName );
	bool getNextFileT( cWCStr *nextFileName );
	//bool getNextFile( char *nextFileName );
	//bool getNextFileW( wchar_t *nextFileName );
	//bool getNextFileT( wchar_t *nextFileName );
	virtual bool FindString( wchar_t *buff, int buffLen, const wchar_t *str, bool getTailOnly = true );

	//
	//	UpdateMapping
	//		Moves mapped window forward on size of one step.
	//		The window size is twice more then step size.
	//		Moving happens if current string is placed at secong half of current window. 
	//				   <here>
	//	|	step	|	step+1	|	step+2	|
	//	[	  current window	]
	//				[		next window		]
	//	
	//	Note, the m_stepSize value has to be assigned before cFileView::Open function is called.
	//	See alse cFileView::MappingNextWindow
	//
	bool UpdateMapping();
};


//
//	Error codes for WriteFileVerify function.
//
enum {
	ewfv_errorNo = 0,
	ewfv_errorSize,
	ewfv_errorInAllocatingMemory,
	ewfv_errorInWrittingFile,
	ewfv_errorInReadingFile,
	ewfv_errorInVerifing,
	ewfv_errorNetwork = ERROR_CONNECTION_ABORTED,
	ewfv_errorDiskFull = ERROR_DISK_FULL,
};

const char *ewfv_str( int err );

//
//	WriteFileVerify
//		Writes data first if verification is on, it reads data from file for verification.
//		If the data is not compatible it tries to write and verify again according with try count.
//
int WriteFileVerify( int tryCount, HANDLE hFile, BYTE *bfWrite, DWORD size, DWORD *pWritten = NULL, bool verify = false, int *pLastError = NULL );

//
//	MoveCurrentFilePointer
//	So it gets current position add(substract) required shift size 
//		and moves result pointer.
//
//	Usual variants do not work to move file pointer back:
//		v.1
//			SetFilePointer( hFile, -size, &(uli.HighPart), FILE_CURRENT );
//			It does not work becouse it uses common signed value.
//		v.2
//			LARGE_INTEGER uli = { 0, 0 };
//			uli.QuardPart -= size;
//			SetFilePointer( hFile, uli.LowPart, &(uli.HighPart), FILE_CURRENT );
//			1. it reports error move before file beginning.
//			2. it do never work for Windows 95 ( as MSDN report uli.HighPart has to be zero ).
//
bool MoveCurrentFilePointer( HANDLE hFile, LONGLONG *ll, int *pLastError = NULL );

bool DeletePrivateProfileString(
		LPCSTR lpAppName,
		LPCSTR lpKeyName,
		LPCSTR lpString,
		LPCSTR lpFileName
	);

#endif // __FileView__
