#include "GetTempPathIH.h"
#include "token_info.h"
#include "UnicodeFuncs.h"
#include "cRegKey.h"
#include "AdvAPI.h"

/////////////////
//
//	GetTempPathIH
//		Gets user temp path by impersonated token.
//

static bool ExpandEnvironmentStringsForUserT(
    IN HANDLE hToken,
    IN LPCWSTR lpSrc,
    OUT LPWSTR lpDest,
    IN DWORD dwSize)
{
	typedef BOOL WINAPI typeExpandEnvironmentStringsForUserW( IN HANDLE hToken, IN LPCWSTR lpSrc, OUT LPWSTR lpDest, IN DWORD dwSize );

	static CDll dll( false, L"userenv.dll", "ExpandEnvironmentStringsForUserW", false );

	if ( dll.error )
		return false;

	return ((typeExpandEnvironmentStringsForUserW*)dll.pProc)( hToken, lpSrc, lpDest, dwSize ) == TRUE;
}

static DWORD GetUserEnvironmentTempPath_ByImpersonatedToken(
    DWORD nBufferLength,
    LPWSTR lpBuffer
    )
{
	//
	//	09.11.2015 Bug #2282
	//	There is 'HKEY_CURRENT_USER\Environment' key with TEM and TMP REG_EXPAND_SZ values, data - %USERPROFILE%\AppData\Local\Temp
	//
	HANDLE hTokenI = GetProcessImpersonatedToken_ODS();

	if ( hTokenI == NULL )
		return 0;

	cWCS wcsSrc( MAX_PATH );
	cWCS wcsDst( MAX_PATH );

	DWORD dwSize = wcsSrc.m_len * sizeof( wchar_t );
	DWORD dwType;

	HKEY hkCU = HKEY_CURRENT_USER;

	//
	//	11.12.2015 Bug #2282
	//		Impersonated user reads registry for SYSTEM as current user.
	//		So we try to read from HKEY_USERS\S-...
	//
	wchar_t *szSid;
	bool ok = CAdvAPIDll::getStatic().GetSidByToken( &szSid, hTokenI );

	cRegKey chkCU( HKEY_USERS, szSid, 0, KEY_READ );
	if ( ok && ( chkCU.getKey() != NULL ) )
		hkCU = chkCU.getKey();

	LONG lRes = RegGetValueT( hkCU, L"Environment", L"TEMP", 0, &dwType, wcsSrc.m_szBody, &dwSize );

	ok = lRes == ERROR_SUCCESS;

	if ( ok )
		ok = ExpandEnvironmentStringsForUserT( hTokenI, wcsSrc.m_szBody, wcsDst.m_szBody, wcsDst.m_len );

	CloseHandle( hTokenI );

	if ( !ok )
		return 0;

	copyString( lpBuffer, wcsDst.m_szBody, nBufferLength );

	return (DWORD)wcslen( wcsDst.m_szBody );
}
//
//	GetUserLocalAppDataPath_ByImpersonatedToken
//
static DWORD GetUserLocalAppDataPath_ByImpersonatedToken(
    DWORD nBufferLength,
    LPWSTR lpBuffer
    )
{
	//
	//	28.06.2013 Both this and SHGetSpecialFolderPathT function return same result for impersonated thread.
	//	In Windows 7.
	//	F.e. 'C:\Users\PAV~1.BC_\AppData\Local'
	//
//#ifdef _DEBUG
//	cWCStr wcsDbg( MAX_PATH );
//	SHGetSpecialFolderPathT( NULL, wcsDbg.m_szBody, CSIDL_LOCAL_APPDATA, !doNotCreate );
//#endif

	HANDLE hTokenI = GetProcessImpersonatedToken_ODS();

	if ( hTokenI == NULL )
		return 0;

	cWCS wcsSrc( MAX_PATH );
	cWCS wcsDst( MAX_PATH );

	DWORD dwSize = wcsSrc.m_len * sizeof( wchar_t );
	DWORD dwType;

	//
	//	Impersonated user reads registry for SYSTEM as current user.
	//	So we try to read from HKEY_USERS\S-...
	//
	HKEY hkCU = HKEY_CURRENT_USER;
	wchar_t *szSid;
	bool ok = CAdvAPIDll::getStatic().GetSidByToken( &szSid, hTokenI );

	cRegKey chkCU( HKEY_USERS, szSid, 0, KEY_READ );
	if ( ok && ( chkCU.getKey() != NULL ) )
		hkCU = chkCU.getKey();

	const wchar_t *szUserShellFolders = L"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\User Shell Folders";

	dwSize = wcsSrc.m_len * sizeof( wchar_t );

	//
	// Windows XP has 'Local Settings' in szUserShellFolders key.
	//	%USERPROFILE%\Local Settings, See V.1 comments.
	//
	LONG lRes = 
		RegGetValueT( 
			hkCU, 
			szUserShellFolders, 
			L"Local Settings", 
			0, 
			&dwType, 
			wcsSrc.m_szBody, 
			&dwSize
		);

	//
	// Windows 7 has 'Local AppData' in szUserShellFolders key.
	//	%USERPROFILE%\AppData\Local
	//
	if ( lRes != ERROR_SUCCESS )
		lRes = 
			RegGetValueT( 
				hkCU, 
				szUserShellFolders, 
				L"Local AppData", 
				0, 
				&dwType, 
				wcsSrc.m_szBody, 
				&dwSize
			);


	ok = lRes == ERROR_SUCCESS;

	if ( ok )
		ok = ExpandEnvironmentStringsForUserT( hTokenI, wcsSrc.m_szBody, wcsDst.m_szBody, wcsDst.m_len );

	CloseHandle( hTokenI );

	copyString( lpBuffer, wcsDst.m_szBody, nBufferLength );

	return (DWORD)wcslen( wcsDst.m_szBody );
}

//
//	GetTempPathIH
//		The standard call GetTempPathT for non impersonated thread returns "C:\Users\PAV~1.BC_\AppData\Local\Temp\", but 
//		for impersonated thread returns 'C:\Windows\Temp'.
//		It was assumed that impersonated thread get temp path for the impersonated user.
//		So we use GetUserLocalAppDataPath_ByImpersonatedToken here. It reads impersonated user registry, for shell folders and expand the environment string for him.
//		As result impersonated thread gets same path as GetTempPathT returns for simple user.
//
DWORD GetTempPathIH(
    DWORD nBufferLength,
    LPWSTR lpBuffer,
	bool doNotCreate // = false
    )
{
	DWORD len = GetUserEnvironmentTempPath_ByImpersonatedToken( nBufferLength, lpBuffer );

	if ( len == 0 )
	{
		len = GetUserLocalAppDataPath_ByImpersonatedToken( nBufferLength, lpBuffer );

		if ( len == 0 )
			return GetTempPathT( nBufferLength, lpBuffer, doNotCreate );

		const wchar_t *szTemp = L"\\Temp";
		int tempLen = wcslen( szTemp );

		if ( len + tempLen + 1 > nBufferLength )
		{ ODS_fls( L"Assertion!" ); return len + tempLen + 1; }

		copyString( lpBuffer + len, szTemp, nBufferLength - len );
		len += tempLen;
	}

	if ( !FileExistsT( lpBuffer ) && !doNotCreate )
		CreateDirectoryW( lpBuffer, NULL );

	copyString( lpBuffer + len, L"\\", nBufferLength - len );
	len ++;

	return len;
}
