#include "IconCache.h"
#include "UnicodeFuncs.h"
#include "shell32dll.h"

const bool useRealExeIcons = true;

#pragma pack (1)

struct sExeType
{
	wchar_t type;
	wchar_t *name;
};

#pragma pack ()

#define exeMark L'*'

// returns a pointer to the extension of a file.
//
// in:
//      qualified or unqualfied file name
//
// returns:
//      pointer to the extension of this file.  if there is no extension
//      as in "foo" we return a pointer to the NULL at the end
//      of the file
//
//      foo.txt     ==> ".txt"
//      foo         ==> ""
//      foo.        ==> "."
//

wchar_t * PathFindExtensionT( const wchar_t *pszPath )
{
	const static wchar_t driveExt[2] = L":";

	if ( ( wcslen( pszPath ) == 2 ) && ( pszPath[1] == ':' ) )
		return (wchar_t *)driveExt;
	if ( ( wcslen( pszPath ) == 3 ) && ( pszPath[1] == ':' ) && ( pszPath[2] == '\\' ) )
		return (wchar_t *)driveExt;

    const wchar_t *pszDot;

    for (pszDot = NULL; *pszPath; pszPath++ )
    {
        switch (*pszPath) {
        case L'.':
            pszDot = pszPath;	// remember the last dot
            break;
        case L'\\':
        case L' ':				// extensions can't have spaces
            pszDot = NULL;		// forget last dot, it was in a directory
            break;
        }
    }

    // if we found the extension, return ptr to the dot, else
    // ptr to end of the string (NULL extension) (cast->non const)
    return pszDot ? (wchar_t *)pszDot : (wchar_t *)pszPath;
}

static bool isNetworkPath( const wchar_t *fileName )
{
	// \\serw10\D\Users\ser\... -> is network
	// \\?\Volume{0cd			-> is not

	if ( fileName[0] != L'\\' ) return false;
	if ( fileName[1] != L'\\' ) return false;
	if ( fileName[2] == L'?' ) return false;

	return true;
}

int cIconCache::GetIconNum( const wchar_t *fileName, wchar_t **pTypeName )
{
	ODSFuncS( fileName );

	DWORD attr = -1;

	// It takes a lot of time if networ drive is currently not accessable. 
	if ( !isNetworkPath( fileName ) )
		attr = GetFileAttributesT( fileName );

	bool bDirectory = false;
	if ( attr == -1 ) // INVALID_FILE_ATTRIBUTES )
	{
		//return -1; // Does not exist!
	}
	else
		bDirectory = ( attr & FILE_ATTRIBUTE_DIRECTORY ) != 0;

	if ( useRealExeIcons && bDirectory && ( m_undefNum == 1 ) )
		return 0; // folder icons was assigned

	//
	//	Check cache if the icon already exists...
	//

	wchar_t *pszExt;
	
	if ( ( wcsncmp( L"http://", fileName, 7 ) == 0 )
		 || ( wcsncmp( L"https://", fileName, 8) == 0 ) 
		 || ( wcsncmp( L"ftp://", fileName, 6) == 0 ) 
		 || ( wcsncmp( L"www.", fileName, 4) == 0 ) 
		)
		pszExt = L".htm";
	else
		pszExt = PathFindExtensionT( fileName );

	int num = GetCachedNum( pszExt, fileName, bDirectory );

	if ( num == m_undefNum )
	{
		//
		//	Get Icon from shell...
		//

		num = AddShellIcon( pszExt, fileName, bDirectory );
	}

	if ( pTypeName )
	{
		*pTypeName = typeNameTbl[ num - ( m_undefNum + 1) ];
	}

	return num;
}

int cIconCache::AddShellIcon( const wchar_t *ext, const wchar_t *filePathName, bool bDirectory )
{
	static const wchar_t szBCArchiveLoc[] = L"BCTMName::";

	cWCStr wcsName( szBCArchiveLoc );
	wcsName.Add( ext );

	bool realFileIcon = true;

	if ( ( _wcsicmp( ext, L".lnk" ) == 0 ) || bDirectory )
	{
		realFileIcon = true;
	}
	else
	if ( ( ext[0] == ':' ) && ( ext[1] == 0 ) )
	{
		int len = wcslen( filePathName );

		if ( wcsName.Set( filePathName ) == 0 )
			return m_undefNum;

		if ( wcsName.m_szBody[ len - 1 ] != '\\' )
			wcsName.Add( L"\\" );

		filePathName = wcsName.m_szBody;
	}
	else
	if ( !useRealExeIcons || (( wcscmp( ext, L".exe" ) != 0 ) && ( wcscmp( ext, L".ico" ) != 0 ) ) )
	{
		filePathName = wcsName.m_szBody;

		realFileIcon = false;
	}

	int is,il;

	UINT uFlags = SHGFI_ICON | SHGFI_USEFILEATTRIBUTES;

	SHFILEINFOW sfi;
	memset( &sfi, 0, sizeof( sfi ) );
	DWORD dwAttr = bDirectory ? FILE_ATTRIBUTE_DIRECTORY : 0;

	if (!SHGetFileInfoT( filePathName, dwAttr, &sfi, sizeof(sfi), uFlags | SHGFI_LARGEICON))
	{
		return m_undefNum;
	}

	il = ImageList_AddIconT( m_hImlLg, sfi.hIcon );

	if (SHGetFileInfoT( filePathName, dwAttr, &sfi, sizeof(sfi), uFlags | SHGFI_SMALLICON | SHGFI_TYPENAME  )) // | SHGFI_ICONLOCATION
	{
		is = ImageList_AddIconT( m_hImlSm, sfi.hIcon );

		int extTblNum = ReallocExtTbl();

		if ( extTblNum == -1 )
		{
			assert( extTblNum != -1 );
			return m_undefNum;
		}

		if ( m_firstNum == -1 )
		{
			if ( extTblNum != 0 ) // The m_firstNum has to be initialized once at first assigment -> extTblNum == 0
			{
				ODS_fls( L"Assertion!" );
			}

			m_firstNum = il;
		}

		if ( m_firstNum + extTblNum != il )
		{
			ODS_fls( L"Assertion! Icon image is moved" );
			m_firstNum = il - extTblNum; // Corect for future images. Early images will be moved.
		}

		if ( realFileIcon )
		{
			sExeType *s = ( sExeType *)extTbl[ extTblNum ];
			s->type = exeMark;
			s->name = newString( filePathName );
		}
		else
			copyString( extTbl[ extTblNum ], ext, maxLen( extTbl[ extTblNum ] ) );

		typeNameTbl[ extTblNum ] = newString( sfi.szTypeName );

		//assert( is == extTblNum + m_undefNum + 1 );
		assert( is == il );

		return is;
	}

	return m_undefNum;
}

int cIconCache::AddIconWithPath( const wchar_t *filePathName, HICON hSmallIcon, HICON hLargeIcon )
{
	if ( hLargeIcon == NULL )
		hLargeIcon = hSmallIcon;

	int il = ImageList_AddIconT( m_hImlLg, hLargeIcon ); // Large icon has to be added here.
	int is = ImageList_AddIconT( m_hImlSm, hSmallIcon );

	int extTblNum = ReallocExtTbl();

	if ( extTblNum == -1 )
	{
		assert( extTblNum != -1 );
		return m_undefNum;
	}

	sExeType *s = ( sExeType *)extTbl[ extTblNum ];
	s->type = exeMark;
	s->name = newString( filePathName );

	typeNameTbl[ extTblNum ] = newString( filePathName );

	assert( is == il );

	return is;
}

int cIconCache::GetCachedNum( const wchar_t *ext, const wchar_t *filePathName, bool bDirectory )
{
	int i;

	bool bUseRealIcon = bDirectory || (wcscmp( ext, L".exe" ) == 0) || (wcscmp( ext, L".ico" ) == 0);

	for ( i = extTblSize; i--; )
	{
		sExeType *s = ( sExeType *)extTbl[ i ];

		if ( useRealExeIcons 
			&& bUseRealIcon 
			&& ( s->type == exeMark ) 
			&& ( wcscmp( s->name, filePathName ) == 0 ) 
			)
			break;

		if ( _wcsicmp( ext, extTbl[ i ] ) == 0 )
		{
			if ( m_firstNum == -1 )
			{
				ODS_fls( L"Asertion!" );
				break;
			}

			return m_firstNum + i;
		}
	}
	
	return m_undefNum;
}

cIconCache::cIconCache( HINSTANCE hInst, HWND hListView, int idiFolder, int idiUnknown, HICON hDefIcon )
{
	HICON hFolder  = LoadIcon( hInst, MAKEINTRESOURCE( idiFolder ) );
	HICON hUnknown = LoadIcon( hInst, MAKEINTRESOURCE( idiUnknown ) );

	CommonConstructor( hInst, hListView, hFolder, hUnknown, hDefIcon );

	DestroyIcon( hFolder );
	DestroyIcon( hUnknown );
}

cIconCache::cIconCache( HINSTANCE hInst, HWND hListView, HICON hFolder, HICON hUnknown, HICON hDefIcon )
{
	CommonConstructor( hInst, hListView, hFolder, hUnknown, hDefIcon );
}

void cIconCache::Init( HINSTANCE hInst, HWND hListView, int idiFolder, int idiUnknown )
{
	memset( this, 0, sizeof( cIconCache ) );

	m_hInst = hInst;

	//
	//      Image lists creation...
	//
	m_hImlLg = ImageList_CreateT( GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), ILC_MASK | ILC_COLOR32 , 8, 8 );
	m_hImlSm = ImageList_CreateT( GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), ILC_MASK | ILC_COLOR32 , 8, 8 );


	//
	//      We have to add folder and unknown icons...
	//

	if ( cShell32Dll::getStatic().error )
	{
			AddIcon( cInstanceHolder::LoadIconT( m_hInst, idiFolder ) ); // IDI_Folder
			AddIcon( cInstanceHolder::LoadIconT( m_hInst, idiUnknown ) ); // IDI_Unknown
	}
	else
	{
			enum { idShell32_Unknown = 1, idShell32_ClosedFolder = 4, idShell32_OpenedFolder = 5 };
			//AddIcon( cInstanceHolder::LoadIconT( cShell32Dll::getStatic().hInst, idShell32_ClosedFolder ) );
			m_undefNum = AddIcon( cInstanceHolder::LoadIconT( cShell32Dll::getStatic().hInst, idShell32_Unknown ) );
	}

	extTbl = NULL;
	extTblSize = 0;

	SetImageList( hListView );
}

void cIconCache::CommonConstructor( HINSTANCE hInst, HWND hListView, HICON hFolder, HICON hUnknown, HICON hDefIcon )
{
	memset( this, 0, sizeof( cIconCache ) );

	m_hInst = hInst;

	//
	//	Image lists creation...
	//
//	m_hImlSm = ImageList_CreateT( GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), ILC_MASK | ILC_COLOR16 , 4, 10 );
	m_hImlLg = ImageList_CreateT( GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), ILC_MASK | ILC_COLOR32 , 8, 8 );
	m_hImlSm = ImageList_CreateT( GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), ILC_MASK | ILC_COLOR32 , 8, 8);


	//
	//	We have to add folder and unknown icons...
	//

	if ( hUnknown != NULL )
	{
		AddIcon( hFolder );
		m_undefNum = AddIcon( hUnknown );
	}
	else
		m_undefNum = -1;

	m_firstNum = -1;

	if ( hDefIcon != NULL )
	{
		m_defNum = AddIcon( hDefIcon );
	}
	else
		m_defNum = 0;

	extTbl = NULL;
	extTblSize = 0;

	ListView_SetImageList( hListView, m_hImlLg, LVSIL_NORMAL );
	ListView_SetImageList( hListView, m_hImlSm, LVSIL_SMALL );
}

int cIconCache::AddIcon( HICON hIcon )
{
	int is,il;

	il = ImageList_AddIconT( m_hImlLg, hIcon );
	is = ImageList_AddIconT( m_hImlSm, hIcon );

	assert( is == il );

	return is;
}

void cIconCache::DestroyImageLists()
{
	if ( m_hImlLg != NULL )
		ImageList_DestroyT( m_hImlLg );

	if ( m_hImlSm != NULL )
		ImageList_DestroyT( m_hImlSm );

	m_hImlLg = NULL;
	m_hImlSm = NULL;
}

cIconCache::~cIconCache()
{
	DestroyImageLists();

	if ( extTbl )
	{
		for ( int i = extTblSize; i--; )
		{
			sExeType *s =  (sExeType *)extTbl[ i ];

			if ( s->type == exeMark )
				deleteString( &(s->name ) );
		}

		free( extTbl );
	}

	if ( typeNameTbl )
	{
		while( extTblSize-- )
		{
			if ( typeNameTbl[ extTblSize ] )
				delete typeNameTbl[ extTblSize ];
		}

		free( typeNameTbl );
	}
}

//
//	Adds new one item to cache table.
//	Returns number of added item.
//
int cIconCache::ReallocExtTbl( )
{
	void *ptr = realloc( extTbl, (extTblSize + 1) * sizeof( *extTbl ) );

	if ( ptr == NULL )
		return -1;
	
	void **pTbl = (void **)&extTbl;

	*pTbl = ptr;

	ptr = realloc( typeNameTbl, (extTblSize + 1) * sizeof( wchar_t * ) );

	if ( ptr == NULL )
		return -1;

	typeNameTbl = (wchar_t**)ptr;

	return extTblSize++;
}

void cIconCache::SetImageList( HWND hListView )
{
	ListView_SetImageList( hListView, m_hImlLg, LVSIL_NORMAL );
	ListView_SetImageList( hListView, m_hImlSm, LVSIL_SMALL );
}

void cIconCache::DestroySharedImageLists()
{
	//
	// 12.10.2011
	// Prev version did not set image lists to zero so both the lists were deleted again at cIconCache destructor.
	//
	if ( m_hImlLg ) { ImageList_DestroyT( m_hImlLg ); m_hImlLg = NULL; }
	if ( m_hImlSm ) { ImageList_DestroyT( m_hImlSm ); m_hImlSm = NULL; }
}