/////////////////////////////////////////////////////////////////
//
//	Copyright  2009-2018 Jetico Inc., Finland
//	All rights reserved.
//
//	File: SignatureCheck.cpp
//
//	Revision:
//		22.03.2012 By Alexander Pichuev.
//		Added to Includes
//		28.09.2018
//		Critical section is used at getStatic function.
//
//	Created:
//		25.11.2009 By Nail Kaipov
//
//	Description: 
//		Verifies signature and checks if signer is Jetico.
//
/////////////////////////////////////////////////////////////////

#include <windows.h>
#include <Softpub.h>
#include <wincrypt.h>
#include <wintrust.h>
#include <tchar.h>
#include <time.h>

#include "SignatureCheck.h"
#include "cSysCriticalSection.h"

#include "ODS.h"
#ifdef _ODS
#include "UnicodeFuncs.h"
#include "cStringBuffer.h"
#endif

#ifdef _useAdvApi
#include "AdvApi_MakeHash.h"
#else
#include "..\CertificateLib\hashAlgs\SHA1\Sha1.h"
#endif

static bool _makeHashSHA1( const BYTE* bf, DWORD bfLen, BYTE *hash, DWORD *pHashLen )
{
	// Several systems have no reeady key set, so we have to create it
	//	Seems simple to use direct code from certificate lib.
#ifdef _useAdvApi
	return cAdvApiForHash::MakeHash(
					bf,
					bfLen,
					hash,
					pHashLen
				);
#else

	if ( *pHashLen < 20 )
	{
		*pHashLen = 20;
		return false;
	}

	*pHashLen = 20;
	return SHA1MakeDigest( bf, bfLen, hash ) == SHA1_ERROR_NO;

#endif
}


typedef LONG					(WINAPI *CWinVerifyTrust)( HWND hWnd,  GUID* pgActionID,  WINTRUST_DATA* pWinTrustData);
typedef CRYPT_PROVIDER_DATA* 	(WINAPI *CWTHelperProvDataFromStateData)(HANDLE hStateData);
typedef CRYPT_PROVIDER_SGNR* 	(WINAPI *CWTHelperGetProvSignerFromChain)(CRYPT_PROVIDER_DATA* pProvData,  DWORD idxSigner,  BOOL fCounterSigner,  DWORD idxCounterSigner);
typedef CRYPT_PROVIDER_CERT*	(WINAPI *CWTHelperGetProvCertFromChain)(CRYPT_PROVIDER_SGNR* pSgnr,  DWORD idxCert);
typedef DWORD					(WINAPI *CCertNameToStr)(DWORD dwCertEncodingType, PCERT_NAME_BLOB pName,  DWORD dwStrType,  LPTSTR psz,  DWORD csz);
typedef BOOL					(WINAPI *CCATAdminCalcHash)(HANDLE hFile, DWORD *pcbHash, BYTE *pbHash, DWORD dwFlags);



class CWinTrustAPI 
{
public:

	// Do not use the in global constructor, and avoid global contruction at all.
	static __declspec(noinline) CWinTrustAPI &getStatic()
	{
		static cSysCriticalSection cs;
		ODSFuncS( _FormatT( L"CS %X", &cs ) );

		SYS_CRITICAL_SECTION_BLOCK( &cs );

		static CWinTrustAPI dll;
		return dll;
	}

	CWinTrustAPI() 
	{
		ODSFunc;
		memset(this, 0, sizeof(CWinTrustAPI));
		Init();
	};

	~CWinTrustAPI() 
	{
		ODSFunc;
		Fini();
	}

	LONG WinVerifyTrust(HWND hWnd,  GUID* pgActionID,  WINTRUST_DATA* pWinTrustData) 
	{
		ODSFunc;

		if ( fWinVerifyTrust )
		{
	ODS_fl;
			LONG lRes = fWinVerifyTrust(hWnd, pgActionID, pWinTrustData);
	ODS_fl;
			return lRes;
		}
		return ERROR_INVALID_FUNCTION;
	}

	CRYPT_PROVIDER_DATA* WTHelperProvDataFromStateData(HANDLE hStateData) 
	{
		if (fWTHelperProvDataFromStateData) {
			return fWTHelperProvDataFromStateData(hStateData);
		}
		return NULL;
	}

	CRYPT_PROVIDER_SGNR* WTHelperGetProvSignerFromChain(CRYPT_PROVIDER_DATA* pProvData,  DWORD idxSigner,  BOOL fCounterSigner,  DWORD idxCounterSigner)
	{
		if (fWTHelperGetProvSignerFromChain) {
			return fWTHelperGetProvSignerFromChain(pProvData, idxSigner, fCounterSigner, idxCounterSigner);
		}
		return NULL;	
	}

	CRYPT_PROVIDER_CERT* WTHelperGetProvCertFromChain(CRYPT_PROVIDER_SGNR* pSgnr, DWORD idxCert) 
	{
		if (fWTHelperGetProvCertFromChain) {
			return fWTHelperGetProvCertFromChain(pSgnr, idxCert);
		}
		return NULL;
	}

	DWORD CertNameToStr(DWORD dwCertEncodingType, PCERT_NAME_BLOB pName,  DWORD dwStrType,  LPTSTR psz,  DWORD csz)
	{
		if (fCertNameToStr) {
			return fCertNameToStr(dwCertEncodingType, pName, dwStrType, psz, csz);
		}
		SetLastError(ERROR_INVALID_FUNCTION);
		return 0;
	}

	BOOL CryptCATAdminCalcHashFromFileHandle(HANDLE hFile, DWORD *pcbHash, BYTE *pbHash, DWORD dwFlags)
	{
		if (fCATAdminCalcHash) {
			return fCATAdminCalcHash(hFile, pcbHash, pbHash, dwFlags);
		}
		SetLastError(ERROR_INVALID_FUNCTION);
		return 0;
	}

	BOOL CryptCreateKeyIdentifierFromCSP(
		DWORD dwCertEncodingType,
		LPCSTR pszPubKeyOID,
		const PUBLICKEYSTRUC *pPubKeyStruc,
		DWORD cbPubKeyStruc,
		DWORD dwFlags,
		void *pvReserved,
		BYTE *pbHash,
		DWORD *pcbHash
	)
	{
		typedef WINCRYPT32API BOOL WINAPI typeCryptCreateKeyIdentifierFromCSP(
			__in DWORD dwCertEncodingType,
			__in_opt LPCSTR pszPubKeyOID,
			__in_bcount(cbPubKeyStruc) const PUBLICKEYSTRUC *pPubKeyStruc,
			__in DWORD cbPubKeyStruc,
			__in DWORD dwFlags,
			__reserved void *pvReserved,
			__out_bcount_part_opt(*pcbHash, *pcbHash) BYTE *pbHash,
			__inout DWORD *pcbHash
	    );

		static typeCryptCreateKeyIdentifierFromCSP *proc = (typeCryptCreateKeyIdentifierFromCSP *)::GetProcAddress(hCrypt32 , "CryptCreateKeyIdentifierFromCSP");

		if ( proc == NULL )
		{
			SetLastError(ERROR_INVALID_FUNCTION);
			return false;
		}

		return proc( dwCertEncodingType, pszPubKeyOID, pPubKeyStruc, cbPubKeyStruc, dwFlags, pvReserved, pbHash, pcbHash );
	}



private:

	void Fini()
	{
		ODSFuncS( _FormatT( L"%X, %X", hCrypt32, hWinTrust ) );

		isOk = false;

		if (hCrypt32) {
			FreeLibrary(hCrypt32);
		}

		if (hWinTrust) {
			FreeLibrary(hWinTrust);
		}

		memset(this, 0, sizeof(CWinTrustAPI));
	}

	void Init()
	{
		ODSFunc;

		hCrypt32 = LoadLibraryA( "Crypt32.dll" );
		hWinTrust = LoadLibraryA( "Wintrust.dll" );

		if (!hCrypt32 || !hWinTrust) {
			Fini();
		}

		fWinVerifyTrust = (CWinVerifyTrust)::GetProcAddress(hWinTrust, "WinVerifyTrust");
		if ( fWinVerifyTrust == NULL ) ODS_fls( L"Assertion!" );

		fWTHelperProvDataFromStateData = (CWTHelperProvDataFromStateData)::GetProcAddress(hWinTrust, "WTHelperProvDataFromStateData");
		fWTHelperGetProvSignerFromChain = (CWTHelperGetProvSignerFromChain)::GetProcAddress(hWinTrust, "WTHelperGetProvSignerFromChain");
		fWTHelperGetProvCertFromChain = (CWTHelperGetProvCertFromChain)::GetProcAddress(hWinTrust, "WTHelperGetProvCertFromChain");
		fCATAdminCalcHash = (CCATAdminCalcHash)::GetProcAddress(hWinTrust, "CryptCATAdminCalcHashFromFileHandle");


#ifdef UNICODE
		fCertNameToStr							= (CCertNameToStr					)::GetProcAddress(hCrypt32 , "CertNameToStrW");
#else
		fCertNameToStr							= (CCertNameToStr					)::GetProcAddress(hCrypt32 , "CertNameToStrA");
#endif

		isOk =	(NULL != fWinVerifyTrust) &&
				(NULL != fWTHelperProvDataFromStateData) &&
				(NULL != fWTHelperGetProvSignerFromChain) &&
				(NULL != fWTHelperGetProvCertFromChain) &&
				(NULL != fCertNameToStr);

		if (!isOk)
		{
			ODS_fls( L"Assertion!" );
			Fini();
		}
	}

	HMODULE	hCrypt32;
	HMODULE	hWinTrust;
	bool	isOk;

	CWinVerifyTrust						fWinVerifyTrust;
	CWTHelperProvDataFromStateData		fWTHelperProvDataFromStateData;
	CWTHelperGetProvSignerFromChain		fWTHelperGetProvSignerFromChain;
	CWTHelperGetProvCertFromChain		fWTHelperGetProvCertFromChain;
	CCertNameToStr						fCertNameToStr;
	CCATAdminCalcHash					fCATAdminCalcHash;
};


/*
typedef BOOL (WINAPI * ENABLE64)(BOOL);
typedef BOOL (WINAPI * DISABLE64)(VOID **);
typedef BOOL (WINAPI * REVERT64)(VOID *);
typedef BOOL (WINAPI * ISWOW64)(HANDLE, BOOL *);

class CFSRedirHelper 
{
public:
	CFSRedirHelper() 
	{
		if (!m_bInit) {
			InitWow64();
		}
		m_bInit = true;
		m_bRedirDisabled = false;
	}

	~CFSRedirHelper() 
	{
		RevertRedirector();
	}

	void	DisableRedirector() 
	{
		if (!m_bWow64 || m_bRedirDisabled) {
			return;
		}
		m_bRedirDisabled = m_fnDisable(&m_RedirValue);
		return;
	}

	void	RevertRedirector() 
	{
		if (m_bWow64 && m_bRedirDisabled) {
			m_fnRevert(m_RedirValue);
		}
		m_bRedirDisabled = false;
	}

	bool	FileExists(LPCTSTR Filename) 
	{
		if (NULL == Filename) {
			return false;
		}
		DWORD	attr;
		if (!m_bWow64 || m_bRedirDisabled) {
			// either wow64 is not available or 
			// redir is already disabled
			attr = ::GetFileAttributes(Filename);
		} else {
			DisableRedirector();
			attr = ::GetFileAttributes(Filename);
			RevertRedirector();
		}
		if (INVALID_FILE_ATTRIBUTES == attr) {
			return false;
		}
		if (FILE_ATTRIBUTE_DIRECTORY & attr) {
			return false;
		}
		if (FILE_ATTRIBUTE_OFFLINE & attr) {
			return false;
		}
		return true;
	}
	static BOOL IsWow64() { return m_bWow64; }
	static BOOL IsFileRedirected(LPCTSTR pszName) 
	{
		if (!IsWow64()) {
			return FALSE;
		}
		if (NULL == pszName) {
			return FALSE;
		}
		int i;
		for (i = 0; ; i++) {
			if (0 == m_System32[i]) {
				return TRUE;
			}
			if (_totupper(pszName[i]) != _totupper(m_System32[i])) {
				return FALSE;
			}
		}
		return FALSE;
	}
private:

	PVOID	m_RedirValue;
	BOOL	m_bRedirDisabled;

private:
	static void InitWow64();

	static BOOL	m_bWow64;
	static BOOL m_bInit;
	static DISABLE64	m_fnDisable;
	static REVERT64		m_fnRevert;
	static TCHAR		m_System32[MAX_PATH];
};


BOOL		CFSRedirHelper::m_bWow64 = FALSE;
BOOL		CFSRedirHelper::m_bInit = FALSE;
DISABLE64	CFSRedirHelper::m_fnDisable;
REVERT64	CFSRedirHelper::m_fnRevert;
TCHAR		CFSRedirHelper::m_System32[MAX_PATH];

void CFSRedirHelper::InitWow64()
{
	HMODULE	hDll = GetModuleHandle(_T("kernel32.dll"));	
	ISWOW64 fnIsWow = (ISWOW64)::GetProcAddress(hDll, "IsWow64Process");
	m_bWow64 = false;
	m_fnDisable = NULL;
	m_fnRevert = NULL;
	::GetSystemDirectory(m_System32, MAX_PATH);

	if (NULL == fnIsWow) {
		return;
	}
	if (!fnIsWow(GetCurrentProcess(), &m_bWow64)) {
		m_bWow64 = false;
		return;
	}
	m_fnDisable = (DISABLE64)::GetProcAddress(hDll, "Wow64DisableWow64FsRedirection");
	if (!m_fnDisable) {
		m_bWow64 = false;
		return;
	}
	m_fnRevert = (REVERT64)::GetProcAddress(hDll, "Wow64RevertWow64FsRedirection");
	if (!m_fnRevert) {
		m_fnDisable = NULL;
		m_bWow64 = false;
		return;
	}
}
*/

CSignatureCheck::CSignatureCheck(void)
{
	memset( m_subjectKeyID, 0, sizeof(m_subjectKeyID) );
}

CSignatureCheck::~CSignatureCheck(void)
{
}

#define COUNT_OF(x) (sizeof(x)/sizeof(x[0]))

TCHAR STR_NO_SIGN[] = _T("File not signed");
TCHAR STR_DISTRUST[] = _T("File signer or publisher is not trusted in system settings");
TCHAR STR_SECURITY[] = _T("File signer or publisher is not trusted in user settings");
TCHAR STR_GENERIC[] = _T("Generic distrust");
TCHAR STR_NO_NAME[] = _T("Failed to retreive signer name");
TCHAR STR_NOT_JETICO[] = _T("File is not signed by Jetico");
TCHAR STR_SELF_FAIL[] = _T("Could not read own signature");
TCHAR STR_SUCCESS[] = _T("Success");


//
// sha1 = 7593fe20d6a8e2ba62969844467a4aceb5cd39fc, Expiration Date 24.09.2018
//	Note!! When certificate is changed we have to change both subjectT and JeticoKeyIdentifierT
//

TCHAR *subjectT[] = {
	_T("C=FI, S=Finland, L=Espoo, O=\"Jetico, Inc.\", OU=Sales, OU=GeoTrust Code Signing, CN=\"Jetico, Inc.\""),
	_T("C=FI, S=Uusimaa, L=Espoo, O=Jetico Inc. Oy, OU=Digital ID Class 3 - Microsoft Software Validation v2, CN=Jetico Inc. Oy"),
	_T("C=FI, S=Uusimaa, L=Espoo, O=Jetico Inc. Oy, CN=Jetico Inc. Oy"),
	_T("C=FI, L=Espoo, O=Jetico Inc. Oy, CN=Jetico Inc. Oy"),

	//
	// Note! 
	//	Problem:
	//		New subject adding and BCUpdate.exe.
	//	Description:
	//		We have to issue BCUpdate.exe with new subject sting in package signed by old certificate.
	//		Otherwise automatic update by old BCUpdate.exe is impassible it refises to install package signed with the new certificate!
	//
};

//
//	Subject public key identifier
//		It was taken from signed executable Properties->Digital Signatures->View Certificate->Details->Subject Key Identifier.
//		It is required for additional check signing, because there is Russian hacker who signed patched executable by 
//		his self generated certificate with jetico name.
//		It is hard to create same public key so we'll check the public key.
//	Really the public key identifier is SHA1 hash of public key info. We use cAdvApiForHash::MakeHash instead of Certificte lib here
//


BYTE JeticoKeyIdentifierT[][20] = {
	/*
	Issuer:
		CN = DigiCert SHA2 High Assurance Code Signing CA
		OU = www.digicert.com
		O = DigiCert Inc
		C = US
	*/
	{ 0x6a, 0xd7, 0xcf, 0xf8, 0x75, 0x02, 0x84, 0xb5, 0xfa, 0x47, 0x9c, 0x8e, 0xc5, 0x2e, 0xc1, 0x5a, 0x39, 0x10, 0x5f, 0xf8 },

	/*
	Issuer:
		CN = DigiCert High Assurance Code Signing CA-1
		OU = www.digicert.com
		O = DigiCert Inc
		C = US
	*/
	{ 0xce, 0x52, 0x08, 0xe7, 0x77, 0x5f, 0xaa, 0x6c, 0x0b, 0xee, 0x80, 0xca, 0x2e, 0x3f, 0x82, 0xa1, 0xae, 0xb3, 0x9f, 0xaa },

	/*
	Issuer:
		CN = DigiCert SHA2 High Assurance Code Signing CA
		OU = www.digicert.com
		O = DigiCert Inc
		C = US
	ValidTo:
		16.10.2021 19:00:00
	Subject:
		CN = Jetico Inc. Oy
		O = Jetico Inc. Oy
		L = Espoo
		C = FI
	*/
	{ 0x06, 0x1a, 0x39, 0x8f, 0x54, 0xeb, 0x25, 0xe9, 0xe2, 0xbb, 0xa3, 0xbe, 0x67, 0xbe, 0x5a, 0x94, 0x8d, 0xca, 0x5f, 0xd5},
};

#ifdef _UNICODE

DWORD CSignatureCheck::ReadCatalogSignatureText(LPCTSTR Filename, LPCTSTR Catalog, LPTSTR Buffer, DWORD &Length)
{
	ODSFuncS( _FormatT( L"%s, %s", _getWcs( Filename ), _getWcs( Catalog ) ) );
	LONG            ret;
    DWORD           err;
	CWinTrustAPI	&wt = CWinTrustAPI::getStatic();

	if (!Filename || !Buffer || 1 > Length) {
		return ERROR_INVALID_PARAMETER;
	}

    WINTRUST_CATALOG_INFO CatalogData;
    memset(&CatalogData, 0, sizeof(CatalogData));
    CatalogData.cbStruct = sizeof(WINTRUST_CATALOG_INFO);
	CatalogData.pcwszCatalogFilePath = Catalog;
	CatalogData.pcwszMemberFilePath = Filename;
	CatalogData.pcwszMemberTag = NULL;

	GUID WVTPolicyGUID = WINTRUST_ACTION_GENERIC_VERIFY_V2;
    WINTRUST_DATA WinTrustData;

    // Initialize the WinVerifyTrust input data structure.
    memset(&WinTrustData, 0, sizeof(WinTrustData));
    WinTrustData.cbStruct = sizeof(WinTrustData);
    WinTrustData.pPolicyCallbackData = NULL;	// Use default code signing EKU.
    WinTrustData.pSIPClientData = NULL;			// No data to pass to SIP.
    WinTrustData.dwUIChoice = WTD_UI_NONE;		// Disable WVT UI.
    WinTrustData.fdwRevocationChecks = WTD_REVOKE_NONE;	// No revocation checking.
	WinTrustData.dwUnionChoice = WTD_CHOICE_CATALOG;		// Verify an embedded signature on a file.
    WinTrustData.dwStateAction = WTD_STATEACTION_VERIFY;
    WinTrustData.hWVTStateData = NULL;			// Not applicable for default verification of embedded signature.
    WinTrustData.pwszURLReference = NULL;		// Not used.
    WinTrustData.dwProvFlags = WTD_REVOCATION_CHECK_NONE | WTD_SAFER_FLAG;
    WinTrustData.dwUIContext = 0;
    WinTrustData.pCatalog = &CatalogData;		// Set pCatalog.


	LPWSTR pszMemberTag = NULL;
	HANDLE hFile = CreateFileW( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );      
    if (INVALID_HANDLE_VALUE != hFile) {      
	    BYTE byHash[100];   
		DWORD dwCnt = 100;      
		wt.CryptCATAdminCalcHashFromFileHandle( hFile, &dwCnt, byHash, 0 );      
		CloseHandle( hFile );      
	     
		pszMemberTag = new WCHAR[dwCnt * 2 + 1];      
		for ( DWORD dw = 0; dw < dwCnt; ++dw )      
		{      
			wsprintfW( &pszMemberTag[dw * 2], L"%02X", byHash[dw] );      
		}      
		CatalogData.pcwszMemberTag = pszMemberTag;
	}      
     

	//fsr.DisableRedirector();
    ret = wt.WinVerifyTrust(NULL, &WVTPolicyGUID, &WinTrustData);
	//fsr.RevertRedirector();

	ODSFuncRes( (DWORD)ret );

	if ( ERROR_SUCCESS == ret )
		err = saveSubjectId( wt, WinTrustData.hWVTStateData, Buffer, Length );
	else
		return ret;

	if (pszMemberTag) {
		delete[] pszMemberTag;
	}

	err = ExtractSignerText(&WinTrustData, Buffer, Length);

	// Close WinTrustData.hWVTStateData handle
	WinTrustData.dwStateAction = WTD_STATEACTION_CLOSE;

	//fsr.DisableRedirector();
	wt.WinVerifyTrust(NULL, &WVTPolicyGUID, &WinTrustData);
	//fsr.RevertRedirector();

	return err ? ERROR_SUCCESS : ERROR_INVALID_DATA;
}

#endif _UNICODE

DWORD CSignatureCheck::ExtractSignerText(WINTRUST_DATA *WinTrust, LPTSTR Buffer, DWORD &Length) 
{
	ODSFunc;

	DWORD err = 0;
	CRYPT_PROVIDER_DATA *pd = NULL;
	CRYPT_PROVIDER_SGNR *ps = NULL;
	CRYPT_PROVIDER_CERT	*pc = NULL;
	CWinTrustAPI	&wt = CWinTrustAPI::getStatic();

	pd = wt.WTHelperProvDataFromStateData(WinTrust->hWVTStateData);
	//ODSF( L"WTHelperProvDataFromStateData = %X", pd );

	if (pd) {
		ps = wt.WTHelperGetProvSignerFromChain(pd, 0, FALSE, 0);
		// ODSF( L"WTHelperGetProvSignerFromChain = %X", ps );
	}
	if (ps) {
		pc = wt.WTHelperGetProvCertFromChain(ps,  0);
		// ODSF( L"WTHelperGetProvCertFromChain = %X", pc );
	}
	if (pc) {
		err = wt.CertNameToStr(X509_ASN_ENCODING, &pc->pCert->pCertInfo->Subject, CERT_X500_NAME_STR, Buffer, Length);
		// ODSF( L"CertNameToStr = %d", err );
		if (err) {
			Length = err;
		}
	}
	return err;
}

DWORD CSignatureCheck::saveSubjectId( CWinTrustAPI &wt, HANDLE hWVTStateData, LPTSTR Buffer, DWORD Length )
{
	CRYPT_PROVIDER_DATA *pd = NULL;
	CRYPT_PROVIDER_SGNR *ps = NULL;
	CRYPT_PROVIDER_CERT	*pc = NULL;

	DWORD err = 0;
	pd = wt.WTHelperProvDataFromStateData( hWVTStateData );
	//ODSF( L"WTHelperProvDataFromStateData = %X", pd );


	if (pd) {
		ps = wt.WTHelperGetProvSignerFromChain(pd, 0, FALSE, 0);
		// ODSF( L"WTHelperGetProvSignerFromChain = %X", ps );
	}

	if (ps) {
		pc = wt.WTHelperGetProvCertFromChain(ps,  0);
		// ODSF( L"WTHelperGetProvCertFromChain = %X", pc );
	}

	if (pc)
	{
		err = wt.CertNameToStr(X509_ASN_ENCODING, &pc->pCert->pCertInfo->Subject, CERT_X500_NAME_STR, Buffer, Length);

		if (err) {
			Length = err;
		}

		DWORD cbHash = sizeof( m_subjectKeyID );

		bool ok = _makeHashSHA1(
				pc->pCert->pCertInfo->SubjectPublicKeyInfo.PublicKey.pbData,
				pc->pCert->pCertInfo->SubjectPublicKeyInfo.PublicKey.cbData,
				m_subjectKeyID,
				&cbHash
			);
#ifdef _ODS
		wchar_t bf[256];
		bytes2string( bf, maxLen( bf ), m_subjectKeyID, sizeof(m_subjectKeyID) );
		ODS_fls( _FormatT( L"m_subjectKeyID %s", bf ) );
#endif
	}

	return err;
}

DWORD CSignatureCheck::ReadFileSignatureText(LPCTSTR Filename, LPTSTR Buffer, DWORD &Length) 
{
	ODSFuncS( _getWcs( Filename ) );

	LONG            ret;
    DWORD           err = ERROR_INVALID_DATA;
	CWinTrustAPI	&wt = CWinTrustAPI::getStatic();

	if (!Filename || !Buffer || 1 > Length) {
		return ERROR_INVALID_PARAMETER;
	}

#ifdef UNICODE
	const wchar_t *FilenameW = Filename;
#else
	wchar_t FilenameW[ MAX_PATH ];
	size_t size;
	mbstowcs_s( &size, FilenameW, Filename, _TRUNCATE );
#endif

    WINTRUST_FILE_INFO FileData;
    memset(&FileData, 0, sizeof(FileData));
    FileData.cbStruct = sizeof(WINTRUST_FILE_INFO);
    FileData.pcwszFilePath = FilenameW;
    FileData.hFile = NULL;
    FileData.pgKnownSubject = NULL;

	GUID WVTPolicyGUID = WINTRUST_ACTION_GENERIC_VERIFY_V2;
    WINTRUST_DATA WinTrustData;

    // Initialize the WinVerifyTrust input data structure.
    memset(&WinTrustData, 0, sizeof(WinTrustData));
    WinTrustData.cbStruct = sizeof(WinTrustData);
    WinTrustData.pPolicyCallbackData = NULL;	// Use default code signing EKU.
    WinTrustData.pSIPClientData = NULL;			// No data to pass to SIP.
    WinTrustData.dwUIChoice = WTD_UI_NONE;		// Disable WVT UI.
	WinTrustData.fdwRevocationChecks = WTD_REVOKE_NONE;	// No revocation checking.
    WinTrustData.dwUnionChoice = WTD_CHOICE_FILE;		// Verify an embedded signature on a file.
    WinTrustData.dwStateAction = WTD_STATEACTION_VERIFY;
    WinTrustData.hWVTStateData = NULL;			// Not applicable for default verification of embedded signature.
    WinTrustData.pwszURLReference = NULL;		// Not used.
    WinTrustData.dwProvFlags = WTD_REVOCATION_CHECK_NONE | WTD_SAFER_FLAG;
    WinTrustData.dwUIContext = 0;
    WinTrustData.pFile = &FileData;				// Set pFile.

ODS_fls( _FormatT( L"WinVerifyTrust start" )  );

	//fsr.DisableRedirector();
    ret = wt.WinVerifyTrust( (HWND)INVALID_HANDLE_VALUE, &WVTPolicyGUID, &WinTrustData);

ODS_fls( _FormatT( L"WinVerifyTrust = %d", ret )  );

	//fsr.RevertRedirector();

	ODSFuncRes( (DWORD)ret );

	if ( ERROR_SUCCESS == ret )
		err = saveSubjectId( wt, WinTrustData.hWVTStateData, Buffer, Length );

	// Close WinTrustData.hWVTStateData handle
	WinTrustData.dwStateAction = WTD_STATEACTION_CLOSE;


	//fsr.DisableRedirector();
	wt.WinVerifyTrust(NULL, &WVTPolicyGUID, &WinTrustData);
	//fsr.RevertRedirector();

	ODSFuncRes( ret );

	if ( ERROR_SUCCESS != ret )
		return ret;

	return err ? ERROR_SUCCESS : ERROR_INVALID_DATA;
}

static bool regReadDW( HKEY hRoot, const char *keyName, const char *valueName, REGSAM flags, DWORD &dwValue )
{
	HKEY hKey;
	LONG lRes = RegOpenKeyExA( hRoot, keyName, 0, KEY_READ | flags, &hKey );

	if ( lRes != ERROR_SUCCESS )
		return false;

	DWORD dwSize = sizeof( dwValue );
	DWORD dwType;

	lRes = RegQueryValueExA( hKey, valueName, NULL, &dwType, (BYTE*)&dwValue, &dwSize );

	RegCloseKey( hKey );

	if ( ( lRes == ERROR_SUCCESS ) && ( dwSize == sizeof( dwValue ) ) && ( dwType == REG_DWORD ) ) 
		return true;

	return false;
}

static bool IgnoreCkeck()
{
	static int iIgnore = -1;

	if ( iIgnore == -1 )
	{
		const char *keyName = "SOFTWARE\\Jetico";
		const char *valueName = "IgnoreSignatureCheck";

		DWORD dwIgnore = 0;
		if ( regReadDW( HKEY_LOCAL_MACHINE, keyName, valueName, KEY_WOW64_64KEY, dwIgnore ) || 
			 regReadDW( HKEY_LOCAL_MACHINE, keyName, valueName,				  0, dwIgnore ) )
		{
			iIgnore = ( dwIgnore == 1 );
		}
		else
			iIgnore = 0;
	}
	
	return iIgnore == 1;
}

static bool cInstanceHolder_isWindowsVista()
{
	ODSFunc;
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	bool bIsWindowsVista = ( osvi.dwPlatformId == VER_PLATFORM_WIN32_NT ) && ( osvi.dwMajorVersion >= 6 );

	ODSFuncRes( _FormatT( L"PlatformId %X, MajorVersion %X", osvi.dwPlatformId, osvi.dwMajorVersion ) );
	return bIsWindowsVista;
}

#ifndef maxLen
#define maxLen( T ) ( sizeof( T ) / sizeof( *T ) )
#endif

bool CSignatureCheck::isValidKeyID()
{
	// 18.01.2017 It has to be checked too here.
	if ( IgnoreCkeck() || !cInstanceHolder_isWindowsVista() )
	{
		ODS_fls( L"Skipped for XP" );
		return true;
	}

	const int keyIdSize = sizeof( JeticoKeyIdentifierT[0] );

	for( int i = 0; i < maxLen( JeticoKeyIdentifierT ); i++ )
	{
		if ( ( memcmp( JeticoKeyIdentifierT[i], m_subjectKeyID, keyIdSize ) ) == 0 )
			return true;
	}

#ifdef _ODS
	wchar_t bf[256];
	bytes2string( bf, maxLen( bf ), m_subjectKeyID, sizeof(m_subjectKeyID) );
	ODS_fls( L"No one compatible key ID is found!" );
	ODS_fls( _FormatT( L"Cur. %s", bf ) );

	int Q = maxLen( JeticoKeyIdentifierT );

	for( int i = 0; i < Q; i++ )
	{
		bytes2string( bf, maxLen( bf ), JeticoKeyIdentifierT[i], keyIdSize );
		ODS_fls( _FormatT( L"%d/%d. %s", i, Q, bf ) );
	}
#endif

	return false;
}


BOOL CSignatureCheck::VerifyEmbeddedSignature(LPCTSTR Filename, LPCTSTR Catalog, HKEY hKey) 
{
	ODSFuncS( cWCStr( Filename ).m_szBody );

	if ( IgnoreCkeck() )
	{
		ODS_fls( L"Check signature is skipped!" );
		return true;
	}

	if ( Filename == NULL )
		return false;

	TCHAR	sign[MAX_PATH];
	DWORD	len;

	if ( hKey != NULL )
	{
		RegDeleteValue(hKey, _T("Status"));
		RegDeleteValue(hKey, _T("Signer"));
	}

	len = MAX_PATH;
	if ( Catalog != NULL )
	{
#ifdef _UNICODE
		m_status = ReadCatalogSignatureText(Filename, Catalog, sign, len);
#else
		ODS_fls( L"Assertion! ReadCatalogSignatureText is unicode!" );
		m_status = -1;
#endif
	}
	else
		m_status = ReadFileSignatureText(Filename, sign, len);

	//len = MAX_PATH;
	//m_status = ReadFileSignatureText(Filename, sign, len);

	m_statusStr = STR_GENERIC;

	if ( ERROR_SUCCESS != m_status )
	{
		switch (m_status)
		{
			case TRUST_E_NOSIGNATURE:		// The file was not signed or had a signature that was not valid.
				m_statusStr = STR_NO_SIGN;
				break;
			case TRUST_E_EXPLICIT_DISTRUST:	// The hash that represents the subject or the publisher is not allowed by the admin or user.
				m_statusStr = STR_DISTRUST;
				break;
			case TRUST_E_SUBJECT_NOT_TRUSTED:	// The user clicked "No" when asked to install and run.
				m_statusStr = STR_DISTRUST;
				break;
			case CRYPT_E_SECURITY_SETTINGS:
				// The hash that represents the subject or the publisher was not explicitly trusted by the admin and the 
				// admin policy has disabled user trust. No signature, publisher or time stamp errors.
				m_statusStr = STR_SECURITY;
				break;
			default: // The UI was disabled in dwUIChoice or the admin policy has disabled user trust. 
				m_statusStr = STR_GENERIC;
				break;
		}

		// A certificate chain could not be built to a trusted root authority.
		if ( IgnoreCkeck() || !cInstanceHolder_isWindowsVista() )
		{
			// v. 3.06.6.4 - 02.11.2016 Disable error for WIndows XP
			//	Bug #3042 (BCWipe for Windows ) - Do not check digital signature on old systems
			ODS_fls( _FormatT( L"Warning! Error 0x%X skipped for XP", m_status ) );
			return true;
		}

		if ( hKey != NULL )
			RegSetValueEx(hKey, m_statusStr, 0, REG_SZ, (const BYTE *)m_statusStr, (DWORD)_tcslen( m_statusStr ) + 1 );

		ODSFuncRes( cWCStr( m_statusStr ).m_szBody );

		return FALSE;
	}

	bool ok = false;

	if ( len != 0 )
	{
		if ( hKey != NULL ) RegSetValueEx(hKey, _T("Signer"), 0, REG_SZ, (const BYTE *)sign, len*sizeof(TCHAR));
		ODSF( L"Signer: %s", cWCStr( sign ).m_szBody );

		bool bFound = false;

		for ( int i = maxLen( subjectT ); i--; )
		{
			if ( _tcscmp( sign, subjectT[i] ) == 0 )
			{
				bFound = true;
				break;
			}
		}

		if ( bFound )
		{
			m_statusStr = STR_SUCCESS;
			ok = true;
		}
		else
		{
			TCHAR self[MAX_PATH], path[MAX_PATH];
			::GetModuleFileName(NULL, path, MAX_PATH);

			len = MAX_PATH;
			m_status = ReadFileSignatureText(path, self, len);

			if ( ERROR_SUCCESS == m_status )
			{
				if  (len > 0 && 0 == _tcscmp(sign, self))
				{
					m_status = eSuccess;
					m_statusStr = STR_SUCCESS;
					ok = true;
				}
				else
				{
					m_status = eNotJetico;
					m_statusStr = STR_NOT_JETICO;
				}
			}
			else
			{
				m_statusStr = STR_SELF_FAIL;
			}
		}
	}
	else
	{
		m_statusStr = STR_NO_NAME;
	}

	if ( hKey != NULL )
		RegSetValueEx(hKey, m_statusStr, 0, REG_SZ, (const BYTE *)m_statusStr, (DWORD)_tcslen( m_statusStr ) + 1 );

#ifdef _ODS
	
	bool bValidPublicKey = isValidKeyID();

	if ( ok && !bValidPublicKey )
	{
		ok = false;
		m_statusStr = _T("Signature Public Key Is Invalid!");
		assert( false );
	}
#endif

	ODSFuncRes( cWCStr( m_statusStr ).m_szBody );

	return ok;
}

//static 
bool CSignatureCheck::getRegDelay()
{
	HKEY hKey;
	LONG lRes = RegOpenKeyExA( HKEY_CURRENT_USER, "Software\\Jetico", 0, KEY_READ | KEY_WRITE, &hKey );
	if ( lRes != ERROR_SUCCESS )
		return false;

	time_t hakedTime;
	DWORD dwSize = sizeof( hakedTime );
	DWORD dwType;

	const char *szVal = "hkTime";

	lRes = RegQueryValueExA( hKey, szVal, NULL, &dwType, (BYTE*)&hakedTime, &dwSize );

	time_t curTime = time( 0 );
	
	bool bWasDelayed = false;

	if ( ( lRes == ERROR_SUCCESS ) && ( dwSize == sizeof( hakedTime ) ) && ( dwType == REG_DWORD ) ) 
	{
		__int64 diff = ( curTime - hakedTime );
		__int64 diffDays = diff / ( 60 * 60 * 24 );

		bWasDelayed = diffDays > 15;
	}
	else
		lRes = RegSetValueExA( hKey, szVal, NULL, REG_DWORD, (BYTE*)&curTime, sizeof( curTime ) );

	RegCloseKey( hKey );

	return bWasDelayed;
}
