#ifndef _UnicodeFuncs_
#define _UnicodeFuncs_

#include "windefs.h"
#include "ODS.h"
#include "AutoDeallocate.h"
#include "cSilentMode.h"
#include "cSysCriticalSection.h"

#ifndef _WIN32
#include "../misc/misc.h"
#endif

//
//	Common functions for _WIN32 and for Unix will be included here
//

template < class T > T RoundUpExp( T value, size_t exp )
{
	T blockSize_minus_1 = ((T)1 << exp) - 1;
	return ( ( (value) + blockSize_minus_1 ) & ~blockSize_minus_1 );
}

template < class T > T RoundDownExp( T value, size_t exp )
{
	T blockSize_minus_1 = ((T)1 << exp) - 1;
	return ( (value) & ~blockSize_minus_1 );
}

template < class T > T RoundUp( T value, T blockSize )
{
	T res = ( ( ( (value) +  (blockSize) - 1 ) / ( blockSize ) ) * ( blockSize ) );
	return res;
}

template < class T > T RoundDown( T value, T blockSize )
{
	return ( ( (value) / ( blockSize ) ) * ( blockSize ) );
}

template < class T > T RoundTail( T value, T blockSize )
{
	return Round( (value), blockSize ) - (value);
}


template < class T > T Round( T value, T blockSize ) { return  RoundUp( value, blockSize ); }

template <size_t _Size> inline wchar_t *copyString( wchar_t (&_Dst)[_Size], const wchar_t *_Src, bool setLastZero = true ) { return copyString( _Dst, _Src, _Size, setLastZero );  }
extern wchar_t *copyString( wchar_t *dst, const wchar_t *src, int dstMaxLen, bool setLastZero = true );

template <size_t _Size> inline wchar_t *catString( wchar_t (&_Dst)[_Size], const wchar_t *_Src, bool setLastZero = true ) { return catString( _Dst, _Src, _Size, setLastZero );  }
extern wchar_t *catString( wchar_t *dst, const wchar_t *src, int dstMaxLen, bool setLastZero = true );

extern wchar_t *newString( const wchar_t *str, int len = -1 );
extern char *newString( const char *str, int len = -1 );
extern void deleteString( const wchar_t **pValue );
extern void deleteString( wchar_t **pValue );
extern void deleteString( const char **pValue );
extern void deleteString( char **pValue );

template < class T > class cAutoDelete
{
public:
	T *m_pObject;

	cAutoDelete( T *pObject = NULL )
	{
		Set( pObject );
	}

	void Set( T *pObject )
	{
		m_pObject = pObject;
	}
	
	~cAutoDelete()
	{
		if ( m_pObject != NULL )
			delete( m_pObject );
	}

	T *Delegate()
	{
		T *obj = m_pObject;
		m_pObject = NULL;
		return obj;

	}
};

class cAutoStringPtr
{
public:
	wchar_t *m_str;
	wchar_t **m_pStr;

	cAutoStringPtr( wchar_t *str = NULL )
	{
		m_str = str;
		m_pStr = &m_str;
	}
	
	cAutoStringPtr( wchar_t **pStr )
	{
		m_str = *pStr;
		m_pStr = pStr;
	}

	~cAutoStringPtr()
	{
		deleteString( m_pStr );
	}
};

extern bool updateString( const wchar_t **pPtr, const wchar_t *val , int len = -1 );
extern bool updateString( wchar_t **pPtr, const wchar_t *val , int len = -1 );
extern bool updateString( wchar_t **ptr, const char *val, int len = -1 );
extern bool updateString( char **pPtr, const char *val , int len = -1 );

class cStringT : public cADBlock < wchar_t * >
{
public:
	cStringT( int Q ) : cADBlock ( Q, true )
	{
	}

	void Free()
	{
		for( int i = m_Q; i--; )
			deleteString( m_T + i );
	}
	
	~cStringT()
	{
		cStringT::Free();
	}

	bool Set( int i, const wchar_t *str, int len = -1 )
	{
		if ( i >= m_Q )
			Add( NULL, m_Q - i );

		if ( i >= m_Q )
			return false;

		return updateString( m_T + i, str, len );
	}

	const wchar_t *GetAt( int i )
	{
		if ( i >= m_Q )
			return NULL;

		return m_T[i];
	}

	size_t GetItemLen( int i )
	{
		const wchar_t *s = GetAt(i);

		if ( s == NULL )
			return 0;

		return wcslen( s );
	}

	int Find( const wchar_t *str )
	{
		int i = m_Q;

		while( i-- )
		{
			if ( wcscmp( m_T[i], str ) == 0 )
				break;
		}

		return i;
	}
};

extern wchar_t *makeValidFileName( wchar_t *name );

#define _UniqueString ( _FormatT( L"%s:%d", _CRT_WIDE( __FUNCTION__ ), __LINE__  ) )
#define sizeofBefore( structname, member ) ((int)((char*)(&((structname*)0)->member) - ((char*)((structname*)0))))
#define sizeofAfter( structname, member ) ( (int) ((char*)(&((structname*)0)->member) - ((char*)((structname*)0)) + sizeof( ((structname*)0)->member ) ) )

#define LODWORD(l) ((DWORD)((DWORDLONG)(l) & 0xFFFFffff))
#define HIDWORD(l) ((DWORD)((DWORDLONG)(l) >> 32))
#define MAKEDWORDLONG( Hi, Lo ) ( DWORDLONG(Hi)<<32 | (Lo) )

template < class T > bool copyBuffer( T *bfOut, int *pBfOutSize, const T *bfIn, int bfInSize )
{
	if ( *pBfOutSize < bfInSize )
	{
		*pBfOutSize = bfInSize;
		return false;
	}

	*pBfOutSize = bfInSize;

	for( int i = 0; i < *pBfOutSize; i ++ )
		bfOut[i] = bfIn[i];

	return true;
}

template < typename T > bool checkZero( T *bf, int length )
{
	__int64 *dwlBf = ( __int64 * )bf;
	int dwlSize = length / sizeof( __int64 );

	int i;

	for ( i = 0; i < dwlSize; i++ )
	{
		if ( dwlBf[i] != 0 )
			return false;
	}

	int tailSize = length % sizeof( __int64 );

	if ( tailSize == 0 )
		return true;

	T *tailPos = (T*)( &(dwlBf[ dwlSize ]) );

	for ( i = 0; i < tailSize; i++ )
	{
		if ( tailPos[i] != 0 )
			return false;
	}

	return true;
}

int bitcount( unsigned int a );

#include "cWCSF.h"

#define _getWcs( _MBS ) ( cWCStr( _MBS ).m_szBody )
#define __getWcs( _MBS ) ( _MBS ? cWCStr( _MBS ).m_szBody : NULL )

class cWCStr : public cWCSF
{
public:

	cWCStr( cWCS &wcs, bool bSwap = false )				: cWCSF( wcs, bSwap )	{}
	cWCStr( const wchar_t *wcs = NULL, int len = -1 )	: cWCSF( wcs, len )		{}
	cWCStr( int len )									: cWCSF( len )			{}

	cWCStr( const char *mbs, int len = -1, bool doubleZeroTerminataed = false, UINT codePage = (UINT)-1 );

	int Set( const wchar_t *s, int len = -1 ) { return cWCSF::Set( s, len ); }
	int Set( const char *mbs, int len = -1, UINT codePade = (UINT)-1 )
	{
		cWCStr wcsNew( mbs, len, false, codePade );

		Swap( &wcsNew );

		return m_len;
	}

	int Add( int len ) 	{ return Add( (const wchar_t *)NULL, len ); }
	int Add( wchar_t wc ) { return Add( &wc, 1 ); }
	int Add( const wchar_t *wcs = NULL, int lenAdd = -1 ) { return cWCSF::Add( wcs, lenAdd ); }
	int Add( const unsigned char *mbs, int len ) { return Add( (const char *)mbs, len ); }
	int Add( const char *mbs, int len = -1, UINT codePade = (UINT)-1 )
	{
		cWCStr wcs( mbs, len, false, codePade );
		return cWCSF::Add( wcs.m_szBody );
	}

	//static void *memAllocate( size_t size ) { return (void *)GlobalAlloc( GMEM_FIXED, size ) };
	//static void	memFree( void *p ); { GlobalFree( (HGLOBAL)p ); }

	//const wchar_t *SetInt( int i, int radix = 10 );
	//int AddInt( int i, int radix = 10 );

	const wchar_t *GetSizeText( DWORDLONG size, bool sizeInBytes = false, const wchar_t *szZero = NULL );
	const wchar_t *GetTimeText( time_t timeInSec, int digits = 2 );

#ifdef _WIN32

	const wchar_t *FormatT( const wchar_t *pattern, ... )
	{
		if ( SetLength( 4096 ) == 0 )
			return L"cWCStr::SetLength( 4096 ) returns <NULL>";

		cVAList vl;
		va_start( vl.marker, pattern );

		::FormatT( m_szBody, m_allocatedLen, &vl, pattern );

		Truncate();

		return m_szBody;
	}

	const wchar_t *FormatT( HINSTANCE hInst, int idsPattern, ... );
	const wchar_t *LoadStringT( HINSTANCE hInst, int ids, bool bTranslate = true, bool bCheckParams	= true );
	const wchar_t *LoadIDS( int ids, bool bTranslate = true, bool bCheckParams	= true );


	int GetWindowTextTI( HWND hWnd );
	const wchar_t *GetSystemError( DWORD dwErr = (DWORD)-1 );

	cWCStr( HWND hWnd )
	{
		//m_szBody = NULL;
		//m_len = 0;
		//m_allocatedLen = 0;

		GetWindowTextTI( hWnd );
	}

	const wchar_t *GetWindowTextT( HWND hWnd );

	const wchar_t *Translate( const char *s );
	cWCStr( HINSTANCE hInst, int ids, bool bTranslate = true, bool bCheckParams	= true )
	{
		LoadStringT( hInst, ids, bTranslate, bCheckParams );
	}

	bool RegGetString( HKEY hRoot, LPCWSTR lpSubKey, LPCWSTR lpValue );
	static bool RegSetString( HKEY hRoot, LPCWSTR lpSubKey, LPCWSTR lpValue, const wchar_t *str );

	const wchar_t *ExpandEnvironmentString( const wchar_t *path, bool *bExpanded = NULL );

#endif // _WIN32
};

#ifndef CP_UTF8

#define CP_ACP                    0           // default to ANSI code page
#define CP_OEMCP                  1           // default to OEM  code page
#define CP_MACCP                  2           // default to MAC  code page
#define CP_THREAD_ACP             3           // current thread's ANSI code page
#define CP_SYMBOL                 42          // SYMBOL translations

#define CP_UTF7                   65000       // UTF-7 translation
#define CP_UTF8                   65001       // UTF-8 translation

#endif // #ifndef CP_ACP

//
//	setCodePage
//
UINT setCodePage( UINT newCP = CP_UTF8 );

//
//	_wcstombs,_mbstowcs
//		1. uses use CP_ACP code page for conversion as default, but the page may be changed by setCodePage
//		2. usual conversion functions stop at first zero, the followings may skip one zero and stop at double only
//
int _wcstombs( char *mbs, int mbsLen, const wchar_t *wcs, int wcsLen = -1, bool doubleZeroTerminated = false, UINT codePage = (UINT)-1 );
int _mbstowcs( wchar_t *wcs, int wcsLen, const char *mbs, int mbsLen = -1, bool doubleZeroTerminated = false, UINT codePage = (UINT)-1 );
int wcsLen( const wchar_t *wcs, int maxLen = -1 );
int strLen( const char *str, int maxLen = -1 );
const wchar_t *wcsMove( wchar_t *dst, const wchar_t *src, int len );

const wchar_t *statusStr( bool bRes );
const wchar_t *getWCStr( bool bRes );
const wchar_t *getWCStr( int iRes, int radix = 10 );
const wchar_t *getWCStr( DWORD dwRes );
const wchar_t *getWCStr( const char *aStr, wchar_t *wStr = NULL, int wStrMaxLen = 0, UINT cp = CP_ACP, int aStrLen = -1 );

const char *getMBStr( const wchar_t *wStr, char *aStr = NULL, int aStrMaxLen = 0, UINT cp = CP_ACP );

extern bool urlEncode( cWCStr *wcs, const char *bf, int bfSize );

//
//	End of common block.
//////////////////////////////////////////////////////////////

#ifdef _WIN32

// _W same as _CRT_WIDE
//(_String) L ## _String
#define _W _CRT_WIDE

#define DBG_REP( s, i, mbid ) MessageBoxT( 0, _FormatT( L"%d", i ), s, mbid )

#include <shellapi.h>
#include <shlobj.h>
#include <commdlg.h>
#include <assert.h>

class cCoInitialize
{
	static FARPROC m_procCoInitialize;
	static FARPROC m_procCoUninitialize;

public:

	HRESULT m_hRes;
	
	cCoInitialize( LPVOID pvReserved = NULL, bool bInitialize = true );
	~cCoInitialize();
};

#ifdef _UseHtmlHelp_

#include <htmlhelp.h>

//
//	HtmlHelpT
//		1. HtmlHelp.lib has to be mantioned at included list.
//		2. _UseHtmlHelp_ has to be definded
//
#define HtmlHelpT HtmlHelpW

bool isProgramFilesDir( const wchar_t *dir );
bool isSystemRootDir( const wchar_t *dir, int len = -1 );

#endif

//
//	GetDriveRootDirT
//		Gets the file's/folder's drive or mount point root folder name by means of GetVolumeName.
//	We can get file mount point by the function so it rather may be called as GetFileMountPoint :)
//	Really it splits file path to several values:
//		file path	-> C:\tmp\text.txt
//		root dir	-> C:\
//		volume name	-> \\?\Volume{0cd30cef-bc98-11e0-ac71-806e6f6e6963}\
//		loca path	-> \tmp\text.txt
//
bool GetDriveRootDirT( wchar_t *rootDirOut, const wchar_t *path, bool silent = false, wchar_t *pVolumeName = NULL, const wchar_t **ppLocalPath = NULL );
bool GetRootDirT( wchar_t *buff, const wchar_t *fileName );
bool GetRootDirT( cWCS &out, const wchar_t *fileName );

DWORD WNetGetConnectionT( LPCWSTR lpLocalName, LPWSTR  lpRemoteName, LPDWORD lpnLength );

#define GetFullPathNameT GetFullPathNameW
#define WaitNamedPipeT WaitNamedPipeW
#define OpenSemaphoreT OpenSemaphoreW
#define CreateSemaphoreT CreateSemaphoreW

//
//	renameT
//		_wrename allows to rename busied file on single disk ( is the difference with MoveFile )
//
#define renameT _wrename
#define SetWindowTextT SetWindowTextW


DWORD GetFullPathNameT( LPCWSTR lpFileName, cWCStr *wcs, LPWSTR *lpFilePart );

//
//	ReleaseSemaphore
//		Help function, it just open the named semaphore and release it state.
//
bool ReleaseSemaphoreT( const wchar_t *semaphoreName, LONG lReleaseCount = 1, LPLONG lpPreviousCount = NULL );


// define _CreateHandle_ODS

#ifdef _CreateHandle_ODS
#define CreateFileT( lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile )\
	CreateFileODS( _CRT_WIDE(__FUNCTION__), __LINE__, lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile )
#else
HANDLE
CreateFileT(
    LPCWSTR lpFileName,
    DWORD dwDesiredAccess,
    DWORD dwShareMode,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    DWORD dwCreationDisposition,
    DWORD dwFlagsAndAttributes,
    HANDLE hTemplateFile
    );
#endif

HANDLE CreateFileAndDirectory(
	LPCWSTR logFileName,
    DWORD dwDesiredAccess,
    DWORD dwShareMode,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    DWORD dwCreationDisposition,
    DWORD dwFlagsAndAttributes,
    HANDLE hTemplateFile,
	DWORD *pLastErr = NULL
 );

BOOL
CreateDirectoryT(
    LPCWSTR lpPathName,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
	DWORD *pError = NULL
    );

BOOL
CreateDirectoryExT(
    IN LPCWSTR lpTemplateDirectory,
    IN LPCWSTR lpNewDirectory,
    IN LPSECURITY_ATTRIBUTES lpSecurityAttributes
    );

//
//	RecCreateDirectory
//		Create directory with subdirectories recursively.
//		Returns true if directory is created or already exists.
//
BOOL RecCreateDirectory( const wchar_t *dir );

//
//	CheckFileDirectory
//		Returns true if the file hold directory exist.
//		Otherwise if bCreate == true craetes it by RecCreateDirectory.
//
bool CheckFileDirectory( const wchar_t *filePath, bool bCreate );

BOOL
MoveFileT(
    LPCWSTR lpExistingFileName,
    LPCWSTR lpNewFileName
    );

DWORD
GetFileAttributesT(
    LPCWSTR lpFileName
    );

DWORDLONG
GetFileSreamsAllocatedSizes( LPCWSTR lpFileName );

DWORDLONG
GetFileSizeT( LPCWSTR lpFileName, DWORD *pErr = NULL );

DWORDLONG
GetFileSizeT( HANDLE hFile  );

DWORDLONG 
GetDirSizeT( const wchar_t *dir, int *pDirCount = NULL, int *pFileCount = NULL, bool bWalkStreams = false, bool bSkipReparsePoint = true );

//
//	SetFilePointer
//		Return new file position if success otherwise (DWORDLONG)-1
//
DWORDLONG SetFilePointerT( HANDLE hFile, LONGLONG llDistanceToMove, DWORD dwMoveMethod );

HANDLE
FindFirstFileT(
    LPCWSTR lpFileName,
    LPWIN32_FIND_DATAW lpFindFileData,
	int *pErrCode = NULL
    );

#define FindNextFileT FindNextFileW

BOOL
CopyFileT(
    LPCWSTR lpExistingFileName,
    LPCWSTR lpNewFileName,
    BOOL bFailIfExists
    );

static inline bool
CopyFileT(
    LPCWSTR lpExistingFileName,
    LPCWSTR lpNewFileName,
    BOOL bFailIfExists,
	int tryCount,
	int sleepMs = 100 // 0.1 Sec
    )
{
		for( int i = tryCount; i--; )
		{
			if ( CopyFileT( lpExistingFileName, lpNewFileName, bFailIfExists ) )
				return true;

			Sleep(100);
		}
	
		return false;
}

// is declared at winbase.h as _WIN32_WINNT >= 0x0400 
#if(_WIN32_WINNT < 0x0400)

#define COPY_FILE_FAIL_IF_EXISTS 0x00000001
#define LPPROGRESS_ROUTINE LPVOID
#endif /* _WIN32_WINNT < 0x0400 */

BOOL
CopyFileExT(
    IN LPCWSTR lpExistingFileName,
    IN LPCWSTR lpNewFileName,
    //IN LPPROGRESS_ROUTINE lpProgressRoutine OPTIONAL,
    IN LPVOID lpProgressRoutine OPTIONAL,
    IN LPVOID lpData OPTIONAL,
    IN LPBOOL pbCancel OPTIONAL,
    IN DWORD dwCopyFlags
    );

BOOL
SetFileAttributesT(
    LPCWSTR lpFileName,
    DWORD dwFileAttributes
    );

BOOL
DeleteFileT(
    LPCWSTR lpFileName
    );

BOOL
MoveFileExT( LPCWSTR lpExistingFileName, LPCWSTR lpNewFileName, DWORD dwFlags, DWORD *pError = NULL );

#define MarkDeleteT( fileName ) MoveFileExT( fileName, NULL, MOVEFILE_DELAY_UNTIL_REBOOT )

BOOL CopyUntilRebootT( const wchar_t *szSrcFile , const wchar_t *szDstFile );

BOOL
RemoveDirectoryT(
    LPCWSTR lpPathName,
	bool removeItems = true, // removes items and subfolders
	DWORD *pError = NULL
);

bool RemoveDirectoryContent(
	LPCWSTR lpPathName,
	DWORD *pError = NULL
);


bool CreateShortestName( wchar_t *newName, const wchar_t *oldName, int diffLen, bool bForce = false );


BOOL FileExistsT( const wchar_t *lpPathName, WIN32_FIND_DATAW *pFindData, int *pLastError = NULL );
BOOL FileExistsT( const wchar_t *lpPathName );
BOOL DirectoryExistsT( const wchar_t *lpPathName );

//
//	FlushFileBuffersT
//		When we just rename file we have no file handle to use original FlushFileBuffers function.
//		But we need the flushing, for save the remanig result to disk.
//
bool FlushFileBuffersT( const wchar_t *path );

//
//	FlushFileBuffersT
//		Uses timeout and extended ODS report.
//
DWORD FlushFileBuffersT( HANDLE hFile, DWORD timeout );

int accessT( const wchar_t *path, int mode );

//
//	Simple translates mbs string.
//
int TranslateT( wchar_t *lpBuffer, int nBufferMax, const char *mbs );
//wchar_t *TranslateOnceT( const char *mbs );

//
//	LoadStringT and LoadStringTA functions
//		Load string with translation, use the SetTranslateDll function before.
//
int
LoadStringT(
    HINSTANCE hInstance,
    UINT uID,
    LPWSTR lpBuffer,
    int nBufferMax,
	bool bTranslate = true,
	bool bCheckParams = true );

int
LoadStringTA(
    HINSTANCE hInstance,
    UINT uID,
    LPSTR lpBuffer,
    int nBufferMax,
	bool translate = true );

#define FindResourceT FindResourceW

//
//	Allocate required space for string and fills if from resources.
//
wchar_t *CreateNewString( HINSTANCE hInst, int ids ); 

#define GetWindowTextLengthT GetWindowTextLengthW
int GetWindowTextT( HWND hWnd, LPWSTR lpString, int nMaxCount );
const wchar_t *GetWindowTextT( cWCStr *wcs );

//
//	It has no difference with usual ImageList_Create in Windows XP and without RTL translation
//		1. It checks if commctrl v.6 and allows ILC_COLOR32 flag, otherwise chenges it to ILC_COLOR24
//		2. It adds mirroring flag for RTL translation.
//
HIMAGELIST ImageList_CreateT( int cx, int cy, UINT flags, int cInitial, int cGrow );
BOOL ImageList_DestroyT( HIMAGELIST himl );

//
//	ImageList_CreateCopyT
//		Creates non mirired copy of existing image list.
//		Returns NULL if error or the existing is empty
//
HIMAGELIST ImageList_CreateCopyT( HIMAGELIST hIml );
int ImageList_ReplaceIconT(HIMAGELIST himl,int i,HICON hicon);
#define ImageList_AddIconT( himl, hicon ) ImageList_ReplaceIconT( himl, -1, hicon )

#define LVI_LAST (0x0fffFFFF)

#define ListView_InsertItemT( hwnd, pitem ) (int)::SendMessageW((hwnd), LVM_INSERTITEMW, 0, (LPARAM)(const LVITEMW*)(pitem))
#define ListView_GetColumnT( hwnd, iCol, pcol ) (BOOL)::SendMessageW((hwnd), LVM_GETCOLUMNW, (WPARAM)iCol, (LPARAM)pcol)
#define ListView_SetColumnT( hwnd, iCol, pcol ) (BOOL)::SendMessageW((hwnd), LVM_SETCOLUMNW, (WPARAM)iCol, (LPARAM)pcol)
#define ListView_InsertColumnT( hwnd, iCol, pcol ) (int)::SendMessageW((hwnd), LVM_INSERTCOLUMNW, (WPARAM)(int)(iCol), (LPARAM)(const LV_COLUMNW *)(pcol))

#define ListView_GetItemT( hwnd, pitem ) (BOOL)::SendMessageW((hwnd), LVM_GETITEMW, 0, (LPARAM)(LV_ITEMW *)(pitem))
#define ListView_SetItemT( hwnd, pitem ) (BOOL)::SendMessageW((hwnd), LVM_SETITEMW, 0, (LPARAM)(const LV_ITEMA FAR*)(pitem))
#define ListView_FindItemT( hwnd, iStart, plvfi ) (int)::SendMessageW((hwnd), LVM_FINDITEMW, (WPARAM)(iStart), (LPARAM)(plvfi) )
int ListView_SetItemTextT( HWND hwndLV, int i, int iSubItem, const wchar_t *pszText );
int ListView_GetItemTextT( HWND hwndLV, int i, int iSubItem, wchar_t *pszText, int cchTextMax );

BOOL ListView_SetItemData( HWND hwnd, int iItem, LPARAM data );
LPARAM ListView_GetItemData( HWND hwnd, int iItem );

#define TreeView_GetItemT( hwnd, pitem ) (BOOL)::SendMessageW((hwnd), TVM_GETITEMW, 0, (LPARAM)(TV_ITEMW *)(pitem))
#define TreeView_SetItemT( hwnd, pitem ) (BOOL)::SendMessageW((hwnd), TVM_SETITEMW, 0, (LPARAM)(const TV_ITEMW *)(pitem))
#define TreeView_InsertItemT( hwnd, lpis ) (HTREEITEM)::SendMessageW((hwnd), TVM_INSERTITEMW, 0, (LPARAM)(LPTV_INSERTSTRUCTW)(lpis))

LPARAM TreeView_GetItemData( HWND hTree, HTREEITEM hItem );
BOOL TreeView_SetItemData(  HWND hTree, HTREEITEM hItem, DWORD data, int iImage = -1, int iSelectedImage = -1 );
BOOL TreeView_GetItemTextT( HWND hTree, HTREEITEM hItem, wchar_t *text, int textMaxLen );
BOOL TreeView_SetItemTextT( HWND hTree, HTREEITEM hItem, const wchar_t *text );

int SendStringMessage( HWND hwndCtl, UINT uMsg, WPARAM wParam, const wchar_t *lpsz );
int SendStringMessage( HWND hwndCtl, UINT uMsgA, UINT uMsgW, WPARAM wParam, const wchar_t *lpsz );
int SendStringMessage_Get( HWND hwndCtl, UINT uMsg, WPARAM wParam, wchar_t *lpsz, int size = -1);

int ComboBoxEx_GetItemTextT		( HWND hCmb, int i, wchar_t *wcs, int sizeInChar );
int ComboBoxEx_GetTextT			( HWND hCmb, wchar_t *wcs, int sizeInChar );
int ComboBoxEx_InsertStringT	( HWND hCmb, int i, const wchar_t *wcs, int indent = 0 );
int ComboBoxEx_SetItemTextT		( HWND hCmb, int i, const wchar_t *wcs, int indent = 0 );

#define ComboBox_AddStringA( hCmb, mbs )		( SendMessageA( hCmb, CB_ADDSTRING,		  0, (LPARAM)mbs ))
#define ComboBox_InsertStringA( hCmb, i, mbs )	( SendMessageA( hCmb, CB_INSERTSTRING,	(i), (LPARAM)mbs ))

#define ComboBox_AddStringT( hCmb, wcs )		(SendStringMessage( hCmb, CB_ADDSTRING,		  0, wcs ))
#define ComboBox_InsertStringT( hCmb, i, wcs)	(SendStringMessage( hCmb, CB_INSERTSTRING,	(i), wcs ))
#define ComboBox_FindStringT( hCmb, i, wcs )	(SendStringMessage( hCmb, CB_FINDSTRING,	(i), wcs ))
#define ListBox_AddStringT( hLB,	wcs )		(SendStringMessage( hLB,  LB_ADDSTRING,		  0, wcs ))
#define ListBox_InsertStringT( hLB,  i,wcs )	(SendStringMessage( hLB,  LB_INSERTSTRING,	(i), wcs ))
#define ListBox_FindStringT( hLB, i, wcs )		(SendStringMessage( hLB,  LB_FINDSTRING,	(i), wcs ))

#define ListBox_GetTextT( hLB, i, wcs, size )	( SendStringMessage_Get( hLB, LB_GETTEXT,	(i), wcs, size ))
#define ComboBox_GetLBTextT( hCmb, i, wcs, size )	( SendStringMessage_Get( hCmb, CB_GETLBTEXT, (i), wcs, size ))

#define ModifyMenuT ModifyMenuW
#define InsertMenuItemT InsertMenuItemW
#define SetMenuItemInfoT SetMenuItemInfoW

//
//	GetMenuStringW
//	works properly for NT 4, instead of GetMenuTextT
//
#define GetMenuStringT GetMenuStringW
#define GetMenuItemInfoT GetMenuItemInfoW

BOOL
GetMenuTextT(
		IN HMENU hMenu,
		IN UINT uPosition,
		bool byPosition,
		wchar_t *text,
		int textMaxLen
	);

ULONG_PTR
	GetMenuItemParam(
		IN HMENU hMenu,
		IN UINT uPosition,
		bool byPosition
	);

BOOL
SetMenuTextT(
    IN HMENU hMenu,
    IN UINT uPosition,
	bool byPosition,
	const wchar_t *text 
	);

//
//	Returns text length in pixels
//
int GetTextSizeT( HWND hWnd, wchar_t *ws );

//
//	PrintClientToDC
//		Sends WM_PRINTCLIENT message to window.
//		The message draws window client aria to position 0,0 of device content.
//		But we need to draw on specified position, so:
//			1. new srcHDC is created
//			2. windowd client region is printed here.
//			3. BitBlt is used to copy srcHDC to dstHDC.
//
bool PrintClientToDC( HDC dstHDC, int dstX, int dstY, HWND srcHWnd, int srcX = 0, int srcY = 0, int cx = -1, int cy = -1 );

#define GetTextExtentPoint32T GetTextExtentPoint32W
#define DrawTextT DrawTextW

BOOL ExtTextOutT( 
		HDC hdc,
		int X, int Y,
		UINT fuOptions,
		CONST RECT* lprc,
		LPCWSTR lpString,
		UINT cbCount,
		CONST INT* lpDx
	);


LONG_PTR SetWindowLongPtrT( HWND hWnd, int nIndex, LONG_PTR dwNewLong );
LONG_PTR GetWindowLongPtrT( HWND hWnd, int nIndex );
ULONG_PTR SetClassLongPtrT( HWND hWnd, int nIndex, LONG_PTR dwNewLong );

#define CallWindowProcT CallWindowProcW
#define GetWindowsDirectoryT GetWindowsDirectoryW
#define GetSystemDirectoryT GetSystemDirectoryW

#ifndef _NoShell32Lib_

#define FindExecutableT FindExecutableW

//
//	ShellSelect
//		Opens a window view with 'path' selected.
//	Works for XP or later.
//	
bool ShellSelect( const wchar_t *path );


//
//	ShellExecuteMailTo
//		Send mail by ShellExecute with "mailto:%s?subject=%s&body=%s" pattern
//	Note, cmd line length is limited by 1000 symbols.
//	If message body size is more the MAPISendMail function has to be used instead,
//	but thatre are problems:
//		1. MSDN reports:
//		'The use of Simple MAPI is discouraged. It may be altered or unavailable in subsequent versions of Windows.'
//		2. Application verifier reports Invalid TSL Index
//
HINSTANCE
ShellExecuteMailTo( HWND hWnd, const wchar_t *szAddress, const wchar_t *szSubject, const wchar_t *szNote, const wchar_t *attacheFile = NULL );

#define ShellExecuteT ShellExecuteW
#define ShellExecuteExT ShellExecuteExW
#define DragQueryFileT DragQueryFileW


bool BrowseForFolderT( HINSTANCE hInst, int idsTitle, HWND hWnd, wchar_t *dir, UINT ulFlags = BIF_RETURNONLYFSDIRS, int clsIdRoot = CSIDL_DRIVES );
bool BrowseForFolderT( const wchar_t *szTitle, HWND hWnd, wchar_t *dir, UINT ulFlags = BIF_RETURNONLYFSDIRS, int clsIdRoot = CSIDL_DRIVES );

#define SHBrowseForFolderT SHBrowseForFolderW
#define SHGetPathFromIDListT SHGetPathFromIDListW
bool SHGetFileInfoT( SHFILEINFOW &sfi, DWORD getFileInfo = SHGFI_DISPLAYNAME | SHGFI_ICON | SHGFI_SMALLICON, DWORD clsid = CSIDL_DRIVES );
DWORD* SHGetFileInfoT(LPCWSTR pszPath, DWORD dwFileAttributes, SHFILEINFOW *psfi, UINT cbFileInfo, UINT uFlags, bool bCoInitialize = false );
#define SHGetSpecialFolderPathT SHGetSpecialFolderPathW
#define Shell_NotifyIconT Shell_NotifyIconW

#endif // _NoShell32Lib_

//
//	GetTempPathT
//		Returns user temporary folder path.
//		Note, for impersonated thread it returns path for original token.
//		F.e. if original process is 'local system' it returns 'C:\Windows\Temp'.
//		Use the GetTempPathIH at GetTempPathIH.cpp instead.
//
DWORD GetTempPathT(
    DWORD nBufferLength,
    LPWSTR lpBuffer,
	bool doNotCreate = false // create the temp folder if it does not exist
    );

DWORD GetTempPathT( cWCStr *path, bool doNotCreate = false );

DWORD
GetShortPathNameT(
    LPCWSTR lpszLongPath,
    LPWSTR  lpszShortPath,
    DWORD    cchBuffer
    );

DWORD GetShortPathNameT( LPCWSTR lpszLongPath, cWCStr *wcs );

#define GetLongPathNameT GetLongPathNameW
DWORD GetLongPathNameT( LPCWSTR lpszShortPath, cWCStr *path );

#define GetDiskFreeSpaceT GetDiskFreeSpaceW

BOOL
GetDiskFreeSpaceExT(
    IN LPCWSTR lpDirectoryName,
    OUT PULARGE_INTEGER lpFreeBytesAvailableToCaller,
    OUT PULARGE_INTEGER lpTotalNumberOfBytes = NULL,
    OUT PULARGE_INTEGER lpTotalNumberOfFreeBytes = NULL
    );

#define WinHelpT WinHelpW

typedef struct tagHH_POPUPW
{
	 int         cbStruct;
	 HINSTANCE   hinst;
	 UINT        idString;
	 wchar_t*    pszText;
	 POINT       pt;
	 COLORREF    clrForeground;
	 COLORREF    clrBackground;
	 RECT        rcMargins;
	 wchar_t*    pszFont;
} HH_POPUPW;

#define GetModuleFileNameT GetModuleFileNameW
#define MessageBoxIndirectT MessageBoxIndirectW

bool GetModuleFileNeighbour( cWCS &wcs, const wchar_t *fileName );

int
MessageBoxT(
    IN HWND hWnd,
    IN LPCWSTR lpText,
    IN LPCWSTR lpCaption,
    IN UINT uType);

//
//	splitPath
//		fills dir ( dir does not include last '\' ) and file buffers if they are not null,
//		returns position number of file name in path
//
int splitPath( const char *path, char *dir = NULL, char *file = NULL );
int splitPath( const wchar_t *path, wchar_t *dir = NULL, wchar_t *file = NULL );
int splitPath( const wchar_t *path, cWCS *dir, cWCS *file = NULL );
void makePath( wchar_t *path, const wchar_t *drive, const wchar_t *dir = NULL, const wchar_t *file = NULL, const wchar_t *ext = NULL );

//
//	isFileStream
//
//		Returns true if the path name contains stream name
//
//		Sample:
//			'QQQ.pdf:Zone.Identifier:$DATA'
//		stream pos -^
//
bool isFileStream( const wchar_t *path, int *pStreamPos = NULL, wchar_t *filePath = NULL, wchar_t *streamName = NULL );

#define wcscpy_S( d, s ) wcscpy_s( (d), maxLen(d), (s)  )

bool isDigital( wchar_t wc, DWORD *pVal = NULL );

#define SetCurrentDirectoryT SetCurrentDirectoryW
#define GetCurrentDirectoryT GetCurrentDirectoryW

DWORD GetCurrentDirectoryT( cWCStr *buffer );

#define GetDriveTypeT GetDriveTypeW

BOOL
GetVolumeInformationT(
    IN LPCWSTR lpRootPathName,
    OUT LPWSTR lpVolumeNameBuffer,
    IN DWORD nVolumeNameSize,
    OUT LPDWORD lpVolumeSerialNumber,
    OUT LPDWORD lpMaximumComponentLength,
    OUT LPDWORD lpFileSystemFlags,
    OUT LPWSTR lpFileSystemNameBuffer,
    IN DWORD nFileSystemNameSize,
	bool silent = false // To prevent the system from displaying message box if floppy or CD drive does not have a disk
    );

UINT
GetTempFileNameT(
    IN LPCWSTR lpPathName,
    IN LPCWSTR lpPrefixString,
    IN UINT uUnique,
    OUT LPWSTR lpTempFileName,
	DWORD *pErr = NULL
  );

bool GetTempFilePathT( wchar_t *buff, wchar_t *prefix, bool useShortName = true );

//
//	CreateTempFileName
//		Works as GetTempFileName but allows to create handle to the temporary file.
//		GetTempFileName is insecure by 'Common weakness enumeration. CWE-377: Insecure Temporary File'.
//
UINT CreateTempFileName(
    IN LPCWSTR lpDirectory,
    IN LPCWSTR lpPrefixString,
    IN UINT uUnique,
    OUT LPWSTR lpTempFileName,
	HANDLE *pHFile,
	DWORD dwShareMode = FILE_SHARE_READ,
	DWORD fileAtributes = FILE_ATTRIBUTE_NORMAL,
	DWORD *pErr = NULL,
	IN LPCWSTR lpPostfixString = L".tmp",
	IN LPCWSTR lpTemplateSrc = NULL // is used for directory while.
 );

//
// CreateTempFilePath
//	Uses CreateTempFileName instead of GetTempFileName
//	GetTempFileName is insecure by 'Common weakness enumeration. CWE-377: Insecure Temporary File'.
//
bool CreateTempFilePath(
		wchar_t *buff,
		wchar_t *prefix,
		bool useShortName,
		HANDLE *phFile,
		DWORD dwShareMode = FILE_SHARE_READ
   );

bool isSparseFile( const wchar_t *filePath );

#define RegCreateKeyExT RegCreateKeyExW
#define RegCreateKeyT RegCreateKeyW
#define RegOpenKeyExT RegOpenKeyExW
#define RegEnumKeyExT RegEnumKeyExW
#define RegEnumValueT RegEnumValueW
#define RegSetValueExT RegSetValueExW
#define RegQueryInfoKeyT RegQueryInfoKeyW

LONG RegGetValueT( HKEY hkey, LPCWSTR lpSubKey, LPCWSTR  lpValue, DWORD dwFlags, LPDWORD pdwType, PVOID   pvData, LPDWORD pcbData );
LONG RegSetValueT( HKEY root, const wchar_t *szSection, const wchar_t *szValue, DWORD dwType, void *bf, int bfSize );
LSTATUS RegDeleteKeyValueT( HKEY hKey, LPCWSTR lpSubKey, LPCWSTR lpValueName );

//
//	RegQueryValueExT
//		In windows 95 environment converts ANSI buffer to WCHAR if REG_EXPAND_SZ, REG_SZ or REG_MULTI_SZ
//
LONG
RegQueryValueExT (
    IN HKEY hKey,
    IN LPCWSTR lpValueName,
    IN LPDWORD lpReserved,
    OUT LPDWORD lpType,
    IN OUT LPBYTE lpData,
    IN OUT LPDWORD lpcbData
    );


//
//	RegDeleteSubKeysAndValuesT
//	It does not delete main key, just subkey and values
//
LONG RegDeleteSubKeysAndValuesT(
	HKEY hRoot,
	LPCWSTR szKeyName
	);

LONG
RegDeleteEntireKeyT(
    IN HKEY hKey,
    IN LPCWSTR lpSubKey
    );

#define RegDeleteKeyT RegDeleteKeyW
#define RegDeleteValueT RegDeleteValueW


//
//	GetTimeText
//		Returns rounded string like '3 Hours 12 Mins'
//									 |value   |rest
//		If main value more then 4 the rest is not shown
//
//	Note, Use SetTimeTextValues for valid translation.
//	The localeNumber allows alternative numerous type
//	f.e. 0x419 Russian: 0 chasov, 1 chas, 2 chasa, 5 chasov
//	Use SetTimeTextValues( true, ... , 0x419 ) - for alternative Russian numerous time like 2 chasa 3 minuty 4 secundy.
//
void GetTimeText( char *buff, int buffSize, DWORD timeInSec, int digits = 2 );
void GetTimeTextT( wchar_t *buff, int buffSize, time_t timeInSec, int digits = 2 );
const wchar_t *GetTimeText( time_t timeInSec , int digits = 2);
const wchar_t *GetIntText( int i, int radix );

enum eTimeType { timeSec = 0, timeMin, timeHour, timeDay, timeWeek, timeLast = timeWeek };
//	int idsT[] = { IDS_sec, IDS_min, IDS_hour, IDS_day, IDS_week } 
enum eNumerousType { entNone = 0, entUsual, entUnusual }; // see getNumerousType()

// The time translation works automatically -> See !TranslationStrategy.txt 
//void SetTimeTextValues( bool numerous, eTimeType timeType, HINSTANCE hInst, int idsValue, int localeNumber = 0 );
//void SetTimeTextValues( bool numerous, eTimeType timeType, wchar_t *szValue, int localeNumber = 0 );
//void SetTimeTextValues( bool numerous, HINSTANCE hInst, int *idsT, int localeNumber = 0 );

int StrFromTimeIntervalT( LPWSTR pszOut, UINT cchMax, DWORD dwTimeMS, int digits = 2 );

char *GetSizeText( char *buff, int buffSize, DWORDLONG dwlSize, bool sizeinBytes = false );
char *GetSizeText( char *buff, int buffSize, DWORD sizeLow, DWORD sizeHigh );
wchar_t *GetSizeTextT( wchar_t *buff, int buffSize, DWORD sizeLow, DWORD sizeHigh );
wchar_t *GetSizeTextT( wchar_t *buff, int buffSize, DWORDLONG size, bool sizeinBytes = false );
wchar_t *GetSizeTextOnceT( DWORDLONG size );

//
//	GetSizeFromText
//		Parses string like '512.5 KB' to DWORDLONG 512 * 1024 + 256
//		Allows probels between figures and dimension -> 1 512 MB
//		Return -1 if text was not properly parsed.
//
DWORDLONG GetSizeFromText( const wchar_t *text, DWORDLONG *pDimention = NULL );

//
//	getFileInfoT
//		Gets file modification date-time ( DD:MM:YY HH:MM ) and file size ( f.e. 27.4 KB ) strings
//
wchar_t *getFileInfoT(
		const wchar_t *fileName,
		wchar_t *dateStr, int dateStrSize,
		wchar_t *sizeStr, int sizeStrSize
	);

char *getFileInfo(
		const char *fileName,
		char *dateStr, int dateStrSize,
		char *sizeStr, int sizeStrSize
	);

//int FormatString( wchar_t *str, const wchar_t *pattern, const wchar_t *val );
//int FormatString( wchar_t *str, HINSTANCE hInst, int ids, const wchar_t *val );
//
//int FormatString2( wchar_t *str, const wchar_t *pattern, const wchar_t *val1, const wchar_t *val2 );
//int FormatString2( wchar_t *str, HINSTANCE hInst, int ids, const wchar_t *val1, const wchar_t *val2 );

//
//	GetVolumeName
//		Gets volume name for corresponded mount point 
//
bool GetVolumeName( wchar_t *volumeName, int length, const wchar_t *mountPoint, DWORD *pErr = NULL );

#ifndef _NoElevation
extern bool runElevated(
		HWND hwnd,
		const wchar_t *pszPath,
		const wchar_t *pszParameters = NULL,
		const wchar_t *pszDirectory = NULL,
		HANDLE *pHProcess = NULL,
		int *pError = NULL
	);

#endif // #ifdef _NoElevation

extern bool userAccountControlIsEnabled( bool bForced = false );

extern bool WaitProcessAsModalDialogue( HANDLE hProcess, DWORD processID = 0 );
extern HWND FindWindowByProcessId( DWORD id );

enum eWaitProcessType { ewptNoWait = 0, ewptWait = 1, ewptWaitAsModalDialogue = 2 };

//
//	Returns:
//		0 if process was not started.
//		-1 if error was encountered.
//		process handle if successfully starded
//	Note, if the procedure does not wait started process so the process handle has to be closed by called process
//	29.06.2011
//		The bRunElevated flag is renamed to bRunAs. It checks adminitrator privileges as well as limited user.
//		As result it asks user name and password for non-admintrative user.
//		Limited user is elevated ( as earlier )
//
HANDLE RunProcessWithWaitingT(
	const wchar_t *cmdLine,
	int waitType = ewptWait,
	bool hidden = false,
	const wchar_t *curDir = NULL,
	int *pRetCode = NULL,
	bool bRunAs = false, // If true it asks elevation or administraotr password for non-privileged user.
	HWND hWnd = NULL,
	HANDLE hToken = NULL, // the token for thread to be impersonated
	HANDLE *pHPrimaryTread = NULL,	// If we uses CREATE_SUSPENDED dwFlags for next value so it suspends primary thread
	DWORD dwFlags = 0,				// The *pHPrimaryTread is required for resuming.
	DWORD *pProcessID = NULL,
	DWORD *pThreadId = NULL,
	HANDLE hOutFile = NULL,
	HANDLE hInFile = NULL
);

//
//	GetProcessOutput
//		Calls RunProcessWithWaitingT, Note argumments are completely same excluding the first 'out'
//		Before calling creates an anonymous pipe and pass writting handle to RunProcessWithWaitingT as outFile
//		After process completes reads from the pipe by ReadToADBlock
//
HANDLE GetProcessOutput(
	cADBlock < BYTE > &out,
	const wchar_t *cmdLine,
	int waitType = ewptWait,
	bool hidden = false,
	const wchar_t *curDir = NULL,
	int *pRetCode = NULL,
	bool bRunAs = false, // If true it asks elevation or administraotr password for non-privileged user.
	HWND hWnd = NULL,
	HANDLE hToken = NULL, // the token for thread to be impersonated
	HANDLE *pHPrimaryTread = NULL,	// If we uses CREATE_SUSPENDED dwFlags for next value so it suspends primary thread
	DWORD dwFlags = 0,				// The *pHPrimaryTread is required for resuming.
	DWORD *pProcessID = NULL,
	DWORD *pThreadId = NULL
);

//
//	ReadFromFile
//		Reads bytes from opened file.
//
bool ReadFromFile( HANDLE hFile, cADBlock < BYTE > &adb );

//
//	15.04.2013
//	runAsLimitedUser
//	Uses scheduled task to run process as limited user.
//	If command line path ( cmdPath ) has no probels it may be send as one string with argumments ( cmdArgs ), the previose variant uses one string.
//	Valid scheduled task format is:  "\" p a t h \" a command arrgumments" 
//	Note, cmdArgs and taskName may be NULL, but were not assigned by default to rise compilation error for old using of the procedure.
//
//bool runAsLimitedUser( const wchar_t *cmdLine, const wchar_t *taskName = NULL, DWORD timeout = 1000, DWORD *pExistCode = NULL );
bool runAsLimitedUser( const wchar_t *cmdPath, const wchar_t *cmdArgs, const wchar_t *taskName, DWORD timeout = 1000, DWORD *pExistCode = NULL );

#define CreateProcessT CreateProcessW

#define _isWinNT ( GetVersion() < 0x80000000 )

class cInstanceHolder
{
public:

	static HINSTANCE m_hInst;
	static HWND m_hMainWnd;
	static HWND getMainWnd() { return IsWindow( m_hMainWnd ) ? m_hMainWnd : (::GetDesktopWindow() ); }
	static cWCStr &getStaticCaption()
	{
		static cWCStr wcs;
		return wcs;
	}

	static const wchar_t *getCaption( const wchar_t *pCaption = NULL )
	{
		if ( pCaption != NULL )
			return pCaption;

		if ( !getStaticCaption().isEmpty() )
			return getStaticCaption().m_szBody;
		
		return NULL;
	}

	operator HINSTANCE () { return m_hInst; }
	operator HWND () { return getMainWnd(); }
	
	cInstanceHolder( HINSTANCE hInst = NULL )
	{
		SetHolder( hInst, NULL );
		m_hSaveCursor = NULL;
	}

	static void SetHolder( HINSTANCE hInst, HWND hWnd = NULL );

	static int ErrorMessage( int idsText, int idsCaption, int mbID = MB_ICONSTOP )
	{
		return ErrorMessage( L"", idsCaption, idsText, mbID );
	}

	static int ErrorMessage( const wchar_t *text, int idsCaption, int idsTextPattern = 0, int mbID = MB_ICONSTOP )
	{
		return MessageBoxT(
					getMainWnd(),
					idsTextPattern == 0 ? 
						text :
						_FormatT( cWCStr().LoadStringT( m_hInst, idsTextPattern ), text ),
					cWCStr().LoadStringT( m_hInst, idsCaption),
					mbID );
	}

	static int cxIcon, cyIcon;
	static HICON LoadIconT( HINSTANCE hInst, int idi, int cx = 0, int cy = 0 );
	static HICON LoadIconT( int idi, int cx = 0, int cy = 0 );

	void WaitLoop( DWORD mSec );

	HCURSOR m_hSaveCursor;
	void BeginCursor( ULONG_PTR idc = (ULONG_PTR)IDC_WAIT );
	void EndCursor();

	static const wchar_t *GetOSVersionStr();

	static bool isWindowsNT() { return _isWinNT; }
	static bool isW7();
	static bool isW8( bool bOrLater = true );
	static bool isW10( bool bOrLater = true );
	static bool isWindowsVista();
	static bool isWindowsXP();
	static bool isWindows2K();
	static bool isWindowsME();
	static bool isWindows98();
	static bool isHomeVersion();

	//
	//	IsWow64Process
	//		returns true if 32-bit application is runned in 64-bit environment.
	//		returns false for 64-bit application or 32-bit windows.
	//
	static bool IsWow64Process();
	static inline bool IsRunningOnWin32()
	{
		return !IsWow64Process();
	}

};

//
//	Useful main instance holder
//		1. It's body has to be declared before implemetation
//			cInstanceHolder IH;
//		2. It has to be inated by SetHolder function
//			IH.SetHolder( hMainInst );
//

extern cInstanceHolder IH;
#define _LoadIDS( __ids ) cWCStr().LoadStringT( IH.m_hInst, __ids )
#define _LoadIDST( __ids, __translate ) ( cWCStr( IH.m_hInst, __ids, __translate ).m_szBody )
#define _LoadIDSNT( __ids ) ( cWCStr( IH.m_hInst, __ids, false ).m_szBody )
#define _GetSizeText( __size ) ( cWCStr().GetSizeText( ( __size) ) )
#define _GetSystemError( __err ) ( cWCStr().GetSystemError( __err ) )
#define _GetLastSystemError() ( cWCStr().GetSystemError( -1 ) )

extern int DlgMessageBox( HINSTANCE hInst, HWND hDlg, int IDS, int MB_ID = MB_OK, const char *pCaption = NULL );
extern int DlgMessageBoxT( HINSTANCE hInst, HWND hDlg, int IDS, int MB_ID = MB_OK, const wchar_t *pCaption = NULL );
extern int DlgMessageBoxT( HWND hDlg, const wchar_t *text, int MB_ID = MB_OK, const wchar_t *pCaption = NULL );
extern int DlgMessageBox( HWND hDlg, const char *text, int MB_ID = MB_OK, const char *pCaption = NULL );
extern const wchar_t *getParentCaption( HWND hWnd );

#define SetDlgItemTextT SetDlgItemTextW
extern int GetDlgItemTextT( HWND hDlg, int nIDDlgItem, wchar_t *lpString, int nMaxCount );
extern bool DlgCheckButton( HWND hDlg, int idBtn, bool bCheck );

class cDlgFuncsHolder : public cInstanceHolder
{
	//cWCStr m_wcsDefCaption;
public:
	HWND m_hDlg;
	const wchar_t *m_defCaption;

	cDlgFuncsHolder()
	{
		m_hDlg = NULL;
		m_defCaption = L"";
		//m_wcsDefCaption.Set( cInstanceHolder::m_defCaption );
		//m_defCaption = m_wcsDefCaption.m_szBody;
	}

	//
	// SetHolder
	//	1. Calls SetHolder for cInstanceHolder parent class
	//	2. Translates the dialog if required
	//	2. Sets default caption
	//
	void SetHolder( HINSTANCE hInst, HWND hDlg, int idsDefCaption = 0 );

	const wchar_t *getCaption( const wchar_t *pCaption )
	{
		if ( pCaption != NULL )
			return pCaption;

		if ( m_defCaption[0] != L'\0' )
			return m_defCaption;

		//m_wcsDefCaption.GetWindowTextT( m_hDlg );
		//if ( !m_wcsDefCaption.isEmpty() )
		//	return m_wcsDefCaption.m_szBody;

		pCaption = getParentCaption();
		if ( pCaption != NULL )
			return pCaption;

		return cInstanceHolder::getCaption();
	}

	const wchar_t *getParentCaption()
	{
		return ::getParentCaption( m_hDlg );
	}

	int DlgMessageBox( int IDS, int MB_ID = MB_OK )
	{
		return ::DlgMessageBoxT( m_hInst, m_hDlg, IDS, MB_ID, getCaption( NULL ) );
	}

	int DlgMessageBoxT( int IDS, int MB_ID = MB_OK, const wchar_t *pCaption = NULL )
	{
		return ::DlgMessageBoxT( m_hInst, m_hDlg, IDS, MB_ID, getCaption( pCaption ) );
	}
	int DlgMessageBoxT( const wchar_t *text, int MB_ID = MB_OK, const wchar_t *pCaption = NULL )
	{
		return ::DlgMessageBoxT( m_hDlg, text, MB_ID, getCaption( pCaption ) );
	}

	HWND GetItem( int nIDDlgItem ) { return ::GetDlgItem( m_hDlg, nIDDlgItem ); }

	bool IsChecked( int nIDDlgItem );

	BOOL SetDlgItemTextT( int nIDDlgItem, const wchar_t *lpString )
	{
		return ::SetDlgItemTextT( m_hDlg, nIDDlgItem, lpString );
	}

	BOOL SetDlgItemTextT( int nIDDlgItem, int ids )
	{	return ::SetDlgItemTextT( m_hDlg, nIDDlgItem, cWCStr().LoadStringT( m_hInst, ids ) ); }

	int GetDlgItemTextT( int nIDDlgItem, wchar_t *lpString, int nMaxCount )
	{	return ::GetDlgItemTextT( m_hDlg, nIDDlgItem, lpString, nMaxCount ); }

	//
	//	Checks if corresponded dlg item text contains the '%s' symbol
	//	if so it forms new string as pattern for txt, otherwise
	//	overwrites entire the item.
	//
	static void UpdateDlgItemTextT( HWND hDlg, int idText, const wchar_t *txt );
	static void UpdateDlgItemText( HWND hDlg, int idText, const char *txt );
};

int findCharPos( const wchar_t *str, wchar_t c, int len = -1 );

#define SendSettingChangeMessageT( hWnd, wParam, section ) ::SendMessageW( hWnd, WM_SETTINGCHANGE, wParam, (LPARAM)section )
#define LoadLibraryExT LoadLibraryExW

#define LoadLibraryT( lpLibFileName ) LoadLibraryExT( lpLibFileName, NULL, 0 );

#define GetModuleHandleT GetModuleHandleW

#ifndef _NoComdlg32Lib_
BOOL GetOpenFileNameT(LPOPENFILENAMEW);
BOOL GetSaveFileNameT(LPOPENFILENAMEW);
#endif // _NoComdlg32Lib_


//
//	GetCommandLineT
//	Returns constant wide char string of command line.
//	It returns WinMain compatible command line if excludeProgramName is true
//	Otherwise returns the command line without changes as the original GetCommandLine function returns.
//
const wchar_t *GetCommandLineT( bool excludeProgramName = false );

// nameFormat 12 is NameDnsDomain
bool GetUserNameExT( int nameFormat, LPWSTR lpNameBuffer, PULONG nSize );

#define GetComputerNameT GetComputerNameW
#define GetEnvironmentVariableT GetEnvironmentVariableW
#define RegisterClassExT RegisterClassExW
#define RegisterClassT RegisterClassW
#define FindWindowExT FindWindowExW
#define FindWindowT FindWindowW
#define CreateWindowExT CreateWindowExW
#define DefWindowProcT DefWindowProcW
#define GetMessageT GetMessageW
#define DispatchMessageT DispatchMessageW
#define TranslateAcceleratorT TranslateAcceleratorW
#define TextOutT TextOutW

//
//	GetDateFormatT
//
//	Sample:
//		SYSTEMTIME sysTime;
//		GetSystemTime( &sysTime );
//
//		int len = GetDateFormatT(
//						LOCALE_USER_DEFAULT,
//						DATE_LONGDATE, // DATE_YEARMONTH, // DATE_SHORTDATE, //LOCALE_NOUSEROVERRIDE
//						&sysTime,
//						NULL,
//						buff, buffLen
//						);
//

#define GetDateFormatT GetDateFormatW
#define GetTimeFormatT GetTimeFormatW

wchar_t *GetFileLocalTime( wchar_t *bf, int bfLen, const FILETIME *fileTime );
wchar_t *GetFileLocalTime( wchar_t *bf, int bfLen, const wchar_t *filePath );
wchar_t *GetSystemTimeStr( wchar_t *bf, int bfLen, const SYSTEMTIME *pSysTime = NULL );
wchar_t *getSystemTimeStr( wchar_t *str = NULL, int strMaxLen = 0, SYSTEMTIME *systime = NULL );
const wchar_t *GetSystemTimeStr( const SYSTEMTIME *pSysTime = NULL );

INT_PTR PropertySheetT( PROPSHEETHEADERW *lppsph );
int PropSheet_SetHeaderTitleT( HWND hDlg, int index, const wchar_t *lpwszText );

HPROPSHEETPAGE CreatePropertySheetPageT( LPCPROPSHEETPAGEW constPropSheetPagePointer );
BOOL DestroyPropertySheetPageT( HPROPSHEETPAGE hpsp );

//
//	PropSheet_Translate
//		It translates TabControl heeadrs too.
//	All is transtated automatically by PropertySheetT and langfile.dll
//
//int PropSheet_Translate( HWND hwndDlg );

#define TabCtrl_InsertItemT( hwnd, iItem, pitem ) (int)::SendMessageW((hwnd), TCM_INSERTITEMW, (WPARAM)(int)(iItem), (LPARAM)(const TC_ITEMW *)(pitem))
#define GetPrivateProfileStringT GetPrivateProfileStringW

BOOL
GetUserNameT (
    OUT LPWSTR lpBuffer,
    IN OUT LPDWORD nSize
    );

BOOL
GetUserNameTA (
    OUT LPSTR lpBuffer,
    IN OUT LPDWORD nSize
    );

#define GetLogicalDriveStringsT GetLogicalDriveStringsW
#define GetCompressedFileSizeT GetCompressedFileSizeW

//
//	Allocates and copies the string to alocated buffer.
//
wchar_t *allocateWCS( const wchar_t *bf, int len = -1 );
char *allocateMBS( const char *bf, int len = -1 );

//
//	Allocates and converts the string to alocated buffer.
//
wchar_t *allocateWCS( const char *bf, int len = -1 );
char *allocateMBS( const wchar_t *bf, int len = -1 );

#define SendMessageT( hWnd, MsgW, MsgA, wParam, lParam ) ::SendMessageW( hWnd, MsgW, (WPARAM)wParam, (LPARAM)lParam )

HWND CreateStatusWindowT(LONG style, LPCWSTR lpszText, HWND hwndParent, UINT wID);

#define ExpandEnvironmentStringsT ExpandEnvironmentStringsW

DWORD
QueryDosDeviceT(
		LPCWSTR lpDeviceName,
		LPWSTR lpTargetPath,
		DWORD ucchMax
	);

HANDLE
FindFirstVolumeT(
    LPWSTR lpszVolumeName,
    DWORD cchBufferLength
    );
BOOL
FindNextVolumeT(
    HANDLE hFindVolume,
    LPWSTR lpszVolumeName,
    DWORD cchBufferLength
    );
BOOL
FindVolumeCloseT(
    HANDLE hFindVolume
    );
HANDLE
FindFirstVolumeMountPointT(
    LPCWSTR lpszRootPathName,
    LPWSTR lpszVolumeMountPoint,
    DWORD cchBufferLength
    );
BOOL
FindNextVolumeMountPointT(
    HANDLE hFindVolumeMountPoint,
    LPWSTR lpszVolumeMountPoint,
    DWORD cchBufferLength
    );
BOOL
FindVolumeMountPointCloseT(
    HANDLE hFindVolumeMountPoint
    );

BOOL
GetVolumePathNameT(
    LPCWSTR lpszFileName,
    LPWSTR lpszVolumePathName,
    DWORD cchBufferLength
    );

BOOL GetVolumeNameForVolumeMountPointT(
	LPCWSTR lpszVolumeMountPoint,
	LPWSTR lpszVolumeName,
	DWORD cchBufferLength
);

BOOL
GetVolumePathNamesForVolumeNameT(
	LPCWSTR lpszVolumeName,
	LPWSTR lpszVolumePathNames,
	DWORD cchBufferLength,
	PDWORD lpcchReturnLength
);

BOOL
FindNextFileNameT(
	HANDLE hFindStream,
	LPDWORD StringLength,
	PWCHAR LinkName
);

HANDLE
FindFirstFileNameT(
	LPCWSTR lpFileName,
	DWORD dwFlags,
	LPDWORD StringLength,
	PWCHAR LinkName
);


#define isFloppy( _name ) ( (towupper( (_name)[0] == L'A' ) || (towupper( (_name)[0] == L'B' ) ) ) )

BOOL
W2K_GetVolumePathNamesForVolumeNameT(
    LPCWSTR lpszVolumeName,
    LPWSTR lpszVolumePathNames,
    DWORD cchBufferLength,
	PDWORD lpcchReturnLength = NULL
);

HICON ExtractIconT(HINSTANCE hInst, LPCWSTR lpszExeFileName, UINT nIconIndex);
UINT ExtractIconExT( LPCWSTR lpszFile, int nIconIndex, HICON *phiconLarge, HICON *phiconSmall, UINT nIcons );

#define Toolbar_InsertButtonT( hToolbar, iButton, tbb ) ( ::SendMessageW( hToolbar, TB_INSERTBUTTONW, (WPARAM)iButton, (LPARAM)tbb ) != 0 )

//
//	GetClient
//		fills ouput rectangle by coordinates relatively to parent window,
//		both input rectangles are in screen coordinates.
//
//void GetClient( RECT *r, const RECT *scrR, const RECT *parentR );
void GetClient( RECT *r, HWND hWnd, HWND hParent = NULL );

UINT getDefaultCodePage();
bool setProcessDefaultLayout( bool rtlLayout );
bool getProcessDefaultLayout( DWORD *pLayout );
bool isRTLProcessLayout();


BOOL GetProgramFilesDirT( wchar_t *maxPathBuff, bool bX64 = false );

enum { eRCStringNotFound = 0, eMemoryAllocation, eRCInstanceInvalid };
wchar_t *getErrorString( int errNum, wchar_t *bf = NULL, int bfMaxLen = 0, DWORD value = 0 );

#include "CSysDll.h"

class cComCtl32Dll : public CSysDll
{
public:
	// Do not use the in global constructor, and avoid global contruction at all.
	static __declspec(noinline) cComCtl32Dll &getStatic()
	{
		static cSysCriticalSection cs;
		SYS_CRITICAL_SECTION_BLOCK( &cs );

		static cComCtl32Dll dll;
		return dll;
	}
	
	cComCtl32Dll()
	{
		LoadOnce();
	}

	BOOL LoadOnce()
	{
		if ( hInst != NULL )
			return true;

		// Note 1.
		// Win 64 the CreatePropertySheetPage function does not work properly.
		// More over it does not load comctl32.dll without path
		// So we specify here full path.
		// Problem appeares at Vista 64 environment, bcshext.dll(64) did not create page for container properties.

		// Note 2.
		// Another problem: in WOW64 for 32 version the directly loaded c:\windows\sysyem32\comctl32.dll
		// reports version 5. So it does not use transparent icons for image list.
		// Default path loads the dll from C:\windows\WinSxS folder.

		char path[ MAX_PATH ] = "";
		bool usePath = false;

		if ( sizeof( HANDLE ) >= 64 / sizeof( BYTE ) )
		{
			usePath = true;
			GetSystemDirectoryA( path, maxLen( path ) );
			strcat_s( path, maxLen( path ), "\\" );
		}
		
		strcat_s( path, maxLen( path ), "comctl32.dll" );

		bool ok = CSysDll::Load( usePath, path, NULL, false );

#ifdef _DEBUG
		if ( ok )
		{
			GetModuleFileNameA( hInst, path, maxLen( path ) );
		}
#endif

		return ok;
	}
};

extern bool isWinXP();
extern bool isWinVista();

typedef bool typeEnumWindowActor( HWND hWnd, void *pParam );
int EnumWindowsByName( const wchar_t *wndName, typeEnumWindowActor *enumWindowActor, void *param, LPTSTR wndClass = WC_DIALOG );

//
//	Use static functions instead
//	'comctl32dll.LoadOnce' -> 'ComCtl32Dll::getStatic().LoadOnce'
//	'shell32dll.m_SHGetSpecialFolderPath' - 'cShell32Dll::getStatic().m_SHGetSpecialFolderPath'
//
//extern cComCtl32Dll comctl32dll;
//extern cShell32Dll shell32dll; // Implementataion has to be once included


//
//	TimetToFileTime, FileTimeToTimet
//		The FILETIME structure is a 64-bit value representing the number of 100-nanosecond intervals since January 1, 1601 (UTC).
//		time() returns the time as seconds elapsed since midnight, January 1, 1970.
//
void TimetToFileTime( FILETIME *pft, time_t t );
time_t FileTimeToTimet( const FILETIME *ft );

time_t SystemTimeToTimet( const SYSTEMTIME * pTime );

BOOL GetFileNameFromHandle(HANDLE hFile, wchar_t *pszFilename, DWORD size = MAX_PATH, bool bFinaleNameOnly = false );
DWORD GetFinalPathNameByHandleT( HANDLE hFile, LPWSTR lpszFilePath, DWORD cchFilePath, DWORD dwFlags = 0 ); // dwFlags == VOLUME_NAME_DOS 
const wchar_t *GetReparsePointFinalPathName( cWCS &finaleFileName, wchar_t *reparsePointName, DWORD *pErr = NULL );

//
//	GetIDListFromPath
//		Creates PID list from specified path. Remember fre the created list by FreeIDList
//
ITEMIDLIST *GetIDListFromPath( LPCWSTR path, bool bCoInitialize = true );
void FreeIDList( ITEMIDLIST *pidl );

//
//	GetDOSNameForHarddiskVolume
//		Input name is '\Device\HarddiskVolumeNN\filePath'
//		Enumerates logical drives GetLogicalDrives 
//		Queries dos device name by QueryDosDeviceT ( X:\ ) -> '\Device\HarddiskVolumeNN'
//		Formats output dos name as 'X:\filePath'
//		Returns true if path is translated.
//
bool GetDOSNameForHarddiskVolume( wchar_t *bf, int bfLen, const wchar_t *hardDiskFilePath );

//
//	isAdministrator
//		Checks administrator privileges by AdvAPI OpenSCManager function.
//
bool isAdministrator();

#ifndef LAYOUT_RTL
	#define LAYOUT_RTL                         0x00000001 // Right to left
#endif

//
//	Set/reset compression of existion file...
//
bool SetCompression( HANDLE hFile, USHORT state = COMPRESSION_FORMAT_DEFAULT );
static inline bool ResetCompression( HANDLE hFile )
{
	return SetCompression( hFile, COMPRESSION_FORMAT_NONE );
}

//
//	createNewRegularFileFile
//		If file is created at folder with ecrypted or compressed attributes so the attributes are automatically inherited.
//		So this function creates new file, checks if it is encrypted or compressed, resets both encryption and compression.
//
int createNewRegularFileFile( const wchar_t *filePath, HANDLE *pHFile = NULL );

//
//	createNonCompressedFile
//		Create file and reset compressed attributes if required.
//
HANDLE createNonCompressedFile( const wchar_t *filePath, DWORD dwDisp = OPEN_ALWAYS );

//
//	SetFileSize
//		Set required file size.
//		Note, if the file is extended so the data after previouse logical end of file up to new file size
//		will be assigned as zero by file system security. Use SetFileSizeAndValidData from AdvAPI.h to validate new tail data.
//
bool SetFileSize( HANDLE hFile, DWORDLONG fileSize, DWORD *pErr = NULL );
bool SetFileSize( const wchar_t *filePath, DWORDLONG fileSize, bool bFlushBuffers = false );

typedef DWORD typeTerminateThreadActor( void *ctx, const wchar_t *szInfo, HANDLE hThread, DWORD exitCode );

DWORD WaitThreadAndTerminate_SA(
		const wchar_t *szFunction, int iLine,
		typeTerminateThreadActor *actor, void *ctx,
		HANDLE hThread, DWORD dwMilliseconds, DWORD exitCode );

//#define WaitThreadAndTerminate( hThread, dwMilliseconds, exitCode ) WaitThreadAndTerminate_S( _CRT_WIDE( __FUNCTION__ ), __LINE__, hThread, dwMilliseconds, exitCode )

class cAutoHandle
{
public:
	HANDLE m_handle;

	bool Set( HANDLE h )
	{
		Close();
		m_handle = h;
		return true;
	}

	cAutoHandle( HANDLE h = NULL )
	{
		m_handle = h;
	}

	~cAutoHandle()
	{
		Close();
	}

	void Close()
	{
		if ( m_handle == NULL )
			return;

		if ( m_handle == INVALID_HANDLE_VALUE )
			return;

		CloseHandle( m_handle );
		m_handle = NULL;
	}
};

BOOL SetFileTimeT(
    const wchar_t *fileName,
    CONST FILETIME *lpCreationTime,
    CONST FILETIME *lpLastAccessTime,
    CONST FILETIME *lpLastWriteTime
    );

BOOL SafeGetFileTime(
	const wchar_t *fileName,
	LPFILETIME lpCreationTime,
	LPFILETIME lpLastAccessTime, 
	LPFILETIME lpLastWriteTime,
	LPDWORD lpDWFileAttributes = NULL,
	DWORDLONG *lpSizeSize = NULL
	);

#if (_WIN32_WINNT < 0x0600)
typedef struct _FILE_BASIC_INFO {
    LARGE_INTEGER CreationTime;
    LARGE_INTEGER LastAccessTime;
    LARGE_INTEGER LastWriteTime;
    LARGE_INTEGER ChangeTime;
    DWORD FileAttributes;
} FILE_BASIC_INFO, *PFILE_BASIC_INFO;
#endif

class cFileAttributes
{
public:
	FILETIME creationTime;
	FILETIME lastAccessTime;
	FILETIME lastWriteTime;
	DWORD attributes;

	FILE_BASIC_INFO m_bi;
	bool m_biOk;
	DWORD m_err;
	
	cFileAttributes()
	{
		m_biOk = false;
		m_err = -1;
	}

	BOOL Get( const wchar_t *fileName );
	BOOL Set( const wchar_t *fileName );
};

extern bool PostCloseAppMessage( const wchar_t *wndClass, const wchar_t *caption = NULL );

void FileTimeToUnixTime( FILETIME ft, DWORDLONG *pMilliSeconds );
void UnixTimeToFileTime( DWORDLONG milliSeconds, LPFILETIME pft);
void SystemTimeToUnixTime( LPSYSTEMTIME pst, DWORDLONG *pMilliSeconds );
void UnixTimeToSystemTime( DWORDLONG milliSeconds, LPSYSTEMTIME pst );

#endif // #ifdef _WIN32

#endif // #ifdef _UnicodeFuncs_


