#include "cATADevice.h"

#include <ntddscsi.h>
#include <winioctl.h>
#include <stddef.h> // offsetof()

#include "UnicodeFuncs.h"
#include "ODS.h"

#pragma pack(1)

typedef struct _GETVERSIONINPARAMS_EX {
  BYTE bVersion;
  BYTE bRevision;
  BYTE bReserved;
  BYTE bIDEDeviceMap;
  DWORD fCapabilities;
  DWORD dwDeviceMapEx;  // 3ware specific: RAID drive bit map
  WORD wIdentifier;     // Vendor specific identifier
  WORD wControllerId;   // 3ware specific: Controller ID (0,1,...)
  ULONG dwReserved[2];
} GETVERSIONINPARAMS_EX;


//
// Note! '_WIN32_WINNT=0x0400' has to be defined for IDEREGS
//
typedef struct _SENDCMDINPARAMS_EX {
  DWORD cBufferSize;
  IDEREGS irDriveRegs;
  BYTE bDriveNumber;
  BYTE bPortNumber;     // 3ware specific: port number
  WORD wIdentifier;     // Vendor specific identifier
  DWORD dwReserved[4];
  BYTE bBuffer[1];
} SENDCMDINPARAMS_EX;

#pragma pack()

// 3ware specific versions of SMART ioctl structs

#define SMART_VENDOR_3WARE      0x13C1  // identifies 3ware specific parameters

// call SMART_GET_VERSION, return device map or -1 on error

static int smart_get_version( HANDLE hdevice, GETVERSIONINPARAMS_EX * ata_version_ex = 0 )
{
	ODSFunc;

	GETVERSIONINPARAMS vers;
	memset(&vers, 0, sizeof(vers));

	const GETVERSIONINPARAMS_EX & vers_ex = (const GETVERSIONINPARAMS_EX &)vers;

	DWORD num_out;

	if ( !DeviceIoControl( hdevice, SMART_GET_VERSION, NULL, 0, &vers, sizeof(vers), &num_out, NULL ) )
	{
		ODSF( L"SMART_GET_VERSION failed, Error=%ld", GetLastError());
		return -1;
	}

	assert(num_out == sizeof(GETVERSIONINPARAMS));

	ODSF( L"SMART_GET_VERSION suceeded, bytes returned: %lu", num_out );
	ODSF( L"Vers = %d.%d, Caps = 0x%lx, DeviceMap = 0x%02x",
			vers.bVersion, vers.bRevision,
			vers.fCapabilities, vers.bIDEDeviceMap);

	if ( vers_ex.wIdentifier == SMART_VENDOR_3WARE )
		ODSF( L"Identifier = %04x(3WARE), ControllerId=%u, DeviceMapEx = 0x%08lx\n",
			vers_ex.wIdentifier, vers_ex.wControllerId, vers_ex.dwDeviceMapEx);

	if ( ata_version_ex != NULL )
		*ata_version_ex = vers_ex;

	// TODO: Check vers.fCapabilities here?
	return vers.bIDEDeviceMap;
}

#define ATA_IDENTIFY_DEVICE             0xec
#define ATA_IDENTIFY_PACKET_DEVICE      0xa1

// Identify Device returns at  83 word bit 2 1 = CFA feature set supported
#define ATA_CFA_TRANSLATE_SECTOR		0x87

//	ATA-8 7.4 CFA TRANSLATE SECTOR - 87h, PIO data-in
//  ATA-6 8.3 CFA TRANSLATE SECTOR - 87h
// 7.4.2 Description
// This command provides information related to a specific logical sector. The data indicates the erased or not
// erased status of the logical sector, and the number of erase and write cycles performed on that sector. Devices
// may return zero in fields that do not apply or that are not supported by the device


#define ATA_CFA_WRITE_SECTORS_WITHOUT_ERASE	0x38
//  ATA-6 8.5.1 CFA WRITE SECTORS WITHOUT ERASE - 38h

// Returns ERORR_ACCESS_DENIED 5 for non adminstrator.
static int ata_pass_through_ioctl( HANDLE hdevice, IDEREGS * regs, IDEREGS * prev_regs, char * data, int datasize );

// It works for NT4
static int ata_via_scsi_pass_through_ioctl( HANDLE hdevice, IDEREGS * regs, char * data, unsigned datasize );

// It does never work
static int ide_pass_through_ioctl( HANDLE hdevice, IDEREGS * regs, char * data, unsigned datasize);

int cDevice::Open( int phydrive, int logdrive )
{
	//
	//	Prepare device path name.
	//
	if ( ( 0 <= phydrive ) && ( phydrive <= 255 ) )
		FormatT( m_devpath, maxLen( m_devpath ), L"\\\\.\\PhysicalDrive%d", phydrive);
	else
	if ( ( 0 <= logdrive ) && ( logdrive <= 'Z'-'A' ) )
		FormatT( m_devpath, maxLen( m_devpath ), L"\\\\.\\%c:", L'A'+logdrive);
	else
	{
		ODSF( L"Invalid argumments: phydrive %d, logdrive %c", phydrive, L'A'+logdrive );
		return -1;
	}

	return Open( m_devpath );
}

int cDevice::Open( const wchar_t *devpath )
{
	cWCStr wcs;

	int len = wcslen( devpath );
	
	if ( len == 0 )
	{
		ODS_fls( _FormatT( L"Assertion! Invalid path '%s'", devpath ) );
		return ERROR_FILE_NOT_FOUND;
	}
	
	//
	//	It cannot open volume root directory by CreateFileT, but allows to open volume.
	//
	if ( devpath[len-1] == L'\\' )
	{
		wcs.Set( devpath, len - 1  );
		devpath = wcs.m_szBody;
	}

	//
	//	Open device.
	//
	HANDLE h = INVALID_HANDLE_VALUE;

	// Open with admin rights
	m_bAdmin = true;
	h = CreateFileT( devpath,
		GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, 0, 0);

	if ( h == INVALID_HANDLE_VALUE )
	{
		// Open without admin rights
		m_bAdmin = false;
		h = CreateFileT( devpath, GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, 0, 0 );
	}

	if ( h == INVALID_HANDLE_VALUE )
	{
		// Open without admin rights
		m_bAdmin = false;
		h = CreateFileT( devpath, 0,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, 0, 0 );
	}

	if ( h == INVALID_HANDLE_VALUE )
	{
		long err = GetLastError();

		if (err == ERROR_FILE_NOT_FOUND)
		{
			//ODSF( L"%s: not found", devpath );
		}
		else if (err == ERROR_ACCESS_DENIED)
			ODS_fls( _FormatT( L"%s: access denied", devpath ) );
		else
			ODS_fls( _FormatT( L"%s: Error=%ld", devpath, err ) );

		return err;
	}

	//ODSF( L"%s: successfully opened%s", devpath, (!m_bAdmin ? L" (without admin rights)" : L"" ));

	m_hDevice = h;

	return 0;
}

static unsigned char byteCheckSum( const void * data, int sizeInByte = 512 )
{
	ODSFuncS( _FormatT( L"%X, %d", data, sizeInByte ) );

	unsigned char sum = 0;

	for (int i = 0; i < sizeInByte; i++)
		sum += ((const unsigned char *)data)[i];

	ODSFuncRes( (DWORD)sum );

	return sum;
}

//static 
int cATADevice::identifyDevice( HANDLE hDevice, ata_identity *ai )
{
	ODSFunc;
	//
	//	It works for Windows 7 but is excessive
	//
	//GETVERSIONINPARAMS_EX vers_ex;

	//int devmap = smart_get_version( h, &vers_ex );

	//if ( devmap < 0 )
	//{
	//	ODSF( L"%s: ATA driver has no SMART support", m_devpath );
	//	::CloseHandle( h );
	//	continue;
	//}

	//
	// Call IDENTIFY DEVICE - ECh, PIO Data-in Device IO control
	//
	memset( ai, 0, sizeof( *ai ) );

	// Set IDEREGS
	IDEREGS regs;//, prev_regs;
	{
		//const ata_in_regs & lo = in.in_regs;
		regs.bFeaturesReg     = 0; //lo.features;
		regs.bSectorCountReg  = 1; //lo.sector_count;
		regs.bSectorNumberReg = 0; //lo.lba_low;
		regs.bCylLowReg       = 0; //lo.lba_mid;
		regs.bCylHighReg      = 0; //lo.lba_high;
		regs.bDriveHeadReg    = 0; //lo.device;
		regs.bCommandReg      = ATA_IDENTIFY_DEVICE; // lo.command;
		regs.bReserved        = 0;
	}

	int res = ata_pass_through_ioctl( hDevice, &regs, NULL, (char *)ai, sizeof( *ai ) );

	//
	//	Identify packet device
	//
	if ( res == -1 )
	{
		//const ata_in_regs & lo = in.in_regs;
		regs.bFeaturesReg     = 0; //lo.features;
		regs.bSectorCountReg  = 1; //lo.sector_count;
		regs.bSectorNumberReg = 0; //lo.lba_low;
		regs.bCylLowReg       = 0; //lo.lba_mid;
		regs.bCylHighReg      = 0; //lo.lba_high;
		regs.bDriveHeadReg    = 0; //lo.device;
		regs.bCommandReg      = ATA_IDENTIFY_PACKET_DEVICE; // lo.command;
		regs.bReserved        = 0;

		//res = ata_pass_through_ioctl( hDevice, &regs, NULL, (char *)&ai, sizeof( ai ) );
		res = ata_pass_through_ioctl( hDevice, &regs, NULL, (char *)ai, sizeof( *ai ) );
	}

	if ( res == -1 ) // for NT4
	{
		//const ata_in_regs & lo = in.in_regs;
		regs.bFeaturesReg     = 0; //lo.features;
		regs.bSectorCountReg  = 1; //lo.sector_count;
		regs.bSectorNumberReg = 0; //lo.lba_low;
		regs.bCylLowReg       = 0; //lo.lba_mid;
		regs.bCylHighReg      = 0; //lo.lba_high;
		regs.bDriveHeadReg    = 0; //lo.device;
		regs.bCommandReg      = ATA_IDENTIFY_DEVICE; // lo.command;
		regs.bReserved        = 0;

		//res = ata_via_scsi_pass_through_ioctl( hDevice, &regs, (char *)&ai, sizeof( ai ) );
		res = ata_via_scsi_pass_through_ioctl( hDevice, &regs, (char *)ai, sizeof( *ai ) );
	}

	//if ( res == -1 ) // IDE
	//{
	//	//const ata_in_regs & lo = in.in_regs;
	//	regs.bFeaturesReg     = 0; //lo.features;
	//	regs.bSectorCountReg  = 1; //lo.sector_count;
	//	regs.bSectorNumberReg = 0; //lo.lba_low;
	//	regs.bCylLowReg       = 0; //lo.lba_mid;
	//	regs.bCylHighReg      = 0; //lo.lba_high;
	//	regs.bDriveHeadReg    = 0; //lo.device;
	//	regs.bCommandReg      = ATA_IDENTIFY_DEVICE; // lo.command;
	//	regs.bReserved        = 0;

	//	res = ide_pass_through_ioctl( h, &regs, (char *)&ai, sizeof( ai ) );
	//}

	// If there is a checksum there, validate it
	unsigned short *rawshort=(unsigned short *)ai;
	unsigned char  *rawbyte =(unsigned char  *)ai;

	if ((rawshort[255] & 0x00ff) == 0x00a5 && byteCheckSum( rawbyte, 512 ))
	{
		ODSF( L"Warning! %s error: invalid SMART checksum.", L"Drive Identity Structure" );
		return -1;
	}

	if ( res != 0 )
		ODS_fls( _FormatT( L"Assertion! Error %d", res ) );

	return res;
}

// static
bool cATADevice::isSSDDrive( ata_identity *ai )
{
	ODSFuncS( getWCStr( (DWORD)ai ) );

	// (word 217) Nominal media rotation rate 
	bool bIsSSD = ( ai->rot_rate == 1 );

	ODS_fls( getWCStr( bIsSSD ) );

	return bIsSSD;
}

//static 
int cATADevice::getPhysDriveNum( int logdrive )
{
	cATADevice ld;

	if ( ld.Open( -1, logdrive ) != 0 )
		return -1;

	VOLUME_DISK_EXTENTS vde;
	DWORD size = sizeof( vde );
	vde.NumberOfDiskExtents = 1;

	BOOL ok = DeviceIoControl(
		ld.m_hDevice,
		IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS,
		NULL,	// lpInBuffer
		0,		// nInBufferSize
		&vde,	// output buffer
		size,	// size of output buffer
		&size,	// number of bytes returned
		NULL	// OVERLAPPED structure
	);

	if ( !ok )
	{
		DWORD err = GetLastError();

		if ( ( err == ERROR_INSUFFICIENT_BUFFER ) || ( err == ERROR_MORE_DATA ) )
		{
			ODS_fls( _FormatT( L"Logical drive locates on %d physical drives", vde.NumberOfDiskExtents ) );
			return -2;
		}

		return -1;
	}

	//
	//	IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS has to return error.
	//
	if ( vde.NumberOfDiskExtents != 1 )
	{
		ODS_fls( _FormatT( L"Assertion! Logical drive locates on %d physical drives", vde.NumberOfDiskExtents ) );
		return -2;
	}

	return vde.Extents[0].DiskNumber;

	//STORAGE_DEVICE_NUMBER sdn;
	//DWORD size = sizeof( sdn );

	//BOOL ok = DeviceIoControl(
	//	ld.m_hDevice,
	//	IOCTL_STORAGE_GET_DEVICE_NUMBER,
	//	NULL,	// lpInBuffer
	//	0,		// nInBufferSize
	//	&sdn,	// output buffer
	//	size,	// size of output buffer
	//	&size,	// number of bytes returned
	//	NULL	// OVERLAPPED structure
	//);

	//if ( !ok )
	//	return -1;

	//return sdn.DeviceNumber;
}

// static
bool cATADevice::isSupportCFA( ata_identity *ai )
{
	// (word 83 ) Bit 2:	CFA Feature Set
	bool bCFAFeaturesSupport = (( ai->command_set_2 & ( 1 << 2 )) != 0 );

	ODSF( L"Nominal media rotation rate %02X, CFA Feature Set bit 2 at %02X - %s", ai->rot_rate, ai->command_set_2, getWCStr( bCFAFeaturesSupport ) );

	//
	//	CFA features were not tested.
	//	
	//if ( bCFAFeaturesSupport  )
	//{
	//	//const ata_in_regs & lo = in.in_regs;
	//	regs.bFeaturesReg     = 0; //lo.features;
	//	regs.bSectorCountReg  = 1; //lo.sector_count;
	//	regs.bSectorNumberReg = 0; //lo.lba_low;
	//	regs.bCylLowReg       = 0; //lo.lba_mid;
	//	regs.bCylHighReg      = 0; //lo.lba_high;
	//	regs.bDriveHeadReg    = 0; //lo.device;
	//	regs.bCommandReg      = ATA_CFA_TRANSLATE_SECTOR; // lo.command;
	//	regs.bReserved        = 0;

	//	res = ata_pass_through_ioctl( h, &regs, NULL, (char *)&ai, sizeof( ai ) );
	//}

	return bCFAFeaturesSupport;
}

bool nonEmpty(const void * data, int size)
{
  for (int i = 0; i < size; i++)
    if (((const unsigned char *)data)[i])
      return true;
  return false;
}

static void print_ide_regs(const IDEREGS * r, int out)
{
	ODSF( L"%s %s=0x%02x,%s=0x%02x, SC=0x%02x, SN=0x%02x, CL=0x%02x, CH=0x%02x, SEL=0x%02x",
		(out?L"Output: ":L"Input : "),
		(out?L"STS":L"CMD"), r->bCommandReg, (out?L"ERR":L" FR"), r->bFeaturesReg,
		r->bSectorCountReg, r->bSectorNumberReg, r->bCylLowReg, r->bCylHighReg, r->bDriveHeadReg);
}

static void print_ide_regs_io(const IDEREGS * ri, const IDEREGS * ro)
{
	print_ide_regs(ri, 0);
	if (ro)
		print_ide_regs(ro, 1);
}

/////////////////////////////////////////////////////////////////////////////
// ATA PASS THROUGH via SCSI PASS THROUGH (WinNT4 only)
#pragma pack ( 1 )

struct SCSI_PASS_THROUGH_WITH_BUFFERS
{
	SCSI_PASS_THROUGH spt;
	ULONG Filler;
	UCHAR ucSenseBuf[32];
	UCHAR ucDataBuf[512];
};

#pragma pack ()

// undocumented SCSI opcode to for ATA passthrough
#define SCSIOP_ATA_PASSTHROUGH    0xCC
//static 
int ata_via_scsi_pass_through_ioctl( HANDLE hdevice, IDEREGS * regs, char * data, unsigned datasize )
{
	ODSFunc;
	ODSF( L"SCSI_PASS_THROUGH %d offsetof %d/%d", sizeof( SCSI_PASS_THROUGH ), offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS, ucDataBuf), sizeof( SCSI_PASS_THROUGH_WITH_BUFFERS ) );

	IDEREGS * cdbregs;
	unsigned int size;
	DWORD num_out;
	const unsigned char magic = 0xcf;

	SCSI_PASS_THROUGH_WITH_BUFFERS sb;

	memset( &sb, 0, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
	sb.spt.Length = sizeof(SCSI_PASS_THROUGH);
	//sb.spt.PathId = 0;
	sb.spt.TargetId = 1;
	//sb.spt.Lun = 0;
	sb.spt.CdbLength = 10; sb.spt.SenseInfoLength = 24;
	sb.spt.TimeOutValue = 10;
	sb.spt.SenseInfoOffset = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS, ucSenseBuf);
	size = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS, ucDataBuf);
	sb.spt.DataBufferOffset = size;

	if (datasize) {
		if (datasize > sizeof(sb.ucDataBuf)) {
			//errno = EINVAL;
			return -1;
		}
		sb.spt.DataIn = SCSI_IOCTL_DATA_IN;
		sb.spt.DataTransferLength = datasize;
		//size += datasize;
		size = sizeof( SCSI_PASS_THROUGH_WITH_BUFFERS );
		sb.ucDataBuf[0] = magic;
	}
	else {
		sb.spt.DataIn = SCSI_IOCTL_DATA_UNSPECIFIED;
		//sb.spt.DataTransferLength = 0;
	}

	// Use pseudo SCSI command followed by registers
	sb.spt.Cdb[0] = SCSIOP_ATA_PASSTHROUGH;
	cdbregs = (IDEREGS *)(sb.spt.Cdb+2);
	*cdbregs = *regs;

	if (!DeviceIoControl(hdevice, IOCTL_SCSI_PASS_THROUGH,
		&sb, size, &sb, size, &num_out, NULL)) {
			long err = GetLastError();
			//if (con->reportataioctl)
				ODSF( L"ATA via IOCTL_SCSI_PASS_THROUGH failed, Error=%ld", err );
			//errno = (err == ERROR_INVALID_FUNCTION || err == ERROR_NOT_SUPPORTED ? ENOSYS : EIO);
			return -1;
	}

	// Cannot check ATA status, because command does not return IDEREGS

	// Check and copy data
	if (datasize)
	{
		//
		// XP reports:
		// ATA via IOCTL_SCSI_PASS_THROUGH output data missing (80), 84
		//

		unsigned int sizeRequired = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS, ucDataBuf) + datasize;

//		if ( ( num_out != size ) ||
		if ( ( num_out < sizeRequired ) ||
		     ( ( sb.ucDataBuf[0] == magic ) && !nonEmpty(sb.ucDataBuf+1, datasize-1) ) )
		{
				ODSF( L"ATA via IOCTL_SCSI_PASS_THROUGH output data missing (%lu), %lu, %lu, %lu, %x", num_out, sizeRequired, size, datasize, (int)sb.ucDataBuf[0] );
				print_ide_regs_io(regs, NULL);
				return -1;
		}

		memcpy( data, sb.ucDataBuf, datasize );
	}

	ODSF( L"ATA via IOCTL_SCSI_PASS_THROUGH suceeded, bytes returned: %lu", num_out );
	print_ide_regs_io(regs, NULL);

	return 0;
}

//static 
int ata_pass_through_ioctl(HANDLE hdevice, IDEREGS * regs, IDEREGS * prev_regs, char * data, int datasize)
{
	const int max_sectors = 32; // TODO: Allocate dynamic buffer

	typedef struct {
		ATA_PASS_THROUGH_EX apt;
		ULONG Filler;
		UCHAR ucDataBuf[max_sectors * 512];
	} ATA_PASS_THROUGH_EX_WITH_BUFFERS;

	const unsigned char magic = 0xcf;

	ATA_PASS_THROUGH_EX_WITH_BUFFERS ab; memset(&ab, 0, sizeof(ab));
	ab.apt.Length = sizeof(ATA_PASS_THROUGH_EX);
	//ab.apt.PathId = 0;
	//ab.apt.TargetId = 0;
	//ab.apt.Lun = 0;
	ab.apt.TimeOutValue = 10;
	unsigned size = offsetof(ATA_PASS_THROUGH_EX_WITH_BUFFERS, ucDataBuf);
	ab.apt.DataBufferOffset = size;

	if (datasize > 0) {
		if (datasize > (int)sizeof(ab.ucDataBuf)) {
		//	errno = EINVAL;
			return -1;
		}
		ab.apt.AtaFlags = ATA_FLAGS_DATA_IN;
		ab.apt.DataTransferLength = datasize;
		size += datasize;
		ab.ucDataBuf[0] = magic;
	}
	else if (datasize < 0) {
		if (-datasize > (int)sizeof(ab.ucDataBuf)) {
			//errno = EINVAL;
			return -1;
		}
		ab.apt.AtaFlags = ATA_FLAGS_DATA_OUT;
		ab.apt.DataTransferLength = -datasize;
		size += -datasize;
		memcpy(ab.ucDataBuf, data, -datasize);
	}
	else {
		assert(ab.apt.AtaFlags == 0);
		assert(ab.apt.DataTransferLength == 0);
	}

	assert(sizeof(ab.apt.CurrentTaskFile) == sizeof(IDEREGS));
	IDEREGS * ctfregs = (IDEREGS *)ab.apt.CurrentTaskFile;
	IDEREGS * ptfregs = (IDEREGS *)ab.apt.PreviousTaskFile;
	*ctfregs = *regs;

	if (prev_regs) {
		*ptfregs = *prev_regs;
		ab.apt.AtaFlags |= ATA_FLAGS_48BIT_COMMAND;
	}

	DWORD num_out;
	if (!DeviceIoControl(hdevice, IOCTL_ATA_PASS_THROUGH,
		&ab, size, &ab, size, &num_out, NULL))
	{
			long err = GetLastError();
			//if (con->reportataioctl) {
				ODSF( L"IOCTL_ATA_PASS_THROUGH (%lu) failed, Error=%ld", (int)ctfregs->bCommandReg, err );
				print_ide_regs_io(regs, NULL);
			//}
			//errno = (err == ERROR_INVALID_FUNCTION || err == ERROR_NOT_SUPPORTED ? ENOSYS : EIO);
			return -1;
	}

	// Check ATA status
	if (ctfregs->bCommandReg/*Status*/ & (0x01/*Err*/|0x08/*DRQ*/)) {
		//if (con->reportataioctl) {
			ODSF( L"IOCTL_ATA_PASS_THROUGH (%lu) command failed: ", (int)ctfregs->bCommandReg );
			print_ide_regs_io(regs, ctfregs);
		//}
		//errno = EIO;
		return -1;
	}

	// Check and copy data
	if (datasize > 0)
	{
		if ( ( num_out != size ) 
			|| ( ab.ucDataBuf[0] == magic && !nonEmpty(ab.ucDataBuf+1, datasize-1 ) ) )
		{
			ODSF( L"IOCTL_ATA_PASS_THROUGH (%lu) output data missing (%lu)", (int)ctfregs->bCommandReg, num_out);
			print_ide_regs_io(regs, ctfregs);
			return -1;
		}

		memcpy( data, ab.ucDataBuf, datasize );
	}

	ODSF( L"IOCTL_ATA_PASS_THROUGH (%lu) suceeded, bytes returned: %lu", (int)ctfregs->bCommandReg, num_out);
	print_ide_regs_io(regs, ctfregs);

	*regs = *ctfregs;
	if (prev_regs)
		*prev_regs = *ptfregs;

	return 0;
}

#pragma pack(1)

typedef struct {
  IDEREGS IdeReg;
  ULONG DataBufferSize;
  UCHAR DataBuffer[1];
} ATA_PASS_THROUGH;

#pragma pack()

//static 
int ide_pass_through_ioctl( HANDLE hdevice, IDEREGS * regs, char * data, unsigned datasize )
{
	if (datasize > 512) {
		//errno = EINVAL;
		return -1;
	}
	unsigned int size = sizeof(ATA_PASS_THROUGH)-1 + datasize;
	ATA_PASS_THROUGH * buf = (ATA_PASS_THROUGH *)VirtualAlloc(NULL, size, MEM_COMMIT, PAGE_READWRITE);
	DWORD num_out;
	const unsigned char magic = 0xcf;

	if (!buf) {
		//errno = ENOMEM;
		return -1;
	}

	buf->IdeReg = *regs;
	buf->DataBufferSize = datasize;
	if (datasize)
		buf->DataBuffer[0] = magic;

	if (!DeviceIoControl(hdevice, IOCTL_IDE_PASS_THROUGH,
		buf, size, buf, size, &num_out, NULL)) {
			long err = GetLastError();
			//if (con->reportataioctl) {
				ODSF( L"IOCTL_IDE_PASS_THROUGH failed, Error=%ld", err);
				print_ide_regs_io(regs, NULL);
			//}
			VirtualFree(buf, 0, MEM_RELEASE);
			//errno = (err == ERROR_INVALID_FUNCTION || err == ERROR_NOT_SUPPORTED ? ENOSYS : EIO);
			return -1;
	}

	// Check ATA status
	if (buf->IdeReg.bCommandReg/*Status*/ & 0x01) {
		//if (con->reportataioctl) {
			ODSF( L"IOCTL_IDE_PASS_THROUGH command failed:");
			print_ide_regs_io(regs, &buf->IdeReg);
		//}
		VirtualFree(buf, 0, MEM_RELEASE);
		//errno = EIO;
		return -1;
	}

	// Check and copy data
	if (datasize) {
		if (   num_out != size
			|| (buf->DataBuffer[0] == magic && !nonEmpty(buf->DataBuffer+1, datasize-1))) {
				//if (con->reportataioctl) {
					ODSF( L"  IOCTL_IDE_PASS_THROUGH output data missing (%lu, %lu)",
						num_out, buf->DataBufferSize);
					print_ide_regs_io(regs, &buf->IdeReg);
				//}
				VirtualFree(buf, 0, MEM_RELEASE);
				//errno = EIO;
				return -1;
		}
		memcpy(data, buf->DataBuffer, datasize);
	}

	//if (con->reportataioctl > 1) {
		ODSF( L"  IOCTL_IDE_PASS_THROUGH suceeded, bytes returned: %lu (buffer %lu)",
			num_out, buf->DataBufferSize);
		print_ide_regs_io(regs, &buf->IdeReg);
	//}
	*regs = buf->IdeReg;

	// Caution: VirtualFree() fails if parameter "dwSize" is nonzero
	VirtualFree(buf, 0, MEM_RELEASE);
	return 0;
}

/////////////////////

static VOLUME_DISK_EXTENTS *getVdeBuff( int extNum, DWORD *pSize )
{
	DWORD size = sizeof( VOLUME_DISK_EXTENTS );
	if ( extNum > 1 )
		size += sizeof( DISK_EXTENT ) * ( extNum - 1 );

	VOLUME_DISK_EXTENTS *ptr = (VOLUME_DISK_EXTENTS *)malloc( size );

	if ( ptr == NULL )
		size = 0;
	else
		ptr->NumberOfDiskExtents = extNum;

	if ( pSize != NULL )
		*pSize = size;

	return ptr;
}

static void freeVdeBuff( VOLUME_DISK_EXTENTS *ptr )
{
	free( ptr );
}

static bool isStorageSpace( HANDLE hDrive )
{
	// ****** Definitions from Win8 ntddstor.h ******
	typedef enum _WINN8_STORAGE_PROPERTY_ID {
		StorageDeviceSeekPenaltyProperty = StorageAccessAlignmentProperty+1,
		StorageDeviceTrimProperty,
		StorageDeviceWriteAggregationProperty,
		StorageDeviceDeviceTelemetryProperty,
		StorageDeviceLBProvisioningProperty,
		StorageDevicePowerProperty,
		StorageDeviceCopyOffloadProperty,
		StorageDeviceResiliencyProperty
	} WINN8_STORAGE_PROPERTY_ID, *PWINN8_STORAGE_PROPERTY_ID;

	typedef struct _DEVICE_LB_PROVISIONING_DESCRIPTOR {
		ULONG       Version;          // keep compatible with STORAGE_DESCRIPTOR_HEADER
		ULONG       Size;             // keep compatible with STORAGE_DESCRIPTOR_HEADER

		UCHAR ThinProvisioningEnabled : 1;
		UCHAR ThinProvisioningReadZeros : 1;
		UCHAR AnchorSupported : 3;
		UCHAR UnmapGranularityAlignmentValid : 1;
		UCHAR Reserved0 : 2;
		UCHAR Reserved1[7];
		ULONGLONG OptimalUnmapGranularity;          // Granularity in bytes.
		ULONGLONG UnmapGranularityAlignment;    // Granularity alignment in bytes.
	} DEVICE_LB_PROVISIONING_DESCRIPTOR, *PDEVICE_LB_PROVISIONING_DESCRIPTOR;

	STORAGE_PROPERTY_QUERY query = {
		(STORAGE_PROPERTY_ID)StorageDeviceLBProvisioningProperty,
		PropertyStandardQuery,
		0
	};
	
	bool storageSpace = false;

	DEVICE_LB_PROVISIONING_DESCRIPTOR provDescr = {0};
	DWORD retRes = 0;
	BOOL res = DeviceIoControl(hDrive, IOCTL_STORAGE_QUERY_PROPERTY, &query, sizeof(query), &provDescr, sizeof(provDescr), &retRes, 0);
	if ( res )
	{
		storageSpace = provDescr.ThinProvisioningEnabled;
		//slabSize = provDescr.OptimalUnmapGranularity;
		ODS_fls( _FormatT( L"ThinProvisioningEnabled = %s", getWCStr( storageSpace ) ) );
	}
	else
		ODS_fls( _FormatT( L"Assertion! IOCTL_STORAGE_QUERY_PROPERTY return 0x%X", retRes ) );

	return storageSpace;
}

//
//	hasSSDDriveExtent_actor -> typeWalkLogDriveActor used by walkLogDriveExtents
//		parameter pSupportCFA has to initialised as true, this function just reset it if at least one extent does not support.
//		flag if found at leas one extent which does not support CFA.
//	If success it return 1.
//	Return 0 if drive is not SSD.
//	Negative is error.
//
int hasSSDDriveExtent_actor( sPhysDriveInfo *pdi, void *param )
{
	bool *pSupportCFA = (bool*)param;

	bool bSSD			= pdi->getFlag( sPhysDriveInfo::eTypeSDD );		// pdi->typeSSD;
	bool bSupport		= pdi->getFlag( sPhysDriveInfo::eSupportCFA );	// pdi->supportCFA;
	//bool bStoragePool	= pdi->getFlag( sPhysDriveInfo::eStoragePool );

	//
	// If at least one drive extent does not support CFA so entire logical drive does not support the function.
	//
	if ( bSSD && !bSupport )
	{
		if ( pSupportCFA != NULL )
			*pSupportCFA = false;
	}

	return ( bSSD ? 1 : 0 );
}

//
//	hasStoragePoolDriveExtent_actor -> typeWalkLogDriveActor used by walkLogDriveExtents
//		parameter NULL
//	If success it return 1.
//	Return 0 if pysical drive does not support thin provision.
//	Negative is error.
//
int hasStoragePoolDriveExtent_actor( sPhysDriveInfo *pdi, void *param )
{
	bool bStoragePool = pdi->getFlag( sPhysDriveInfo::eStoragePool );

	return ( bStoragePool ? 1 : 0 );
}

static sPhysDriveInfo *addPhysDriveInfo( DWORD diskNumber, cADTable < sPhysDriveInfo > *pdiT )
{
	ODSFuncS( getWCStr( diskNumber ) );

	if ( !pdiT->Resize( pdiT->m_Q + 1 ) )
	{	ODS_fls( L"Assertion!" );	return NULL;	}

	sPhysDriveInfo *pdi = sPhysDriveInfo::find( sPhysDriveInfo::pdEmpty, pdiT->m_T, pdiT->m_Q );

	if ( pdi == NULL )
	{	ODS_fls( L"Assertion!" );	return NULL;	}
	
	pdi->diskNumber = diskNumber;

	cATADevice ataDev;

	if ( ataDev.Open( diskNumber, -1 ) != 0 )
	{
		ODS_fls( _FormatT( L"Error in opening disk %d", diskNumber ) );
		return pdi; // Return pdi with initially empty flags...
	}

	ata_identity ai;
	if ( cATADevice::identifyDevice( ataDev.m_hDevice, &ai ) != 0 )
	{
		ODS_fls( _FormatT( L"Error in ATA device %X identification", ataDev.m_hDevice ) );
	}
	else
	{
		pdi->setFlag( sPhysDriveInfo::eTypeSDD, ataDev.isSSDDrive( &ai ) );
		pdi->setFlag( sPhysDriveInfo::eSupportCFA, ataDev.isSupportCFA( &ai ) );
	}

	pdi->setFlag( sPhysDriveInfo::eStoragePool, isStorageSpace( ataDev.m_hDevice ) );

	return pdi;
}

//
//	walkLogDriveExtents
//	Walks all extent of corresponded logical drive:
//		1. checks if the extent physical drive is SSD
//		2. sets pSupportCFA as false if at least one SDD drive does not support CFA functions
//	Returns:
//		-1 if error
//		0 - does not have extent with SDD drive.
//		otherwise number of extents with SDD drive.
//
int walkLogDriveExtents( HANDLE hLogDrive, typeWalkLogDriveActor *actor, void *param, cADTable < sPhysDriveInfo > *pdiT ) // bool *pSupportCFA,
{
	ODSFunc;
	
	if ( pdiT == NULL )
	{
		ODS_fls( L"Assertion! pdiT is NULL" );
		return -1;
	}

	DWORD size = 0;
	VOLUME_DISK_EXTENTS *pVde = getVdeBuff( 1, &size );

	if ( size == 0 )
	{
		ODS_fls( L"Assertion! Error in memory allocation!" );
		return -1;
	}

	BOOL ok = DeviceIoControl(
		hLogDrive,
		IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS,
		NULL,	// lpInBuffer
		0,		// nInBufferSize
		pVde,	// output buffer
		size,	// size of output buffer
		&size,	// number of bytes returned
		NULL	// OVERLAPPED structure
	);

	if ( !ok )
	{
		DWORD err = GetLastError();

		if ( ( err != ERROR_INSUFFICIENT_BUFFER ) && ( err != ERROR_MORE_DATA ) )
		{
			ODS_fls( _FormatT( L"Assertion! Error %d in DeviceIoControl( IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS )", err ) );

			freeVdeBuff( pVde );
			return -1;
		}

		int numReqired = pVde->NumberOfDiskExtents;

		pVde = getVdeBuff( numReqired, &size );

		ok = DeviceIoControl(
			hLogDrive,
			IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS,
			NULL,	// lpInBuffer
			0,		// nInBufferSize
			pVde,	// output buffer
			size,	// size of output buffer
			&size,	// number of bytes returned
			NULL	// OVERLAPPED structure
		);
	}

	if ( !ok )
	{
		DWORD err = GetLastError();

		ODS_fls( _FormatT( L"Assertion! Error %d in DeviceIoControl( IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS )", err ) );
		freeVdeBuff( pVde );
		return -1;
	}

	int res = 0;
	//int countSSD = 0;
	//bool bSupportCFA = true;

	for( DWORD i = 0; i < pVde->NumberOfDiskExtents; i++ )
	{
		sPhysDriveInfo *pdi = NULL;

		// 1. Non admin has to read this information from registry.
		// 2. It does not add disk number several times.
		pdi = sPhysDriveInfo::find( pVde->Extents[i].DiskNumber, pdiT->m_T, pdiT->m_Q );
		ODS_fls( getWCStr( (DWORD)pdi ) );

		// Administrator mode may gat this information even if pdiT is NULL.
		if ( pdi == NULL )
			pdi = addPhysDriveInfo( pVde->Extents[i].DiskNumber, pdiT );

		// Not found not added.
		if ( pdi == NULL )
		{
			res = -1;
			break;
		}

		int lRes = actor( pdi, param );

		if ( lRes < 0 )
		{
			res = lRes;
			break;
		}

		res += lRes;
	}

	freeVdeBuff( pVde );
	
	return res;
}

//
//	hasStoragePoolDriveExtent
//	Walks all extent of corresponded logical drive:
//		returns number of extents which support thin provision.
//
int hasStoragePoolDriveExtent( int logDrive, cADTable < sPhysDriveInfo > *pdiT )
{
	cDevice ld;

	if ( ld.Open( -1, logDrive ) != 0 )
		return -1;

	cADTable < sPhysDriveInfo > adt;

	if ( pdiT == NULL )
		pdiT = &adt;

	int res = walkLogDriveExtents( ld.m_hDevice, hasStoragePoolDriveExtent_actor, NULL, pdiT );

	return res;
}

//
//	hasSSDDriveExtent
//	Walks all extent of corresponded logical drive:
//		1. checks if the extent physical drive is SSD
//		2. sets pSupportCFA as false if at least one SDD drive does not support CFA functions
//	Returns:
//		-1 if error
//		0 - does not have extent with SDD drive.
//		otherwise number of extents with SDD drive.
//
int hasSSDDriveExtent( int logDrive, bool *pSupportCFA )
{
	cDevice ld;

	if ( ld.Open( -1, logDrive ) != 0 )
		return -1;

	//return hasSSDDriveExtent( ld.m_hDevice, pSupportCFA, NULL, 0 );

	cADTable < sPhysDriveInfo > adt;

	int res = walkLogDriveExtents( ld.m_hDevice, hasSSDDriveExtent_actor, pSupportCFA, &adt );

	return res;
}

int getLogicalDriveNumber( const wchar_t *filePath )
{
	if ( filePath[0] == L'\\' )
	{
		ODS_fls( L"Assertion!" );
		return -1;
	}

	return towupper( filePath[0] ) - L'A';
}

bool getLogicalDriveName( wchar_t *bf, int bfSize, const wchar_t *filePath )
{
	wchar_t mountPoint[ MAX_PATH ];

	if ( !GetVolumePathNameT( filePath, mountPoint, maxLen( mountPoint ) ) )
		return false;

	if( !GetVolumeName( bf, bfSize, mountPoint ) )
		return false;

	return true;
}

//
//	szPDInfo value is used to save physical drive info propared at administrative mode.
//
const wchar_t *szPDInfo = L"PhysDriveInfo";

//static 
cADTable < sPhysDriveInfo > cPhysDriveTable::m_adt;

int cPhysDriveTable::SavePDInfoToRegistry( const wchar_t *szBCWipeKey )
{
	//ODSFunc;

	cAutoDeallocate < BYTE > adb; // PlainTextBlock ptb;
	
	for( int i = 0; i < m_adt.m_Q; i++ )
	{
		sPhysDriveInfo *pdi = m_adt.m_T + i;

		if ( pdi->diskNumber != sPhysDriveInfo::pdEmpty )
			//ptb.WriteBytes( (BYTE*)pdi, sizeof( *pdi ) );
			adb.Add( (BYTE*)pdi, sizeof( *pdi ) );
	}

	// RegSetValueT( HKEY_LOCAL_MACHINE, szBCWipeKey, szPDInfo, REG_BINARY, ptb.buffer, ptb.bufferSize );
	RegSetValueT( HKEY_LOCAL_MACHINE, szBCWipeKey, szPDInfo, REG_BINARY, adb.m_ptr, adb.m_len );

	//ODSFuncRes( m_adt.m_Q );
	return m_adt.m_Q;
}

int cPhysDriveTable::ReadPDInfoFromRegistry( const wchar_t *szBCWipeKey )
{
	//ODSFunc;

	m_adt.Free();

	DWORD size = 0;

	LONG lRes = RegGetValueT( HKEY_LOCAL_MACHINE, szBCWipeKey, szPDInfo, REG_BINARY, 0, NULL, &size );

	if ( ( lRes != ERROR_MORE_DATA ) && ( lRes != ERROR_SUCCESS ) )
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	//cPlainTextBlock ptb;
	cAutoDeallocate < BYTE > adb;

	//if ( ptb.Allocate( size ) == NULL )
	if ( adb.Allocate( size ) == NULL )
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	//lRes = RegGetValueT( HKEY_LOCAL_MACHINE, szBCWipeKey, szPDInfo, REG_BINARY, 0, ptb.buffer, &size );
	lRes = RegGetValueT( HKEY_LOCAL_MACHINE, szBCWipeKey, szPDInfo, REG_BINARY, 0, adb.m_ptr, &size );

	if ( lRes != ERROR_SUCCESS )
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	//m_adt.Add( (sPhysDriveInfo *)ptb.buffer, size / sizeof( sPhysDriveInfo ) );
	m_adt.Add( (sPhysDriveInfo *)adb.m_ptr, size / sizeof( sPhysDriveInfo ) );

	//ODSFuncRes( m_adt.m_Q );

	return m_adt.m_Q;
}
