#pragma once

#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\AutoDeallocate.h"

//
//	Class cInterceptor is required for redirection of child window procedure to parent.
//	The parent dialog class has to be derived with cInterceptor.
//	1. It attachs the windows to be intercepted at dialog initialisation.
//	2. The virtual interceptorActor function is called before original function is called.
//		If the actor return false the original procedure is not called, the lRes is returned.
//	3. All intercepted windows have to by detached at dialog is destroing.
//	Note, it changes GWLP_USERDATA of intercepted window.
//
//	Usage:
//		Declaration:
//			class cAnyDlg : public cInterceptor
//			{
//				...
//				virtual bool interceptorActor( LRESULT *pRes, HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
//				{
//					return true; // allows to call original procedure.
//				}
//				...
//			}
//
//		WM_INITDIALOG
//			interceptorAttach( GetDlgItem( m_hDlg, IDC_Item ) );
//		WM_DESTROY
//			interceptorDetach( NULL ); // Detach all
//			
//

struct sIterceptorItem
{
	HWND hWnd;
	LONG_PTR originalWndProc;
};

class cInterceptor
{
	cADBlock < sIterceptorItem > adt;

public:

	cInterceptor()
	{
		adt.Free();
	}

	~cInterceptor()
	{
		assert( adt.m_Q == 0 );
	}

	static LRESULT CALLBACK staticWndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		cInterceptor *pThis = (cInterceptor *)GetWindowLongPtrT( hWnd, GWLP_USERDATA );

		if ( pThis == NULL )
			return (LRESULT)-1;

		return pThis->realWndProc( hWnd, uMsg, wParam, lParam );
	}

	int findItem( HWND hWnd )
	{
		int i = adt.m_Q;

		while( i-- )
		{
			if ( adt.m_T[i].hWnd == hWnd )
				break;
		}

		return i;
	}
	
	LRESULT CALLBACK realWndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		int i = findItem( hWnd );
		if ( i == -1 ) // No same window found
			return (LRESULT)-1;

		LRESULT lRes;

		if ( !interceptorActor( &lRes, hWnd, uMsg, wParam, lParam ) )
		{
			return lRes; // original procedure is blocked.
		}

		return CallWindowProc( (WNDPROC)adt.m_T[i].originalWndProc, hWnd, uMsg, wParam, lParam );
	}

	bool interceptorAttach( HWND hWnd )
	{
		sIterceptorItem ii; 

		ii.hWnd = hWnd;
		ii.originalWndProc = SetWindowLongPtrT( hWnd, GWLP_WNDPROC, (LONG_PTR)staticWndProc );
		SetWindowLongPtrT( hWnd, GWLP_USERDATA, (LONG_PTR)this );

		return adt.Add( ii );		
	}

	bool interceptorDetach( int i )
	{
		if ( i >= adt.m_Q )
		{
			assert( false );
			return false;
		}

		SetWindowLongPtrT( adt.m_T[i].hWnd, GWLP_WNDPROC, (LONG_PTR)adt.m_T[i].originalWndProc );
		adt.m_T[i].hWnd = NULL;

		return true;
	}

	bool interceptorDetach( HWND hWnd = NULL )
	{
		if ( hWnd != NULL )
		{
			int i = findItem( hWnd );
			if ( i == -1 ) // No same window found
				return false;
			
			return interceptorDetach( i );
		}

		int i = adt.m_Q;

		while( i-- )
		{
			if ( adt.m_T[i].hWnd != NULL )
				interceptorDetach( i );
		}

		adt.Free();

		return true;
	}

	virtual bool interceptorActor( LRESULT *pRes, HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		return true; // allows to call original procedure.
	}
};