/*
 *	Copyright 2017 Jetico Inc. All rights reserved.
 *	cThreadTask.h
 *	Created:	08.09.2017
 *	Author:		Alexander Pichuev.
 */

#pragma once

//
//	class cThreadTask
//		It is used as object by once created thread with static 'cThreadTask::staticRun' function.
//		The function does: 
//			1. Waits Start semaphore to start
//			2. Runs virtual function 'run' or stops depending on task flag
//			3. Set Finish semathore to signaled state
//
//	Sample:
//
//	class cShowText : public cThreadTask
//	{
//	public:
//		const char *text;
//		cShowText()
//		{
//			text = "no text!";
//		}
//
//		virtual DWORD run()
//		{
//			printf( text );
//		}
//
//	} newTask;
//
//	newTask.text = "first\n";
//	newTask.execute();
//
//	newTask.text = "second\n";
//	newTask.execute();
//	
//	newTask.stop();
//


#include "cSemaphore.h"

class cThreadTask
{
	cSemaphore m_sStart;
	cSemaphore m_sStop;

	enum { efStart = 0, efStop };
	int m_flag;

	DWORD m_res;

	HANDLE m_hThread;
	DWORD m_dwId;

public:

	virtual DWORD actor() = 0;

	cThreadTask()
	{
		m_flag = efStop;
		m_res = -1;

		//
		//	Creates both start stop semaphore in non sigaled state, so all wait functions will wait when the semaphore became signaled.
		//
		m_sStart.createSemaphore( NULL, false );
		m_sStop .createSemaphore( NULL, false );

		m_hThread = CreateThread( 0, 0, staticRun, (void *)this, 0, &m_dwId );
	}

	virtual ~cThreadTask()
	{
		if ( m_hThread == NULL )
			return;

		stop();

		DWORD dwWait = WaitForSingleObject_withODS( m_hThread, INFINITE, NULL );

		CloseHandle( m_hThread );
		
		m_hThread = NULL;
	}

	static DWORD WINAPI staticRun( LPVOID lpThreadParameter )
	{
		cThreadTask *obj = (cThreadTask *)lpThreadParameter;

		if ( obj == NULL )
		{
			ODS_fls( L"Assertion! obj == NULL" );
			return -1;
		}

		ODSFuncS( _FormatT( L"%04X", obj->m_hThread ) );

		while( true )
		{
			obj->m_sStart.waitForSemaphore();

			if ( obj->m_flag == efStart )
				obj->m_res = obj->actor();

			obj->m_sStop.releaseSemaphore();

			if ( obj->m_flag == efStop )
				break;
		}

		ODSFuncRes( _FormatT( L"%04X", obj->m_hThread ) );
		return 0;
	}

	int execute()
	{
		m_flag = efStart;
		m_sStart.releaseSemaphore();
		return m_res;
	}

	int stop()
	{
		ODSFuncS( _FormatT( L"%04X", m_hThread ) );
		m_flag = efStop;
		m_sStart.releaseSemaphore();
		return m_res;
	}

	bool wait( DWORD dwMilliseconds = INFINITE )
	{
		return m_sStop.waitForSemaphore( dwMilliseconds ) == WAIT_OBJECT_0;
	}
};
