/*
 *	Copyright 2013-2014 Jetico Inc. All rights reserved.
 *	cWCS.h
 *	Base class for wide character string classes like cWCSF, cWCStr ..
 *	Created:	30.10.2013
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include <wchar.h>

class cWCS
{
public:

	wchar_t *m_szBody;
	int m_allocatedLen;
	int m_len; // Current len is used for adding string

	operator const wchar_t * ()
	{
		return m_szBody;
	}

	cWCS& operator= ( const cWCS& other)
	{
		if (this == &other) {
            return *this;
        }

		Set( other.m_szBody );
        return *this;
	}

	bool isEmpty();
	void Swap( cWCS *wcs );

	template < class T> const wchar_t *SetBody( const T *s, int len = -1 )
	{
		Set( s, len );
		return m_szBody;
	}

	virtual int Add( const char *str, int lenAdd );

	template < class T> int Set( const T *s, int len = -1 )
	{
		//
		// 22.08.2013
		//	It is not required to reallocate buffer if we just set by same pointer.
		//
		if ( (wchar_t *)s == m_szBody )
		{
			if ( len != -1 )
				SetLength( len );

			return m_len;
		}

		cWCS wcsNew;

		//
		//	15.09.2010
		//	Prev version of Set returns 0 for "" string ( but zero indicates error ),
		//	the version returns number of symbols are added.
		//
		int added = wcsNew.Add( s, len );

		Swap( &wcsNew );

		return added;
	}

	~cWCS()
	{
		Free();
	}

	void Free();

	//
	//	Add
	//		Concatenates new string to existing body.
	//		Always adds zero, but next added position depends on len parameter.
	//		If len == -1 two string are concatenated without zero separator.
	//		Use strlen + 1 to add string with zero separator.
	//
	int Add( int len ) 	{ return Add( (const wchar_t *)NULL, len ); }
	int Add( wchar_t wc ) { return Add( &wc, 1 ); }

	int Add( const wchar_t *wcs = NULL, int lenAdd = -1 );

	//
	//	SetLength
	//		If string is more then 'len' so string length is truncated.
	//		If allocated buffer is less then 'len' so the buffer is extended.
	//
	int SetLength( int len );

	//
	//	Safely gets length of buffer.
	//
	int GetLength();

	//
	//	AddLength
	//		adds zero buffer specified len to string tail
	//		returns pointer to added tail if success otherwise returns NULL 
	//
	wchar_t *AddLength( int lenAdd );

	int Truncate() { return SetLength( -1 ); }

	//
	// The reassigment is required to exclude copy constructor.
	//	Otherwise it copies m_szBody as pointer.
	//
	cWCS( cWCS &wcs, bool bSwap = false );
	cWCS( const wchar_t *wcs = NULL, int len = -1 );
	cWCS( int len );

	static void *memAllocate( size_t size ) { return (void *)new char[ size ]; }	// GlobalAlloc( GMEM_FIXED, size );
	static void	memFree( void *p )			{ delete (char*)p; }					// GlobalFree( (HGLOBAL)p );

	static wchar_t *allocateStr( int wcsLen, int *pAllocatedLen );
	static void freeStr( wchar_t *wcs, int *pAllocatedLen );


	//
	//	Inserts string at position.
	//	If len is -1 so zero termination symbol is added.
	//
	const wchar_t *InsertAt( int pos, const wchar_t *str, int len = -1 );

	static int cmp( const wchar_t *wcs0, const wchar_t *wcs1, int len = -1 );
	static int cmpI( const wchar_t *wcs0, const wchar_t *wcs1, int len = -1 );
	static bool equ( const wchar_t *wcs0, const wchar_t *wcs1, int len = -1 ) { return cmp( wcs0, wcs1, len ) == 0; }
	static bool equI( const wchar_t *wcs0, const wchar_t *wcs1, int len = -1 ) { return cmpI( wcs0, wcs1, len ) == 0; }

	int cmp( const wchar_t *wcs, int len = -1 )	{ return cmp( m_szBody, wcs, len ); }
	int cmpI( const wchar_t *wcs, int len = -1 )	{ return cmpI( m_szBody, wcs, len ); }
	bool equ( const wchar_t *wcs, int len = -1 ) { return cmp( wcs, len ) == 0; }
	bool equI( const wchar_t *wcs, int len = -1 ) { return cmpI( wcs, len ) == 0; }

	wchar_t getLast( unsigned int last = 0 );
	wchar_t getAt( int pos );
	wchar_t setAt( int pos, wchar_t ch );

	//
	//	ChangeSymbols
	//		Sample: ChangeSymbols( '\n', '$' );
	//			encode == true:  $xx\n -> $$xx$ - saves new symbol by doubling
	//			encode == false: $$xx$ -> $xx\n - remove saved symbol 
	//
	enum eEncodeType { eetUndefined = -1, eetDecode = 0, eetEncode = 1 };
	const wchar_t *ChangeSymbols( wchar_t existingSymbol, wchar_t newSymbol, eEncodeType encodeType = eetUndefined );

	const wchar_t *ChangeTabToNewLine()
	{ return ChangeSymbols( L'\t', L'\n' ); }

	const wchar_t *ChangeNewLineToTab();

	//
	//	ChangeWords
	//		Sample:
	//			body = "user_pref('QUOTE'browser.cache.disk.parent_directory'QUOTE'"
	//			presentWord = "'QUOTE'"
	//			newWord		= "\""
	//		Result = "user_pref(\"browser.cache.disk.parent_directory\""
	//
	const wchar_t *ChangeWords( const wchar_t *presentWord, const wchar_t *newWord );

	//
	//	ExtractString
	//		Finds szBegin and szEnd at szSource and set this content as body.
	//
	bool ExtractString( const wchar_t *szSource, const wchar_t *szBegin, const wchar_t *szEnd );

	//
	//	GetString
	//		Walks new line separated strings
	//		Cuts off stringQ by GetFirstString.
	//
	const wchar_t *GetString( const wchar_t *str, int pos, int stringQ = 1 );

	//
	//	GetFirstString
	//		Cuts off first stringQ from new line separated string list.
	//
	const wchar_t *GetFirstString( int stringQ = 1 );

	const wchar_t *SetInt( int i, int radix = 10 );
	int AddInt( int i, int radix = 10 );

	const wchar_t *MakeLower();
	const wchar_t *MakeUpper();
	const wchar_t *MakePath( const wchar_t *dir, const wchar_t *element );
};