/*
 *	Copyright 2014 Jetico Inc. All rights reserved.
 *
 *	File: cWndProcInterceptor.h
 *
 *	Class cWndProcInterceptor declaration.
 *
 *	Description:
 *		Intercepts windows procedure.
 *		1. Function Set()
 *			- Change window procedure to new static and saves 
 *			- Saves the old one.
 *			- Adds this object pointer to static table.
 *		2. New static window procedure:
 *			- Finds object in static table
 *			- Calls it's virtual windows function.
 *			- If this function return false, calls saved original windows procedure.
 *		3. Function Reset()
 *			- Remove object from static table.
 *
 *	Note, this class change windows procedure ( GWL_WNDPROC ) only and does not change another window data values like GWL_USERDATA.
 *
 *	Created:	04.06.2013
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include "AutoDeallocate.h"

class cWndProcInterceptor
{
	WNDPROC m_savedWndProc;

public:
	HWND m_hCtrl;
	
	cWndProcInterceptor()
	{
		m_hCtrl = NULL;
		m_savedWndProc = NULL;
	}

	static cADBlock < cWndProcInterceptor * > &getTable()
	{
		static cADBlock < cWndProcInterceptor * > T ( 0, true );
		return T;
	}

	static int find( HWND hCtrl, cADBlock < cWndProcInterceptor * > &T )
	{
		int i = T.m_Q;

		while( i-- )
		{
			if ( T.m_T[i]->m_hCtrl == hCtrl )
				break;
		}

		return i;
	}

	void addToTable()
	{
		cADBlock < cWndProcInterceptor * > &T = getTable();
		T.Add( this );
	}

	void removeFromTable( )
	{
		cADBlock < cWndProcInterceptor * > &T = getTable();

		int i = find( m_hCtrl, T );

		if ( i != -1 )
			T.RemoveByNumber( i );
	}

	~cWndProcInterceptor()
	{
		if ( m_hCtrl != NULL )
			Detach();

		Free();
	}

	void Free()
	{
		m_hCtrl = NULL;
		m_savedWndProc = NULL;
	}

	void Attach( HWND hCtrl )
	{
		m_hCtrl = hCtrl;
		addToTable();
		m_savedWndProc = (WNDPROC)SetWindowLong( hCtrl, GWL_WNDPROC, (LONG)newWndProc );
	}

	void Detach()
	{
		SetWindowLong( m_hCtrl, GWL_WNDPROC, (LONG)m_savedWndProc );
		removeFromTable();
	}

	static LRESULT newWndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		cADBlock < cWndProcInterceptor * > &T = getTable();
		int i = find( hWnd, T );

		if ( i == -1 )
			return 0;

		cWndProcInterceptor *pThis = dynamic_cast < cWndProcInterceptor * > ( T.m_T[i] );

		if ( pThis == NULL )
			return 0;

		LRESULT lRes = 0;

		if ( pThis->wndProc( &lRes, uMsg, wParam, lParam ) )
			return lRes;

		return CallWindowProc( pThis->m_savedWndProc, hWnd, uMsg, wParam, lParam );
	}

	//
	//	Returns true if message is processed and original procedure call is not required.
	//
	virtual bool wndProc( LRESULT *pResult, UINT uMsg, WPARAM wParam, LPARAM lParam ) = 0;
 };
