/*******************************************************************\
*
*	Copyright (c) 2005-2019 Jetico Inc., Finland
*	All rights reserved.
*
*	File: pwdlg.cpp
*	Revision: 
*	Created:
*	Description:
*		implementation of procedures that support 
*		dialog boxes for accepting user's advanced 
*		key block parameters for
*			1. Password based encryption	- PBE,
*			2. Public key encryption		- PKE,
*			3. Shamir secret shared scheme	- SSS 
*
*	26.08.2014
*	1. Support iteration count other then 256. 
*		See: http://nsk.jetico.com/redmine/projects/best-crypt-for-windo/wiki/Wiki
*
*	02.03.2015
*	Addvanced password settings uses key files.
*
\*******************************************************************/


#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <string.h>

#include <stdio.h>

#include "UnicodeFuncs.h"
#include "cArgs.h"
#include "pwdlg.h"

#include "cDlgKBParams.h"
#include "resource.h"
#include "diskDriver.h"
#include "KgGhost.h"

static bool convertTextToFilenamePath( wchar_t *fileName, wchar_t *pathName, const wchar_t *text )
{
	ODSFuncS( text );

	fileName[0] = L'\0';
	pathName[0] = L'\0';

	if ( FileExistsT( text ) )
	{
		if ( splitPath( text, pathName, fileName ) != 0 )
			return true;
	}

	// input Text has a form: "Container:  D:\path\container_name.jbc"
	// we return: ContainerName = "container_name.jbc", 
	//            ContainerPath = "D:\path\"

	int  i,j;
	const wchar_t *pathPtr = NULL, *namePtr = NULL;

	if (!text) return FALSE;

	for( i = (int)wcslen( text );  ; i-- )
	{
		if ( i <= 0 )
			return false;

		if ((text[i] == L' ') && ( text[i-1] == L':'))
		{
			pathPtr = &(text[i]);
			j = i;
			while (text[j++] == L' ') pathPtr = &(text[j]);
			break;
		}
	}

	if ( !pathPtr )
		pathPtr = text;

	return ( splitPath( pathPtr, pathName, fileName ) != 0 );
}

static bool checkMountPoint(
		wchar_t *mountPoint,
		DWORD   *pFlags
	)
{
	if ( mountPoint == NULL )
		return false;

	//// Get Logical drive letter

	DWORD curLogicalDisks = 0, firstAvailableDrive = 0;

	if ( ( pFlags != NULL ) &&
		 ( ( *pFlags & FLAG_MOUNT_ON_SINGLE_LETTER ) != 0 )
		 )
	{
		// leave the drive letter as it was set
		return TRUE;
	}

	// Ckecks if the mount point is not mounted.
	DISKINFO_BUFFER_V8_0 dBuf;
	if ( !GetVDiskInfo( &dBuf, mountPoint, true ) )
		return true;

	// Otherwise it gets first available logical drive as old versions did

	curLogicalDisks = _GetLogicalDrives();

	for ( int i=2; i<26; i++ )
	{
		if ((curLogicalDisks & (1 << i)) == 0)
		{
			firstAvailableDrive = i;
			break;
		}
	}

	if (!firstAvailableDrive) return FALSE;

	wcsncpy( mountPoint, L"A:\\", MAX_PATH );
	mountPoint[0] = L'A' + (int)firstAvailableDrive;

	return true;
}

static bool getPasswordFromArgs( 
		wchar_t *Password,	// buffer size is 256 symbols
		cArgs &args
	)
{
	///// Try to get password from a command-line: BestCrypt.exe Open G:\30.jbc X: -p:=password=-p

	const wchar_t *value = NULL;

	if ( !args.ReadAndRemoveArgumentWithValue( L"-p:=", &value ) )
		return false;

	int len = wcslen( value );
	if ( len < 3 )
		return false;

	len -= 3;

	if ( wcscmp( value + len, L"=-p" ) != 0 )
		return false;

	if ( len >= 256 )
		return false;

	copyString( Password, value, len, false );
	Password[len] = L'\0';

	return true;
}

bool GetPasswordFromCommandLine(
			wchar_t *Password,	// buffer size is 256 symbols
			DWORD *pFlags,
			cKeyFileList *keyFileList 
	)
{
	
	const wchar_t *cmdLine = GetCommandLineT( true );
	cArgs args( cmdLine );

	// L"  Open <container name> [<mount point>] [-KB <key block file name>] [-KF <key file> ...]\n"

	//
	// Version 2.01.6.2 - 18.12.2017
	//	Bug #3729 Mounting with key-files from command line - wrong password issue
	//	Adds key file from paasword only once at first call ( see bFirstCall ). Otherwise it adds keyfile if dialog is restarted ( f.e. if wrong password was inserted befor )
	//
	// v. 2.01.6.1
	//	Key files are retrieved from command line even if password was not assigned.
	//

	static bool bFirstCall = true;

	if ( bFirstCall && ( keyFileList != NULL ) )
	{
		const wchar_t *path;

		while( args.ReadAndRemoveArgumentWithValue( L"-KF", &path ) )
		{
			keyFileList->m_sa.addString( path );
		}
	}

	bFirstCall = false;

	if ( !getPasswordFromArgs( Password, args ) )
		return false;

	if ( pFlags != NULL )
		*pFlags |= FLAG_CMD_LINE_PASSWORD;

	return true;
}

static bool ods_FindWindow( const wchar_t *name )
{
	ODSFuncS( name );
	HWND hWnd = FindWindowT( name, NULL);

	bool bActive = hWnd != NULL;

	ODSFuncRes( _FormatT( L"%04X, is active = %s", (DWORD)hWnd, getWCStr( bActive ) ) );

	return bActive;
}

bool OnScreenKeyboardActive()
{
	// mhk.sys has IOCTL_PRIVATE_MHK_CONFIRM_KEY_PRESSED and WM_KEYDOWN checks if keyboard drived send a symbol.
	return false;

	ODSFunc;

	//
	//	v.2.01.7.6 - 19.03.2019
	//		The "IPTIP_Main_Window" is always active for all Tablet PC's ( it is main window for 'touch keyboard' )
	//		In this case the touch keyboard according settting appiars automatically as well as activated by double click on edit box.
	//		So user may use both hardware and touch keyboard simultaneousely.
	//		Temporary solution:
	//			we cannot disable keyboard filter here because we don't know what keyboard will be used. Just notify user at 
	//			antykeyloger assigment dialogue to use only hardware keyboard
	//		Solution for future:
	//			if driver will report current number of posted symbols KBD_WM_CHAR may check if driver sends a new one, if not
	//			so the symbol was send by software keyboard and it is not decoded.
	//

	//bool bFirst  = ods_FindWindow( L"IPTIP_Main_Window" ); // 'touch keyboard'
	bool bFirst = false;

	bool bSecond = ods_FindWindow( L"OSKMainClass" );		// 'On-Screen Keyboard'

	bool bActive = false;

	if ( bFirst || bSecond )
		bActive = true;

/*
	if ( FindWindowT( L"IPTIP_Main_Window", NULL) || 
		 FindWindowT( L"OSKMainClass", NULL) ) 
			bActive = true;
*/

	ODSFuncRes( bActive );

	return bActive;
}

DWORD GetKeyBlockParamsWithVerification_Advanced(
		HWND   hWndParent,
		const wchar_t *text,	// Text, containing filename of the container
		const wchar_t *caption,	// Caption of the "get password" dialog window
		cKeyBlockParams *kbParams,
		DWORD  *pFlags,
		DWORD maxFreeSize,
		bool simple // = false
	)
{
	ODSFuncS( text );

	wchar_t fileName[ MAX_PATH ] = L"";
	wchar_t location[ MAX_PATH ] = L"";

	convertTextToFilenamePath( fileName, location, text );

	cDlgKBParams dlg( kbParams, caption, fileName, location  );

	dlg.m_useSecureTyping	= ((*pFlags & FLAG_ADVANCED_SECURE_TYPING) > 0) && !OnScreenKeyboardActive();
	ODS_fls( _FormatT( L"dlg.m_useSecureTyping = %s, *pFlags 0x%X", getWCStr( dlg.m_useSecureTyping ), *pFlags ) );
	dlg.m_keyGenFlags		= *pFlags;

	int idd = simple ? IDD_PasswdDlgV : IDD_KBParamsAdvanced;

	INT_PTR res = dlg.DoModal( IH.m_hInst, idd, hWndParent );

	*pFlags = dlg.m_keyGenFlags;

	if (res == -1) // system error
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	return (DWORD)res;
}

bool bCmdLinePassword = false;

DWORD GetKeyBlockParamsWithVerification(
		dbHeaderBase &header,
		HWND hWnd,
		const wchar_t	*text,
		const wchar_t	*caption,
		cKeyBlockParams *kbParams,
		DWORD *pFlags,
		DWORD maxFreeSize, // = -1, -1 means max available
		bool bDataBlockIsDamaged // = false, see comment at function declaration. 
	)
{
	ODSFuncS( text );

	if ( maxFreeSize == -1 )
		maxFreeSize = DB_maxKeyQ;

	kbParams->kbType = kbtPBE;

	kbParams->kbSSS.Init( maxFreeSize, kbParams->kbPBE.m_algId, kbParams->kbPBE.m_encModeId, kbParams->kbPBE.m_hashAlgId );

	cPassword tmpPwd;
	bCmdLinePassword = GetPasswordFromCommandLine( tmpPwd.wcBuffer, pFlags, &(kbParams->kbPBE.m_keyFileList) );

	if ( bCmdLinePassword )
	{
		kbParams->kbPBE.m_pwd.ReadFromBuffer( tmpPwd.wcBuffer );
		return ERROR_NO;
	}

	DWORD dwRes = GetKeyBlockParamsWithVerification_Advanced( hWnd, text, caption, kbParams, pFlags, maxFreeSize, true );

	while ( dwRes == ERROR_AdvancedDlgRequired )
	{
		if ( !bDataBlockIsDamaged )
		{
			dwRes = GetKeyBlockParamsWithVerification_Advanced( hWnd, text, caption, kbParams, pFlags, maxFreeSize, false );
			break;
		}

		DlgMessageBoxT( hWnd, _LoadIDS( IDS_DBIsDamaged ), MB_ICONEXCLAMATION, _LoadIDS( IDS_Attention ) );
		dwRes = GetKeyBlockParamsWithVerification_Advanced( hWnd, text, caption, kbParams, pFlags, maxFreeSize, true );
	}

	if ( dwRes != ERROR_NO )
		return dwRes;

	return ERROR_NO;
}

static bool isTraveller()
{
	static int iTraveller = -1;

	if ( iTraveller == -1 )
	{
		cWCS wcs ( 1024 );
		int len = GetModuleFileNameT( NULL, wcs.m_szBody, wcs.m_len );
		len = wcslen( wcs.m_szBody );
 
		const wchar_t *szTraveller = L"Traveller.exe";

		const wchar_t *namePos = wcs.m_szBody + len - wcslen( szTraveller );
		if ( _wcsicmp( namePos, szTraveller ) == 0 )
			iTraveller = 1;
		else
			iTraveller = 0;
	}

	return iTraveller == 1;
}

DWORD GetPassword( 
		HWND   hWnd,
		const wchar_t *text,
		const wchar_t *caption,
		cPassword *pwd,
		cKeyFileList *keyFileList,
		DWORD  *pFlags,		// = NULL,
		sCKHParams *ps, //wchar_t *mountPoint// = NULL		
		cWCS *wcsKbb
	)
{
	ODSFunc;

	cPassword tmpPwd;
	bCmdLinePassword = GetPasswordFromCommandLine( tmpPwd.wcBuffer, pFlags, keyFileList ); // mountPoint
	if ( bCmdLinePassword )
	{
		pwd->ReadFromBuffer( tmpPwd.wcBuffer );

		checkMountPoint(  ps->getMountPoint(), pFlags  );

		ODS_fls( L"pwd->ReadFromBuffer ERROR_NO" );
		return ERROR_NO;
	}

	wchar_t fileName[ MAX_PATH ] = L"";
	wchar_t location[ MAX_PATH ] = L"";

	bool converted = convertTextToFilenamePath( fileName, location, text );

	cDlgGetPasswordAndMount dlg( caption, converted ? fileName : text, converted ? location : L"" );

	if ( pFlags != NULL )
	{
		dlg.m_useSecureTyping	= ((*pFlags & FLAG_ADVANCED_SECURE_TYPING) > 0) && !OnScreenKeyboardActive();
		ODS_fls( _FormatT( L"dlg.m_useSecureTyping = %s, *pFlags 0x%X", getWCStr( dlg.m_useSecureTyping ), *pFlags ) );
		dlg.m_keyGenFlags		= *pFlags;
		dlg.m_bTravellerMode	= ((*pFlags & FLAG_TRAVELER_MODE ) != 0 );
	}

	if ( ps != NULL )
	{
		ODS_fl;
		dlg.m_mountPoint = ps->getMountPoint();
		ODS_fl;
	}

	int dlgID = IDD_PasswdDlgSimple;
	
	sCKHParamsAndMountPointEx *psEx = NULL;

	ODS_fls( _FormatT( L"dlg.m_mountPoint %X, %s", dlg.m_mountPoint, dlg.m_mountPoint ) );

	if ( dlg.m_mountPoint != NULL )
	{
		//
		// Version 2.01.7.4 - 24.10.2018
		//	Checks ps == NULL again, but if dlg.m_mountPoint != NULL ps has to be not NULL too.
		//
		if ( ( ps == NULL  ) || ( ps->size < sizeof( sCKHParamsAndMountPointEx ) ) )
		{ assert( false ); return ERROR_INTERNAL_PROBLEM; }

		psEx = (sCKHParamsAndMountPointEx *)ps;

		dlg.m_bAdvanced = ( ( psEx->flags & sCKHParamsAndMountPointEx::eFlagAdvancedMode ) != 0 );;
		//dlgID = dlg.m_bAdvanced ? IDD_PasswdDlgMount_Advanced : IDD_PasswdDlgMount;
		dlgID = IDD_PasswdDlgMount_Advanced;
		if ( isTraveller() )
			dlgID = IDD_PasswdDlgMount_Advanced_Traveller;

		dlg.m_bMountForCurrentUser = psEx->getFlagMountForCurrentUser();
		dlg.m_bMountAsRemovable = ( ( psEx->flags & sCKHParamsAndMountPointEx::eFlagMountAsRemovable ) != 0 );
		
		if ( keyFileList != NULL )
			dlg.m_keyFileList.Set( *keyFileList );

		if ( wcsKbb!= NULL )
			wcsKbb->Swap( &dlg.m_kbbPath );
	}
	else
	if ( converted ) // The text has no valid path and container location
		dlgID = IDD_PasswdDlgSimple;
	else
		dlgID = IDD_PasswdDlgText; // The dialog has text only
	
	ODS_fls( _FormatT( L"dlg.m_useSecureTyping = %s", getWCStr( dlg.m_useSecureTyping ) ) );

	INT_PTR res = dlg.DoModal( IH.m_hInst, dlgID, hWnd );

	if ( pFlags != NULL )
		*pFlags = dlg.m_keyGenFlags;

	if (res == -1) // system error
	{
		ODS_fls( L"Assertion! ERROR_INTERNAL_PROBLEM" );
		return ERROR_INTERNAL_PROBLEM;
	}

	pwd->ReadFromBuffer( dlg.m_pwd.wcBuffer, CP_ACP );

	if  ( keyFileList != NULL )
		keyFileList->Set( dlg.m_keyFileList );

	if ( psEx != NULL )
	{
		psEx->setFlag( sCKHParamsAndMountPointEx::eFlagMountForCurrentUser, dlg.m_bMountForCurrentUser );
		psEx->setFlag( sCKHParamsAndMountPointEx::eFlagMountAsRemovable, dlg.m_bMountAsRemovable );
		psEx->setFlag( sCKHParamsAndMountPointEx::eFlagAdvancedMode, dlg.m_bAdvanced );
	}

	if ( wcsKbb != NULL )
		wcsKbb->Swap( &dlg.m_kbbPath );

	ODSFuncRes( res );
	return (DWORD)res;
}

DWORD GetKeyBlockParams(
		HWND hWnd,
		const wchar_t	*text,
		const wchar_t	*caption,
		cKeyBlockParams *kbParams,
		DWORD *pFlags,
		sCKHParams *ps, // wchar_t *mountPoint		
		cWCS *wcsKbb
	)
{
//	kbParams->kbType = kbtPBE;
	return GetPassword( hWnd, text, caption, &( kbParams->kbPBE.m_pwd ), &(kbParams->kbPBE.m_keyFileList), pFlags, ps, wcsKbb ); // mountPoint );
}

DWORD GetAdvancedFlags( DWORD *pFlags, HWND  hWndParent )
{
	cDlgGetAdvancedFlags dlg( true );

	dlg.m_useSecureTyping	= TRUE;
	dlg.m_keyGenFlags		= *pFlags;

	dlg.SetRCValues( IDC_ColorLine );

	INT_PTR res = dlg.DoModal( IH.m_hInst, IDD_SetFlags, hWndParent );

	if (res == -1) // system error
	{
		int err = GetLastError();
		ODS_fls( _FormatT( L"Assertioin! %s ", _GetSystemError( err ) ) );
		return ERROR_INTERNAL_PROBLEM;
	}

	*pFlags = dlg.m_keyGenFlags;

	return (DWORD)res;
}