/*******************************************************************\
*
*	Copyright (c) 2006-2013 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: cKeyBlockParams.h
*	Revision: 
*	Created:		6.10.2004
*	Description:	The key block is used for getting new and existing key parameters
*					It works with three different types of key parameters.
*						1. Password based encryption
*						2. Public key encryption
*						3. Shamir secret shared scheme
*
\*******************************************************************/

#pragma once

#include "..\..\..\CertificateLib\publicKeyAlgs\Power\Windefs.h"
#include "..\..\..\CertificateLib\SMemory\SBlocks.h"
#include "UnicodeFuncs.h"
#include "..\..\..\CertificateLib\SMemory\cPassword.h"
#include "KBlock.h"
#include "..\..\Comm_def\Kg_defs.h"
#include "..\..\..\CertificateLib\PGPPkts\PKTDEFS.H"
#include "..\..\..\BCKeyMan\KM_Funcs.h"
#include "alg.h"
#include "cKeyFileList.h"

//
//	Password Based Encryption parameters
//

class cKeyBlockParams;

struct sEncIDs
{
	DWORD m_algId;	// algorithm for key encryption

	// Encryption mode see bc_ioctl.h
	// #define EX_ENCRYPT_LRW_MODE_FLAG  0xBC000001
	// #define EX_ENCRYPT_CBC_MODE_FLAG  0xBC000002
	DWORD m_encModeId;

	DWORD m_hashAlgId;

	void SetEncIds( sEncIDs *ps = NULL )
	{
		if ( ps != NULL )
		{
			memcpy( this, ps, sizeof( sEncIDs ) );
			return;
		}

		m_algId		= -1;
		m_encModeId = -1;
		m_hashAlgId = -1;
	}
};

class cParamsPBE : public sEncIDs
{
public:

	cPassword m_pwd;
	cPassword m_pwdConf; // Just to save confirmard it when move to advanced view.

	//
	// Key generator V5 and BEstCrypt V9 new features.
	//

	//
	//	Old version of KG-Ghost supports iteration count 256 only.
	//
	int m_iterations;

	cKeyFileList m_keyFileList;

	cPTB m_salt;
	cPwdKeyV5 m_pwdKey;
	enum eDecodeType { eDecodeType_Undefined = -1, eDecodeType_V4, eDecodeType_V5  } m_decodeType;

	int setSalt( BYTE *bf, int bfSize );
	int setSaltSize( int size = 0 ); // if size is 0 it is assigned as hash algorithm digest size.
	int getSaltSize();
	bool isKeyGenV5();
	bool setBCV9Default();
	bool setAsBCV9();

	cKeyBlockParams *m_pKBParams;

	cParamsPBE();
	~cParamsPBE();

	//
	//	decodeKey
	//		Gets key from password with salt and decrypts session key to keyBuff
	//		kbEncrypted consists of 1. random salt and 2. encrypted session key, see PBE_KeyBlock
	//
	bool decodeKey(
				BYTE *keyBuff, int keyBuffSize,
				BYTE *kbEncrypted, int kbEncryptedSize,
				int algBlockSize
			);

	//
	//	Gets key from password with salt and encrypts key to encrypted key part of kbEncrypted
	//	kbEncrypted consists of :
	//		1. random salt
	//		2. encrypted key, see PBE_KeyBlock
	//
	bool encodeKey( BYTE *kbEncrypted, int kbEncryptedSize, BYTE *keyBuff, int keyBuffSize );

	//
	//	generateKeyFromPassword
	//		Uses P12_producePseudoRandomBytes functions for key generation.
	//
	static bool generateKeyFromPassword(
				BYTE *outBuff, int outBuffSize, 
				cPassword *pwd,
				DWORD hashAlgId = bchaSHA256,
				BYTE *salt = NULL, int saltSize = 0,
				int iteration = 1,
				bool key = true // key or IV
			);

	//
	//	updateKey
	//		1. Forms key and IV from password
	//		2. Encrypts ( or decrypts ) inBuff to outBuff
	//
	bool encryptKey( PBE_KeyBlock *encOut, sKeyInfo *keyInfo );
	bool decryptKey( sKeyInfo *keyInfo, PBE_KeyBlock *encIn );
	bool encryptKey( PBE_KeyBlock_V5 *encOut, sKeyInfo *keyInfo );
	bool decryptKey( sKeyInfo *keyInfo, PBE_KeyBlock_V5 *encIn );

	//
	//	getEncodedKeySize
	//
	int getEncodedKeySize( int keyLength );

	bool updateSalt();
};

typedef class cPKA * HPKA; // Handle to public/private key algorithm

//
//	Public Key Encryption parameters
//
class cParamsPKE
{
public:
	HPKA pka, ska;

	cPTB ptbKey;
	cPTB ptbKeyID;
	type_kff kff;

	bool bCancelled;

	cParamsPKE()
	{
		pka = NULL;
		ska = NULL;
		kff = kffNone;

		bCancelled = false;
	}

	~cParamsPKE();

	bool encodeKey( BYTE *kbEncrypted, int kbEncryptedSize, BYTE *keyBuff, int keyBuffSize );
	bool decodeKey( BYTE *keyBuff, int keyBuffSize, BYTE *kbEncrypted, int kbEncryptedSize );
	int getEncodedKeySize( int keyLength );
	bool openSecretKey( HWND hWnd, cPassword *pwd, BYTE *keyIDSBuffer = NULL, bool checkPwdOnly = false );
};

//
//	SSS table item.
//
struct sssPwdItem
{
	int numberOfShares;
	cPassword pwd;
};

//
//	Shamir Secret Shared scheme parameters
//
class cParamsSSS : public sEncIDs
{
	cPTB ptbTab;

public:

	enum { defaultSSSVersion = 2 }; // see sShareV2 and sShareV1 differences

	cPTB ptbSeed;
	DWORD		*pErrorCode;
	DWORD		*pFlags;

	HWND		hWnd;

	sssPwdItem	*itemT;
	int			itemQ;
	BYTE		version;
	int			reqQ;			// Number of different shares are required for key recovering 
								// ( polinom of the degree is generated )

	DWORD		maxBufferSize;	// max encrypted buffer size, is required for generate shares 
								// in encode process

	int			encodedSize;	// real encoded size, it is filled when information about shared scheme was once decoded ( see _sShareV2::totalQ ).

	cParamsSSS()
	{
		memset( this, 0, sizeof( *this ) );
		version = (BYTE)-1;
		m_algId	= -1;
		m_encModeId = -1;
		m_hashAlgId = -1;
		reqQ	= -1;
		maxBufferSize = 0;
	}

	void SetEncIds( sEncIDs *ps = NULL, BYTE pVersion = (BYTE)-1 )
	{
		sEncIDs::SetEncIds( ps );

		// Version 2.01.5.2 - 2.11.2016
		//	Bug #3043 SSS: undefined key generator error
		//	version depends of hash Id if HASH block size is 64 bytes it is v.3
		if ( pVersion == (BYTE)-1 )
			pVersion = getDefultVersion( m_hashAlgId );

		version = pVersion;
	}

	void Init( DWORD pMaxBufferSize, DWORD pAlgId, DWORD pEncModeId, DWORD pHashAlgId, BYTE pVersion = (BYTE)-1 ) // defaultSSSVersion
	{
		if ( pVersion == (BYTE)-1 )
			pVersion = getDefultVersion( pHashAlgId );

		maxBufferSize = pMaxBufferSize;
		version = pVersion;
		m_algId = pAlgId;
		m_encModeId = pEncModeId;
		m_hashAlgId = pHashAlgId;
		reqQ = 2;
	}

	~cParamsSSS()
	{
		Free();
	}

	void Free()
	{
		while( itemQ-- )
			itemT[ itemQ ].pwd.ClearBuffers();

		//memset( this, 0, sizeof( *this ) );
	}

	//
	//	getTotalSharesQ
	//		walks table of password and count total shares
	//		returns total shares currently assigned
	//
	int getTotalSharesQ();

	//
	//	getMaxSharesQ
	//		Checks encode buffer size and returns number of maximum shares quantity could be filled.
	//
	int getMaxSharesQ();

	//
	//	getSeedSize
	//		calkulates size of required seed, the seed may be generated togather with common seed generation
	//
	int getSeedSize( int keyBuffSize );

	//
	// Checks if all required shares are assigned.
	//
	bool isReady( int *pReadyQ = NULL );

	//
	//	All added passwords have to be deleted at the class destructor.
	//
	bool addShare( int numberOfShares, cPassword *pwd );

	//
	//	Frees table and zeroes counts
	//
	bool resetAll();

	//
	//	checkPwdIsDifferent
	//		Checks if any previously added password is not same with the newPwd
	//
	bool checkPwdIsDifferent( cPassword *newPwd );

	bool encodeKey( BYTE *kbEncrypted, int kbEncryptedSize, BYTE *keyBuff, int keyBuffSize );
	bool decodeKey( BYTE *keyBuff, int keyBuffSize, BYTE *kbEncrypted, int kbEncryptedSize );
	int getEncodedKeySize( int keyLength );

	//
	//	cParamsSSS::decodeKey
	//		1. Tryes first password, 
	//		2. If first pwd is compatible, it asks passwords while required shares will not collected
	//
	bool decodeKeyWithFirstPwd( cPassword *firstPwd, BYTE *keyBuff, int keyBuffSize, BYTE *kbEncrypted, int *kbEncryptedSize, bool checkPasswordOnly );

	static int getDefultVersion( DWORD hashAlgId );
};

//
//	Class key block params used by GetKeyBlockParamsWithVerification to get new key parameters
//		and by GetKeyBlockParams to get existing key parameters without verification.
//

struct sKeyInfo;

class cKeyBlockParams
{
public:

	//BYTE	hashAlgId;	// Digest algorithm used for key block verification ( same as pgp id ( f.e. pgphaSHA1 see pktdefs.h )
	//DWORD	keyAlgId;	// Algorithm used for container and key block encryption same as BestCrypt Service Manager defines and algorithm driver uses
	//DWORD	encModeId;	// Encryption mode see bc_ioctl.h
	//					// #define EX_ENCRYPT_LRW_MODE_FLAG  0xBC000001
	//					// #define EX_ENCRYPT_CBC_MODE_FLAG  0xBC000002


	cSymAlgorithm algInfo;

	//
	//	We may create new key block with a password,
	//	it is useful to check if the password is compatible with a key block.
	//	See DataBlockCheckPwdIsDifferent.
	//
	DATA_BLOCK *dataBlock;
	DWORD dataBlockSize;

	bool KeyBlockParamsInit(
			DATA_BLOCK *dataBlockArray,
			DWORD dataBlockArraySizeInBytes,
			int		contAlgId	= -1,
			DWORD	encMode		= -1,		
			int		contHashAlgId = -1,
			int		iterations	= -1,
			DWORD	*pErrorCode	= NULL,
			DWORD	*pFlags		= NULL,
			HWND	hWndParent	= NULL
		);

	ekbType kbType; // Determinates what class is vallidly filled.

	cParamsPBE kbPBE;
	cParamsPKE kbPKE;
	cParamsSSS kbSSS;

	void setHashAlgId( DWORD id )
	{
		ODSFunc;
		kbPBE.m_hashAlgId = id;
		kbSSS.m_hashAlgId = id;
		kbPBE.setSaltSize();
	}

	void setBCV9Default()
	{
		kbPBE.setBCV9Default();
		kbSSS.m_hashAlgId = kbPBE.m_hashAlgId;
	}


	//
	//	The new data block creating allows to change algorithms ids.
	//	If new password is added of hidden part is created the algorithm id cannot be changed.
	//
	bool m_allowChangeAlgorithms;

	cKeyBlockParams();
	~cKeyBlockParams();

	//
	//	Calls decodeKey function for corresponded block type.
	//
	bool decodeKey( sKeyInfo *keyBuff, int keyBuffSize, BYTE *kbEncoded, int kbEncodedSize, bool checkPasswordOnly = false );

	//
	//	Calls encodeKey function for corresponded block type.
	//
	bool encodeKey( BYTE *kbEncoded, int kbEncodedSize, sKeyInfo *keyBuff, int keyBuffSize );

	//
	//	Calls getEncodedKeySize function for corresponded block type.
	//
	int getEncodedKeySize( int keyLength );

	//
	//	Tries to get a password from key block parameters.
	//	The password is required as client network password for corporate version.
	//
	bool getPassword( char *pwdBuff );
};