/*******************************************************************\
*
*	Copyright (c) 2015 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: cKeyFileList.cpp
*	Revision: 
*	Created:		03.03.2015
*	Description:	Class with list of key files.
*
\*******************************************************************/

#include "cKeyFileList.h"
#include "..\..\..\Includes\FileView.h"
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"

//
//	getFileQ
//		returns 1 if path is file, otherwise number of files in the directory
//
//static
int cKeyFileList::processPath( const wchar_t *path, cProcessPath *process )
{
	WIN32_FIND_DATAW fd;

	if ( !FileExistsT( path, &fd ) )
		return 0;

	if ( ( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
	{
		if ( process == NULL )
			return 1;

		int res = process->actor( path );

		if ( res < 0 )
			process->showError( path );

		return res;
	}

	cFindFile ff( path );
	int Q = 0;

	if ( ff.findFirstFile() )
	{
		do
		{
			if ( !ff.isDir() )
			{
				int res = process ? process->actor( ff.getCurrentPath() ) : 1;
				if ( res < 0 )
				{
					if ( process != NULL )
						process->showError( ff.getCurrentPath() );

					return res;
				}

				Q += res;
			}
		} 
		while( ff.findNextFile() );
	}

	return Q;
}

int cKeyFileList::processPathes( cProcessPath *process )
{
	int Q = 0;

	for( int i = 0; i < m_sa.Q; i++ )
	{
		int res = processPath( m_sa.T[i], process );

		if ( res < 0 )
			return res;

		Q += res;
	}

	return Q;
}

class cGetSize : public cProcessPath
{
public:
	DWORDLONG m_dwlSize;

	cGetSize()
	{
		m_dwlSize = 0;
	}

	//
	//	Allocates required digest size at 
	//
	virtual int actor( const wchar_t *path )
	{
		DWORDLONG dwl = GetFileSizeT( path, &m_sysErr );

		if ( dwl == (DWORDLONG)-1 )
			return -1;

		m_dwlSize += dwl;

		return 1;
	}
};

//static 
int cKeyFileList::getFileQ( const wchar_t *path, DWORDLONG *pSize )
{
	cGetSize cgs;
	
	int res = processPath( path, &cgs );

	if ( pSize != NULL )
		*pSize = cgs.m_dwlSize;

	return res;
}

int cKeyFileList::getFileQ( DWORDLONG *pSize )
{
	cGetSize cgs;
	
	int res = processPathes( &cgs );

	if ( pSize != NULL )
		*pSize = cgs.m_dwlSize;

	return res;
}


//static 
int cKeyFileList::makeFileDigest( int hashAlgId, BYTE *digest, const wchar_t *path, DWORD maxSize ) // = 1024 * 1024 )
{
	cFileView fv( path, true, maxSize );
	if ( fv.m_err != cFileView::cfv_ok )
		return fv.m_sysErr;

	if ( fv.m_size.QuadPart == 0 )
		return 0;

	DWORDLONG size = min( maxSize, fv.m_size.QuadPart );

	int res = ha_makeDigestEx( hashAlgId, digest, fv.m_ptr, (DWORD)size );
	if ( res != hae_ok )
	{ assert( false ); return ERROR_INVALID_DATA; } // res is not reported as system error.

	return 0;
}

static int compare( BYTE *bf, int p0, int p1, int size )
{
	return memcmp( bf + p0 * size, bf + p1 * size, size );
}

static bool swap( BYTE *bf, int p0, int p1, int size )
{
	BYTE * p = new BYTE[ size ];
	if ( p == NULL )
	{ assert( false ); return false; }

	memcpy( p, bf + p0 * size, size );
	memcpy( bf + p0 * size, bf + p1 * size, size );
	memcpy( bf + p1 * size, p, size );

	return true;
}

static bool sortBuffer( BYTE *bf, int bfSize, int sortSize )
{
	if ( ( bfSize % sortSize ) != 0 )
	{ assert( false ); return false; }
	
	// Booble sort..
	for( int q = bfSize / sortSize; q > 1; q-- )
	{
		for( int i = 0; i < q - 1; i++ )
		{
			if ( compare( bf, i, i + 1, sortSize ) > 0 )
				if ( !swap( bf, i, i + 1, sortSize ) )
					return false;
		}
	}

	return true;
}

//
//	fillHashTable
//		1. Walks files and gets it hashes upto 1M
//		2. Sorts prepared buffer.
//	Note, sorting exclude invalid order of key files.
//
bool cKeyFileList::fillHashTable( cPTB &ptbHashes, int hashAlgId )
{
	class cFillHashes : public cProcessPath
	{
		cPTB &m_ptbHashes;
		int m_hashAlgId;

	public:
		int m_digestSize;

		cFillHashes( cPTB &ptbHashes, int hashAlgId )
			: m_ptbHashes( ptbHashes )
		{
			m_hashAlgId = -1;
			m_digestSize = 0;
			if ( ha_getDigestSize( &m_digestSize, hashAlgId ) != hae_ok )
			{ assert( false );	return; }

			m_hashAlgId = hashAlgId;
		}

		//
		//	Allocates additional digest at m_ptbHashes
		//	Calculates the file digest.
		//
		virtual int actor( const wchar_t *path )
		{
			if ( ( m_hashAlgId == -1 ) || ( m_digestSize == 0 ) )
			{ assert( false );	return -1; }
			
			BYTE *digest = m_ptbHashes.AddBytes( NULL, m_digestSize );

			if ( digest == NULL )
			{ assert( false );	return -1; }

			m_sysErr = makeFileDigest( m_hashAlgId, digest, path );

			if ( m_sysErr != 0 )
				return -1;

			return 1;
		}
	};
	
	cFillHashes fh( ptbHashes, hashAlgId );
	
	if ( processPathes( &fh ) == -1 )
		return false;

	return sortBuffer( ptbHashes.buffer, ptbHashes.bufferSize, fh.m_digestSize );
}

int cKeyFileList::updateWindowText( HWND hWnd )
{
	if ( m_pattern.isEmpty() )
		m_pattern.GetWindowTextT( hWnd );

	int Q = getFileQ();

	SetWindowTextT( hWnd, _FormatT( m_pattern.m_szBody, Q ) );
	return Q;
}

//
//	generateSaltBuffer
//		1. Walks files and gets it hashes upto 1M
//		2. Sorts prepared buffer. Note, sorting exclude invalid order of key files.
//		3. Adds real salt to the hash buffer tail.
//
bool cKeyFileList::generateSaltBuffer( int hashAlgId, BYTE *saltBuffer, int saltSize )
{
	m_ptbSalt.Free();

	if ( !fillHashTable( m_ptbSalt, hashAlgId ) )
		return false;

	if ( m_ptbSalt.WriteBytes( saltBuffer, saltSize ) != saltSize )
	{ assert( false ); return false; }

	return true;
}
