/*******************************************************************\
*
*	Copyright (c) 2004-2014 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: cParamsPBE.h
*	Revision: 
*	Created:		21.10.2004
*	Description:	Password based encryption parameters
*						2. Public key encryption
*						3. Shamir secret shared scheme
*
\*******************************************************************/

#include "ods.h"
#include "cKeyBlockParams.h"
#include "..\..\..\CertificateLib\PGPPkts\PKTDEFS.H"
#include "..\..\..\CertificateLib\x509pkts\pkcs12_funcs.h"
//#include "..\..\..\CertificateLib\symmetricAlgs\symmetricAlgs.h"
#include "Alg.h"
#include <assert.h>
#include "..\..\..\CertificateLib\PGPPkts\string2key.h"
#include "KgGhost.h"

//
// Key generator V5 and BEstCrypt V9 new features.
//

int cParamsPBE::setSalt( BYTE *bf, int bfSize )
{
#ifdef _ODS
	ODSFuncS( _FormatT( L"%X, %X, %d", this, bf, bfSize ) );

	if ( bfSize != 0 && ( bf[0] == 0 ) && ( bf[1] == 0 ) )
		ODS_fls( _FormatT( L"Zero salt, bfSize %d", bfSize ) );
#endif

	if ( m_hashAlgId == (DWORD)-1 )
		m_salt.Allocate( bfSize );
	else
	{
		int digestSize;

		if ( ha_getDigestSize( &digestSize, m_hashAlgId ) != hae_ok )
		{ assert( false ); return 0; }

		if ( bfSize < digestSize )
		{ assert( false ); return 0; }

		if ( setSaltSize( digestSize ) == 0 )
		{ assert( false ); return 0; }
	}

	memcpy( m_salt.buffer, bf, getSaltSize() );
	return getSaltSize();
}

int cParamsPBE::setSaltSize( int size )
{
	ODSFuncS( _FormatT( L"%X, size", this, size )  );
#ifdef _ODS
	int sizeIn = size;
#endif

	if ( ( size == 0 ) && ( ha_getDigestSize( &size, m_hashAlgId ) != hae_ok ) )
	{
		assert( false );
		ODS_fls( _FormatT( L"Invalid hash algorithm %d", m_hashAlgId ) );
		return 0;
	}
	
	bool regenerateSalt = ( m_salt.allocatedSize == 0 );

	if ( m_salt.allocatedSize < size )
	{
		if ( m_salt.ReAllocate( size ) == NULL )
			return 0;
	}

	regenerateSalt |= ( m_salt.bufferSize < size );

	m_salt.bufferSize = size;

	if ( regenerateSalt )
		updateSalt();

#ifdef _ODS
	if ( ( m_salt.buffer[0] == 0 ) && ( m_salt.buffer[1] == 0 ) )
		ODS_fls( _FormatT( L"Zero salt, sizeIn %d, size %d, hash %d, regenerateSalt %s", sizeIn, size, m_hashAlgId, getWCStr(regenerateSalt) ) );
#endif

	return size;
}

int cParamsPBE::getSaltSize()
{
	return m_salt.bufferSize;
}

bool cParamsPBE::isKeyGenV5()
{
	if ( m_iterations != 256 )			return true;
	if ( getSaltSize() != SALT_SIZE )	return true;
	if ( !m_keyFileList.m_sa.isEmpty() )	return true;

	return false;
}

bool cParamsPBE::setAsBCV9()
{
	m_iterations = DEFAULT_ITERATIONS_V5;

	if ( setSaltSize( 0 ) == 0 )
		return false;

	updateSalt();
	return true;
}

bool cParamsPBE::setBCV9Default()
{
	ODSFunc;
//	m_algId = bcsaRIJN;	// algorithm for key encryption
	m_iterations = DEFAULT_ITERATIONS_V5;
	m_hashAlgId = bchaWhirlpool512;

	if ( setSaltSize( 0 ) == 0 )
		return false;

	updateSalt();
	return true;
}

cParamsPBE::cParamsPBE()
{
	ODSFuncS( _FormatT( L"%X", this )  );
//	m_algId = pgpsaBlowfish;	// algorithm for key encryption
	m_iterations = DEFAULT_ITERATIONS_V4;
	m_hashAlgId = bchaSHA256;
	setSaltSize( SALT_SIZE );
	updateSalt();
	m_decodeType = eDecodeType_Undefined;
}

cParamsPBE::~cParamsPBE()
{
	ODSFunc;

	m_salt.Free();
	m_pwdKey.Free();
	m_pwdConf.ClearBuffers();
	m_pwd.ClearBuffers();

	//memset( this, 0, sizeof( *this ) );
}

bool cParamsPBE::updateSalt()
{
	int digestSize;
	ha_getDigestSize( &digestSize, m_hashAlgId );

	cPTB bf;
	if ( bf.Allocate( digestSize * 3 ) == NULL )
	{
		ODS_fls( _FormatT( L"Assertion! Error in memory allocation." ) );
		assert( false );
		return false;
	}

	ShredData( bf.buffer + digestSize, bf.bufferSize - digestSize );

	if ( hae_ok != ha_makeDigestEx( m_hashAlgId, bf.buffer, bf.buffer + digestSize, bf.bufferSize - digestSize ) )
	{
		ODS_fls( _FormatT( L"Assertion! Invalid algorithm %d", m_hashAlgId ) );
		assert( false );
		return false;
	}

	int saltSize = getSaltSize();
	memset( m_salt.buffer, 0, saltSize );

	for( int i = saltSize; i--; )
		m_salt.buffer[i] ^= bf.buffer[i];

	return true;
}

//
//	generateKeyFromPassword
//		Uses P12_producePseudoRandomBytes functions for key generation.
//
bool cParamsPBE::generateKeyFromPassword(
		BYTE *outBuff,	int outBuffSize,
		cPassword *pwd,
		DWORD hashAlgId, // = bchaSHA256,
		BYTE *salt,		int saltSize, // NULL
		int iteration, // == 1
		bool key // key or IV
	)
{
	return pwd->generateKeyFromPassword(
		outBuff, outBuffSize,
		(BYTE)hashAlgId,
		salt, saltSize,
		iteration,
		key );
}

bool cParamsPBE::encryptKey( PBE_KeyBlock_V5 *encOut, sKeyInfo *keyInfo )
{
	if ( getSaltSize() == SALT_SIZE )
	{ assert( false ); return false; }

	BYTE *saltBuffer =  m_salt.buffer;
	int saltSize = getSaltSize();

	if ( m_keyFileList.generateSaltBuffer( m_hashAlgId, m_salt.buffer, getSaltSize( ) ) )
	{
		saltBuffer = m_keyFileList.getSaltBuffer();
		saltSize = m_keyFileList.getSaltSize();
	}
	else
	{ assert( false ); return false; }

	cPwdAlgorithm cAlg( m_algId, m_encModeId, &m_pwd, m_hashAlgId, saltBuffer, saltSize, m_iterations );
	if ( cAlg.m_err )
		return false;
	
	return encOut->keyInfo.encryptFromBody( keyInfo, m_algId, m_encModeId, m_hashAlgId, cAlg.m_keyHandle );
}


bool cParamsPBE::decryptKey( sKeyInfo *keyInfo, PBE_KeyBlock_V5 *encIn )
{
	if ( getSaltSize() == SALT_SIZE )
	{ assert( false ); return false; }

	cPwdAlgorithm cAlg;

	cAlg.m_pwdKey = &m_pwdKey;

	BYTE *saltBuffer =  m_salt.buffer;
	int saltSize = getSaltSize();

	if ( m_keyFileList.generateSaltBuffer( m_hashAlgId, m_salt.buffer, getSaltSize( ) ) )
	{
		saltBuffer = m_keyFileList.getSaltBuffer();
		saltSize = m_keyFileList.getSaltSize();
	}
	else
	{
		*m_pKBParams->kbSSS.pErrorCode = ERROR_InReadingKeyFiles;
		return false;
	}

	if ( !cAlg.generateKey( m_algId, m_encModeId, &m_pwd, m_hashAlgId, saltBuffer, saltSize, m_iterations ) )
	{
		assert( false );
		return false;
	}

	if ( encIn->keyInfo.decryptBodyTo( keyInfo, m_algId, m_encModeId, cAlg.m_keyHandle ) )
		return true;

	return false;
}

bool cParamsPBE::encryptKey( PBE_KeyBlock *encOut, sKeyInfo *keyInfo )
{
	if ( getSaltSize() != SALT_SIZE )
	{ assert( false ); return false; }

	memcpy( encOut->salt, m_salt.buffer, sizeof( encOut->salt ) );

	cPwdAlgorithm cAlg( m_algId, m_encModeId, &m_pwd, m_hashAlgId, encOut->salt, sizeof( encOut->salt ), m_iterations );
	if ( cAlg.m_err )
		return false;
	
	return encOut->keyInfo.encryptFromBody( keyInfo, m_algId, m_encModeId, m_hashAlgId, cAlg.m_keyHandle );
}


bool cParamsPBE::decryptKey( sKeyInfo *keyInfo, PBE_KeyBlock *encIn )
{
	cPwdAlgorithm cAlg( m_algId, m_encModeId, &m_pwd, m_hashAlgId, encIn->salt, sizeof( encIn->salt ), m_iterations );
	if ( cAlg.m_err )
	{
		assert( false );
		return false;
	}

	if ( encIn->keyInfo.decryptBodyTo( keyInfo, m_algId, m_encModeId, cAlg.m_keyHandle ) )
		return true;

	return false;
}

//
//	Gets key from password with salt and decrypts key to keyBuff
//	kbEncoded consists of :
//		1. random salt
//		2. encrypted key, see PBE_KeyBlock
//
bool cParamsPBE::decodeKey( BYTE *keyBuff, int keyBuffSize, BYTE *kbEncoded, int kbEncodedSize, int algBlockSize )
{
	if ( kbEncodedSize < ( sizeof( PBE_KeyBlock_V5 ) - tail32tail64diff ) )
	{
		//assert( false ); salt slot may be shorter
		return false;
	}

	if ( keyBuffSize < sizeof( sKeyInfo ) )
	{
		assert( false );
		return false;
	}

	if ( m_decodeType == eDecodeType_V5 )
		return decryptKey( (sKeyInfo *)keyBuff, (PBE_KeyBlock_V5 *)kbEncoded );
	else
	if ( m_decodeType == eDecodeType_V4 )
		return decryptKey( (sKeyInfo *)keyBuff, (PBE_KeyBlock *)kbEncoded );
	else
	if ( getSaltSize() == SALT_SIZE )
		return decryptKey( (sKeyInfo *)keyBuff, (PBE_KeyBlock *)kbEncoded );
	else
	{
		if ( decryptKey( (sKeyInfo *)keyBuff, (PBE_KeyBlock_V5 *)kbEncoded ) )
			return true;
		else
			return decryptKey( (sKeyInfo *)keyBuff, (PBE_KeyBlock *)kbEncoded );
	}
}

//
//	Gets key from password with salt and encrypts key to encrypted key part of kbEncoded
//	kbEncoded consists of :
//		1. random salt
//		2. encrypted key, see PBE_KeyBlock
//
bool cParamsPBE::encodeKey( BYTE *kbEncoded, int kbEncodedSize, BYTE *keyBuff, int keyBuffSize )
{
	if ( kbEncodedSize < getEncodedKeySize( -1 ) )
		return false;

	if ( keyBuffSize < sizeof( sKeyInfo ) )
		return false;

	//
	//	If key fileas are used at data block creation so db version has to be upgraded to V5 automatically 
	//
	if ( getSaltSize() == SALT_SIZE )
		return encryptKey( (PBE_KeyBlock *)kbEncoded, (sKeyInfo *)keyBuff );
	else
		return encryptKey( (PBE_KeyBlock_V5 *)kbEncoded, (sKeyInfo *)keyBuff );
}

//
//	getEncodedKeySize
//		returns size of buffer required for key encoding.
//
int cParamsPBE::getEncodedKeySize( int keyLength )
{
	int digetSize;
	if ( ha_getDigestSize( &digetSize, m_hashAlgId ) != hae_ok )
	{
		assert( false );
		return -1;
	}

	int keySize;

	if ( getSaltSize() == SALT_SIZE )
		keySize = sizeof ( PBE_KeyBlock );
	else
		keySize = sizeof ( PBE_KeyBlock_V5 );

	if ( digetSize <= 32 )
		keySize -= tail32tail64diff;

	return keySize;
}
