#include "cDlg.h"
#include "cKeyBlockParams.h"
#include "..\..\..\CertificateLib\SSS\Shamir\shamir.h"
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
#include "ALG.H"
#include "..\..\Comm_def\GenerateSeed.h"
#include "resource.h"
#include "Pwdlg.h"

static int getTotalSharesQ( 
	sssPwdItem	*itemT,	// Table with passwords and heghts of holder
	int			itemQ	// Number of items in the table
	)
{
	int totalSharesQ = 0;

	for( int i = 0; i < itemQ; i++ )
		totalSharesQ += itemT[i].numberOfShares;

	return totalSharesQ;
}

int cParamsSSS::getTotalSharesQ()
{
	return ::getTotalSharesQ( itemT, itemQ );
}

bool cParamsSSS::isReady( int *pReadyQ )
{
	int shareCount = getTotalSharesQ();

	assert( reqQ >= 2 );

	if ( pReadyQ != NULL )
		*pReadyQ = shareCount;

	if ( shareCount >= reqQ )
		return true;

	return false;
}

bool cParamsSSS::checkPwdIsDifferent( cPassword *pwd )
{
	for ( int i = itemQ; i--; )
	{
		if ( pwd->Compare( &itemT[i].pwd ) == 0 )
			return false;
	}

	return true;
}

bool cParamsSSS::resetAll()
{
	ptbTab.Free();
	itemT = NULL;
	itemQ = 0;
	return true;
}

bool cParamsSSS::addShare( int numberOfShares, cPassword *pwd )
{
	if ( !checkPwdIsDifferent( pwd ) )
		return false;

	if ( ptbTab.WriteBytes( NULL, sizeof( sssPwdItem ), false ) == 0 )
	{
		assert( false );
		return false;
	}
	
	itemT = ( sssPwdItem * )ptbTab.buffer;

	int last = itemQ;

	itemQ ++;
	assert( itemQ == ( ptbTab.bufferSize / sizeof( sssPwdItem ) ) );

	itemT[ last ].numberOfShares = numberOfShares; 

	//
	// 28.06.2011
	//	1. The cUniversalString::GenerateUnicodeString uses wcBuffer for valid unicode string generation.
	//	If wcBuffer is NULL it simple adds zero bytes to every symbol from buffer for unicode string generation.
	//	But SetUnicode does not fill wcBuffer, we have to use Set function instead.
	//	As result the shared scheme works with simple latin symbols bu does not work with real unicode.
	//
	//itemT[ last ].pwd.Set( pwd->buffer ); 
	//itemT[ last ].pwd.SetUnicode( pwd->unicodeBuffer ); 
	//itemT[ last ].pwd.SetUnicode( pwd->wcBuffer );
	//
	//	The Set function is changed at Certificate lib v. 1.03.4 - 28.06.2011 
	//	So we can use here the updated function.
	//
	itemT[ last ].pwd.Set( pwd );

	return true;
}

#pragma pack (1)

//	the version encodes entire sKeyInfo
#define MaxEncodedSize ( sizeof( sKeyInfo ) )

//
//	MaxPrimeSize - size in bytes
//	Prime size is one byte more then size of the key to be encoded ( algorithm defined )
//
#define MaxPrimeSize ( MaxEncodedSize + 1 )

struct _sShareV1
{
	BYTE version;
	BYTE reqNum;
	BYTE totalQ;
	BYTE headOfPrime; // One first byte of prime always contain elder bit, so we hide it in encrypted part
	BYTE valX[ MaxPrimeSize ];
	BYTE valY[ MaxPrimeSize ]; // MAXIMUM_KEY_SIZE_BYTES + 1 
	BYTE hashAlgId;
};

struct sShareV1 : public _sShareV1
{
	BYTE reserved[ DRoundTail( sizeof( _sShareV1 ), DB_MaxEncBlockSize ) ];
	BYTE digest[ MAXIMUM_DIGEST_SIZE_BYTES ];
};

struct _sShareV2
{
	BYTE version;
	BYTE reqNum;
	BYTE totalQ;		// The value is required to check ( or detect real key block size busied by the scheme ).
	BYTE headOfPrime;	// One first byte of prime always contain elder bit, so we hide it in encrypted part
	BYTE valX[ 1 ];		// share order number is used for x value.
	BYTE valY[ MaxPrimeSize ];	// MAXIMUM_KEY_SIZE_BYTES + 1
	BYTE hashAlgId;
};

struct sShareV2 : public _sShareV2
{
	BYTE reserved[ DRoundTail( sizeof( _sShareV2 ), DB_MaxEncBlockSize ) ];
	BYTE digest[ MAXIMUM_DIGEST_SIZE_BYTES ];
};

struct sShareV3 : public _sShareV2
{
	BYTE reserved[ DRoundTail( sizeof( _sShareV2 ), DB_MaxEncBlockSize ) ];
	BYTE digest[ MAXIMUM_DIGEST_SIZE_BYTES_64 ];
};

#pragma pack ()

//
// prime elder byte is hidden and moved to encrypted part, prime firest 8 bytes is used as salt.
//
static const int primeAndSaltSize = MaxPrimeSize - 1; 
static const int saltSize = 8;

static int getShareSize( BYTE ver )
{
	switch ( ver )
	{
		case 1: return sizeof( sShareV1 );
		case 2: return sizeof( sShareV2 );
		case 3: return sizeof( sShareV3 );
		default: return -1;
	}
}

static bool fillValue( BYTE *value, DWORD size, pw_int *p )
{
	if ( pw_getSizeInBytes( p ) > (int)size )
	{
		assert( false );
		return false;
	}

	if ( pw_int2bytes( value, size, p ) != pw_OK )
	{
		assert( false );
		return false;
	}

	return true;
}

//static int hashAlgID = pgphaSHA1;

//
//	sssCalculateDigest
//	 Calculates digest of
//		1. plain text part of prime.
//		2. the share info excluding the digest part.
//
template < class TT >
	int sssCalculateDigest( BYTE *primeBuffer, TT *sv, BYTE hashAlgId )
//	int sssCalculateDigest( BYTE *primeBuffer, sShareV2 *sv, BYTE hashAlgId )
{
	ODSFuncS( _FormatT( L"%d(%d), %d", sizeof(TT), sizeof( sv->digest ), hashAlgId ) );

	sv->hashAlgId = hashAlgId;

	HHA ha;
	int res = ha_createAlgorithm( &ha, hashAlgId );

	if ( res != hae_ok )
	{
		assert( false );
		ODS_fls( _FormatT( L"Assertion! Error %d", res ) );
		return ERROR_INVALID_ALGORITHM;
	}

//	ShredData( sv->reserved, sizeof( sv->reserved ) );
	ShredData( sv->digest, sizeof( sv->digest ) );

	bool err = ( ha_getDigestSize( ha ) > sizeof( sv->digest ) );
	if ( err )
		ODS_fls( _FormatT( L"Assertion! Digest size %d > buffer size %d", ha_getDigestSize( ha ), sizeof( sv->digest ) ) );
	else
		err =
		// Processes prime low bytes
			( ha_process( ha, primeBuffer, primeAndSaltSize ) != hae_ok ) ||
		// Processes prepared share
			( ha_process( ha, (BYTE *)sv, sizeof( *sv ) - sizeof( sv->digest ) ) != hae_ok ) ||
			( ha_final( ha, sv->digest ) != hae_ok );

	ha_free( ha );

	if ( err )
	{
		assert( false );
		ODS_fls( _FormatT( L"Assertion! Error %d", err ) );
		return ERROR_INTERNAL_PROBLEM;
	}

	return ERROR_NO;
}

//
// Calculates digest of
//	1. plain text part of prime.
//	2. the share info excluding the digest part.
//	3. verify digest
//
template < class TT >
	int sssVerifydigest( BYTE *primeBuff, TT *sv )
//	int sssVerifydigest( BYTE *primeBuff, sShareV2 *sv )
{
	HHA ha;
	int res = ha_createAlgorithm( &ha, sv->hashAlgId );

	if ( res != hae_ok )
	{
//		assert( false );
		return ERROR_INCORRECT_HASH_ALGORITHM;
	}
	
	int digestSize = ha_getDigestSize( ha );
	cADBlock < BYTE > adbDigest( digestSize );

	bool err = 
			( digestSize > sizeof( sv->digest ) ) ||
		// Processes prime low bytes
			( ha_process( ha, primeBuff, primeAndSaltSize ) != hae_ok ) ||
		// Processes prepared share
			( ha_process( ha, (BYTE *)sv, sizeof( *sv ) - sizeof( sv->digest ) ) != hae_ok ) ||
			( ha_final( ha, adbDigest.m_T ) != hae_ok );

	ha_free( ha );

	if ( err )
	{
		if ( digestSize > sizeof( sv->digest ) )
			return ERROR_INCORRECT_HASH_ALGORITHM;

		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	if ( memcmp( adbDigest.m_T, sv->digest, digestSize ) != 0 )
		return ERROR_INCORRECT_PASSWORD;

	return ERROR_NO;
}


//
//	encodeSharesEx
//		walks tables with share's passwords ( itemT )
//			1. creates symmetric algorithm from password
//			2. fills shares and calculates it's digest
//				the digest is started from plain text primes low bytes 
//				and finished by prepared shares, so we may check the prime integrity
//				together with the share in decoding porcess.
//			3. encodes corresponded share quantity from generated shares table ( sss )
//
template < class TT >
	int encodeSharesEx(
		BYTE		version,
		BYTE		*buff,		// Buffer for encoded key
		int			*pBuffSize,

		sssPwdItem	*itemT,		// Table with passwords and heights of holder
		int			itemQ,		// Number of items in the table
		SSShares	*sss,		// Allocated and generated shares' table
		DWORD		algId,		// Identifier of algorithm for password based share encryption
		DWORD		encModeId,	// Encryption mode
		DWORD		hashAlgId,	// Identifier of algorithm for generation key from password
		size_t		reqN		// Number of required shares, to be stored in encoded part.
	)
{
	ODSFuncI( version );

	int totalSharesQ = getTotalSharesQ( itemT, itemQ );

	int requiredSize = primeAndSaltSize + totalSharesQ * getShareSize( version );
	
	if ( *pBuffSize < requiredSize )
	{
		*pBuffSize = requiredSize;
		ODS_fls( L"Warinig!" );
		return ERROR_BUFFER_TOO_SMALL;
	}

	*pBuffSize = requiredSize;

	// Symmetric algorithm for share encryption, key is generated from corresponded shares' password.
	cSymAlgorithm sPwdAlg;

	if ( !sPwdAlg.Init( algId, encModeId ) )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INVALID_ALGORITHM;
	}

	int pwdKeySizeBytes = ( sPwdAlg.m_keyLengthInBites + 7 ) / 8;

	//	Writes prime's low bytes to buffer (where the password key will be stored later )
	//	and save elder byte to be hidden to encrypted part.

	int buffSize = max( pwdKeySizeBytes, primeAndSaltSize + 1 );

	// Deploys prime low bytes to buffer start and saves elder byte 
	cPlainTextBlock ptb;
	if ( ptb.Allocate( buffSize ) == NULL )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}


	int primeSizeBytes = pw_getSizeInBytes( &sss->prime );
	if ( buffSize < primeSizeBytes )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	pw_int2bytes( ptb.buffer, primeSizeBytes, &sss->prime );

	BYTE *pelderByte = &ptb.buffer[0];

	int pos = 0;
	memcpy( buff + pos, ptb.buffer + 1, primeSizeBytes );
	pos += primeAndSaltSize;

	// Prepares buffer for plain text share
	int shareSize = getShareSize( version );
	if ( shareSize <= 0 )
		return ERROR_INCORRECT_FORMAT_VERSION;

	cPlainTextBlock ptbShare;
	if ( ptbShare.Allocate( shareSize ) == NULL )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	TT *sv = ( TT * )ptbShare.buffer;

	assert( sizeof( *sv ) == shareSize );
	assert( sizeof( TT ) == shareSize );

	sv->version		= version;
	sv->reqNum		= (BYTE)reqN;
	sv->totalQ		= (BYTE)totalSharesQ;
	sv->headOfPrime	= *pelderByte;

	if ( primeAndSaltSize - saltSize < 0 )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	BYTE *salt = buff + primeAndSaltSize - saltSize;

	int itemN = 0;	// position in itemT
	int sssN = 0;	// position in sss

	BYTE iv[ IVECTOR_LENGTH_DB ];
	memset( iv, 0, sizeof( iv ) );

	// Password key buffer uses same temporary ptb where prime has been deployed
	BYTE *pwdKeyBuff = ptb.buffer;

	for ( ; itemN < itemQ; itemN++ )
	{
		if ( !cParamsPBE::generateKeyFromPassword( pwdKeyBuff, pwdKeySizeBytes, &itemT[ itemN ].pwd, hashAlgId, salt, saltSize ) ||
			 !sPwdAlg.createKey( algId, encModeId, pwdKeyBuff, pwdKeySizeBytes, NULL ) 
			)
		{
			assert( false );
			ODS_fls( L"Assertion!" );
			return ERROR_INTERNAL_PROBLEM;
		}

		for( int i = 0; i < itemT[ itemN ].numberOfShares; i++, sssN++ )
		{
			if ( sssN >= (int)sss->arraySize )
			{
				assert( false );
				ODS_fls( L"Assertion!" );
				return ERROR_INTERNAL_PROBLEM;
			}

			// Fills x and y values

			if ( !fillValue( sv->valX, sizeof( sv->valX ), &sss->array[ sssN ].x )  ||
					!fillValue( sv->valY, sizeof( sv->valY ), &sss->array[ sssN ].y ) 
					)
			{
				assert( false );
				ODS_fls( L"Assertion!" );
				return ERROR_INTERNAL_PROBLEM;
			}

			if ( pos + shareSize > requiredSize )
			{
				assert( false );
				ODS_fls( L"Assertion!" );
				return ERROR_INTERNAL_PROBLEM;
			}
			
			// Calculates digest of
			//	1. plain text part of prime.
			//	2. the share info exluding the digest part.
			int err = sssCalculateDigest( buff, sv, (BYTE)hashAlgId );

			if ( err != ERROR_NO )
			{
				assert( false );
				ODS_fls( _FormatT( L"Assertion error %x!", err ) );
				return ERROR_INTERNAL_PROBLEM;
			}

			// Encode share directly to represent buffer
			// It uses number of share as initial vector

			iv[0] = LOBYTE( WORD( sssN ) );
			iv[1] = HIBYTE( WORD( sssN ) );

			if ( !sPwdAlg.encrypt( iv, buff + pos, ( BYTE * )sv, shareSize ) )
			{
				assert( false );
				ODS_fls( L"Assertion!" );
				return ERROR_INTERNAL_PROBLEM;
			}

			pos += shareSize;
		}
	}

	if ( pos != requiredSize )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	ODSFuncRes( ERROR_NO );
	return ERROR_NO;
}

//
//	encodeShares
//	 0. Checks for different passwords in itemT table
//	 1. Generates prime p one byte more than the key size
//	 2. Generates (recNum - 1) degree polinom P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= recNum -1,
//	 3. Generates shares array by Xi and Yi where P(Xi) == Yi, the array quantity is Summ of itemT[i].sharesQ, where 0 <= i < itemQ
//	 4. For every i gets N = itemT[i].sharesQ shares from generated shares array and encrypts the shares by passwords from itemT[i].pwd
//
//		Encrypted item format is:
//			encrypted:
//				Ver + RN + PrimeelderByte + Xi + Yi + hashAlgId + reservedBytes + digest
//		where:
//			Ver ( 1 byte ) is version number
//				Ver 1 has 152 bytes size of encrypted share ( see getShareSize() )
//			RN  ( 1 byte ) is number of required shares.
//			PrimeelderByte - prime is saved in open state but it's elder bytes has assigned elder bit,
//							it is a reason why the byte was hidden to encrypted part.
//			The values is required to recovering key,
//			but it has to be encrypted, so we save it here.
//
//			reservedBytes is random filled value it used for normal encryption with any block based algorithm 
//
//		digest size is defined as MAXIMUM_DIGEST_SIZE_BYTES
//
//		prime and Xi Yi sizes is defined by key size, see MaxPrimeSize
//
//		SSS encrypted key block format:
//			prime small bytes ( without elder ) ( note,	lowest 8 bytes of prime is common salt for all shares )
//			N Encrypted items
//
//		Maximum estimated shares size for 3K block:
//			V1 is -> 70 + N * ( sizeof( sShareV1 ) == 192 ) -> N = ( 3K - 70 ) ) / 192 > 15 shares
//			V2 is -> 70 + N * ( sizeof( sShareV2 ) == 112  )-> N = ( 3K - 70 ) ) / 112 > 26 shares
//
//	Recovery process looks like:
//		1. Gets table with passwords and implements every password to every encrypted item
//			if decrypted digest is compatible the shared Xi,Yi are added to new share table
//		2. Checks if required shared quantity achived
//		3. Implements A0 - polinom member recovering, that is our key.
//
static int encodeShares(
		BYTE		version,

		BYTE *buff,				// Buffer for encoded key
		int *pBuffSize,

		sssPwdItem	*itemT,		// Table with passwords and heghts of holder
		int			itemQ,		// Number of items in the table
		SSShares	*sss = NULL,// Allocated and generated shares' table
		DWORD		algId = 0,	// Identifier of algorithm for password based share encryption
		DWORD		encModeId = 0,
		DWORD		hashAlgId = 0,	// Identifier of algorithm for generation key from password
		size_t		reqN = 0	// Number of required shares, to be stored in encoded part.
	)
{
	int res = ERROR_INCORRECT_FORMAT_VERSION;

	if ( version == (BYTE)-1 )
		version = cParamsSSS::getDefultVersion( hashAlgId ); // defaultSSSVersion;

	if ( version == 3 )
		res = encodeSharesEx< sShareV3 >( version, buff, pBuffSize, itemT, itemQ, sss, algId, encModeId, hashAlgId, reqN );
	else
	if ( version == 2 )
		res = encodeSharesEx< sShareV2 >( version, buff, pBuffSize, itemT, itemQ, sss, algId, encModeId, hashAlgId, reqN );
	else
	if ( version == 1 )
		res = encodeSharesEx< sShareV1 >( version, buff, pBuffSize, itemT, itemQ, sss, algId, encModeId, hashAlgId, reqN );
	else
		res = ERROR_INCORRECT_FORMAT_VERSION;

	return res;
}

static int getMaxSharesQ( BYTE version, DWORD maxBufferSize )
{
	return ( maxBufferSize - primeAndSaltSize ) / getShareSize( version );
}

int cParamsSSS::getMaxSharesQ()
{
	return ::getMaxSharesQ( version, maxBufferSize );
}

int cParamsSSS::getSeedSize( int keyBuffSize )
{
	SSShares *s = NULL;
	int totalSharesQ = getTotalSharesQ();

	int res = SSSAllocateShares( &s, itemQ );
	if ( res != ppe_ok )
		return -1;

	if ( version == (BYTE)-1 )
		version = getDefultVersion( m_hashAlgId );

	// Version 2 uses one byte x values, so theirs are generated from order number
	if ( ( version != 1 ) && 
			!prepareOrderedX( s ) )
	{
		SSSFree( &s );
		return -1;
	}

	int seedSize = 0;

	res = SSSGenerateShares( s, totalSharesQ, NULL, keyBuffSize, NULL, &seedSize );
	SSSFree( &s );

	if ( res != ppe_randomBufferRequired )
	{
		return -1;
	}

	return seedSize;
}

//INT_PTR CALLBACK simpleDialogProc(
//		HWND hwndDlg,
//		UINT uMsg,
//		WPARAM wParam,
//		LPARAM lParam
//	)
//{
//	if ( uMsg == WM_INITDIALOG )
//	{
//		if ( pTranslateDll != NULL )
//			pTranslateDll->TranslateDialog( hwnd );
//
//		return true;
//	}
//
//	return false;
//}

class cProgressDlg_ShareGen : public cProgressBlock, public cDlg
{
	SSShares *m_s;
	int m_reqQ;
	BYTE *m_keyBuff;
	int m_keyBuffSize;
	BYTE *m_seed;
	int m_seedSize;

	bool m_cancelled;
	HWND m_hProgress;
//	bool m_useMarqueeMode; 

public:
	int m_res;

	cProgressDlg_ShareGen( SSShares *s, int reqQ, BYTE *keyBuff, int keyBuffSize, BYTE *seed, int seedSize )
	{
		m_s = s;
		m_reqQ = reqQ;
		m_keyBuff = keyBuff;
		m_keyBuffSize = keyBuffSize;
		m_seed = seed;
		m_seedSize = seedSize;
		m_res = -1;
		m_cancelled = false;
	}

	BOOL InitDialog()
	{
		m_hProgress = GetDlgItem( m_hDlg, IDC_PROGRESS );

		SendMessage( m_hProgress, PBM_SETRANGE, 0, MAKELPARAM( 0, 100 ) );
//		m_useMarqueeMode = SendMessage( m_hProgress, (WM_USER+10), true, 0 ); // PBM_SETMARQUEE
		SendMessage( m_hProgress, PBM_SETSTEP, 1, 0 );

		DWORD threadId;
		HANDLE hThread = CreateThread( NULL, 0, threadFunc, this, 0, &threadId );

		if ( hThread == NULL )
		{
			MessageBoxT( m_hDlg, L"Error in thread creating", _LoadIDS( IDS_KeyGen ), MB_ICONSTOP );
			EndDialog( m_hDlg, -1 );
			return false;
		}

		CloseHandle( hThread );

		return TRUE;
	}

	int Generate()
	{
		m_res = SSSGenerateShares( m_s, m_reqQ, m_keyBuff, m_keyBuffSize, m_seed, &m_seedSize, this );
		SendMessage( m_hProgress, PBM_SETPOS, true, 0 );
		return m_res;
	}

	static DWORD WINAPI threadFunc( void *ptr )
	{
		cProgressDlg_ShareGen *pThis = ( cProgressDlg_ShareGen * )ptr;

		pThis->Generate();

		PostMessage( pThis->m_hDlg, WM_COMMAND, IDOK, 0 );

		return 0;
	}
	
	void SetText( const char *text )
	{
		SetDlgItemTextA( m_hDlg, IDC_Text, text );
	}

	// Progress test result enum
	// enum { stepResult_False_ByTable = -1, stepResult_False_ByPrimeTest = 0, stepResult_True = 1 };
	
	//
	// SetCurrentStepResult
	//	Reports result of every step, see enum above.
	//	Returns false to stop the process, otherwise returns non zero.
	//
	bool SetCurrentStepResult( int res )
	{
		if ( m_cancelled )
			return false;

//		if ( res == stepResult_False_ByPrimeTest )
//			SendMessage( m_hProgress, PBM_SETPOS, 0, 0 );
//		SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

		if ( res == stepResult_False_ByPrimeTest )
			SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

		return true;
	}

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		switch (uMsg) 
		{ 
			case WM_INITDIALOG:
				return InitDialog();

	        case WM_COMMAND: 
			{
				switch(wParam) 
				{ 
					case IDOK: 
						EndDialog( m_hDlg, m_res );
						return TRUE; 
					case IDCANCEL:
					{
						int res = MessageBoxT(
									m_hDlg,
									_LoadIDS( IDS_Q_StopGenerationProcess ),
									_LoadIDS( IDS_KeyGen ),
									MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2 );

						if ( res != IDYES )
								return 0;

						m_cancelled = true;
						EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), false );
						return TRUE;
					}
					default:
						return 0; 
				}
			}
			default:
				return 0; 
		}

		return 0;
	}
};


bool cParamsSSS::encodeKey( BYTE *kbEncrypted, int kbEncryptedSize, BYTE *keyBuff, int keyBuffSize )
{
	ODSFunc;

	SSShares *s = NULL;

	int totalSharesQ = getTotalSharesQ();

	int res = SSSAllocateShares( &s, totalSharesQ ); // itemQ );
	if ( res != ppe_ok )
	{ ODS_fls( L"Assertion!" ); return false; }

	if ( version == (BYTE)-1 )
	{
		version = getDefultVersion( m_hashAlgId ); // defaultSSSVersion;
	}

	// Version 2 and 3 use one byte x values, so theirs are generated from order number
	if ( ( version != 1 ) && 
			!prepareOrderedX( s ) )
	{
		SSSFree( &s );
		ODS_fls( L"Assertion!" );
		return false;
	}

	int seedSize = 0;

	res = SSSGenerateShares( s, totalSharesQ, keyBuff, keyBuffSize, NULL, &seedSize );
	if ( res != ppe_randomBufferRequired )
	{
		SSSFree( &s );
		ODS_fls( L"Assertion!" );
		return false;
	}

	//
	//	Check if seed value has been generated early.
	//
	if ( seedSize > ptbSeed.bufferSize )
	{
		if ( ptbSeed.Allocate( seedSize ) == NULL )
		{
			SSSFree( &s );
			ODS_fls( L"Assertion!" );
			return false;
		}

		res = GetSeedValue( hWnd, ptbSeed.buffer, seedSize );
		if ( res != ERROR_NO )
		{
			SSSFree( &s );
			ODS_fls( L"Assertion!" );
			return false;
		}
	}

	BYTE *seed = ptbSeed.buffer;

	cProgressDlg_ShareGen dlgMT( s, reqQ, keyBuff, keyBuffSize, seed, seedSize );
	dlgMT.DoModal( IH.m_hInst, IDD_WaitShareGenerationDlg, hWnd );

	res = dlgMT.m_res;

//	res = SSSGenerateShares( s, reqQ, keyBuff, keyBuffSize, seed, &seedSize, progress );

	if ( res != ppe_ok )
	{
		SSSFree( &s );
		ODS_fls( L"Assertion!" );
		return false;
	}

#ifdef _DEBUG
	BYTE *tmpBuff = new BYTE[ keyBuffSize ];
	bool same = false;
	while( !same )
	{
		res = SSSRestoreKey( s, tmpBuff, keyBuffSize );
		same = memcmp( tmpBuff, keyBuff, keyBuffSize ) == 0;
		assert( same );
	}
	delete tmpBuff;
#endif

	res = encodeShares( version, kbEncrypted, &kbEncryptedSize, itemT, itemQ, s, m_algId, m_encModeId, m_hashAlgId, reqQ );

	SSSFree( &s );

	ODSFuncRes( res );
	return res == ERROR_NO;
}

int cParamsSSS::getEncodedKeySize( int keyLength )
{
	if ( keyLength > sizeof( sKeyInfo ) ) // MAXIMUM_KEY_SIZE_BYTES )
		return - 1;

	int buffSize = 0;
	int res = encodeShares( version, NULL, &buffSize, itemT, itemQ );

	if ( res == ERROR_BUFFER_TOO_SMALL )
	{
		if ( encodedSize > buffSize )
			return encodedSize;

		return buffSize;
	}

	return -1;
}

//
//	decodeSingleShare
//		1. Walks all encrypted shares and trys to decrypt by sPwdAlg
//		2. If decrypted shares is properly verified it adds the share to share table sss
//		3. If share table is not allocate yet ( first share decoded ), so:
//			pSSS is allocated upto decoded required size
//			(*pSSS)->prime value is filled
//
//	Return values:
//		1. If no one shares was properly decoded it returns ERROR_INCORRECT_PASSWORD
//		2. If shares is decoded but it is not enought shares in table for key decoding
//			so it returns ERROR_NOT_ENOUGH_INIT_KEYS
//		3. If share table is completely filled it returns ERROR_NO
//
//		Otherwise returns ERROR_INTERNAL_PROBLEM.
//
template < class TT >
	int decodeSingleShare(
		BYTE &version,
		BYTE *kbEncrypted, int *kbEncryptedSize,
		int keySize,
		SSShares	**pSSS,		// Allocated shares' table to added by single item if is decoded.
		cSymAlgorithm	*sPwdAlg
	)
{
	bool found = false;

	BYTE *primeLowBytes = kbEncrypted; // + keySize;

	int shareSize = getShareSize( version );

	if ( shareSize <= 0 )
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	cPlainTextBlock ptb;
	if ( ptb.Allocate( shareSize ) == NULL )// + MAXIMUM_DIGEST_SIZE_BYTES
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	TT *sv = ( TT *)ptb.buffer;
//	BYTE *tmpDigest = ptb.buffer + shareSize;

	BYTE iv[ IVECTOR_LENGTH_DB ];
	memset( iv, 0, sizeof( iv ) );

	WORD sssN = 0;

	for(
		BYTE *pos = kbEncrypted + primeAndSaltSize;
			(( sssN + 1 ) * shareSize) <= (*kbEncryptedSize - primeAndSaltSize);
				pos += shareSize, sssN++ )
	{
		// Decrypt share to allocated buffer
		// It uses number of share as initial vector

		iv[0] = LOBYTE( WORD( sssN ) );
		iv[1] = HIBYTE( WORD( sssN ) );
		if ( !sPwdAlg->decrypt( iv, (BYTE *)sv, pos, shareSize ) )
		{
			assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}

		int err = sssVerifydigest( kbEncrypted, sv );

		if ( ( err == ERROR_INCORRECT_PASSWORD ) ||
			 ( err == ERROR_INCORRECT_HASH_ALGORITHM ) )
			continue;

		if ( err != ERROR_NO )
		{
			assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}

		// The shares was properly decrypted and verified

		found = true;

		//
		// 24.09.14 
		//	Version 3 properly decodes shares version 2, but real share version is stored in decryped body.
		//		So we just update this version.
		//
		version = sv->version;

		//
		//	12.11.2014
		//	Bug #1734 SSS creation error,
		//		if version is changed so share size may be changed too.
		//
		shareSize = getShareSize( version );

		// Checks encoded size.
		sssPwdItem pwdItem;
		pwdItem.numberOfShares = sv->totalQ;
		int checkSize = 0;
        int res = encodeShares( sv->version, NULL, &checkSize, &pwdItem, 1 );
		if ( res == ERROR_BUFFER_TOO_SMALL )
		{
//			assert( false );
			if ( checkSize > *kbEncryptedSize )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			*kbEncryptedSize = checkSize;
		}

		// Allocates the share table
		if ( *pSSS == NULL )
		{
			int res = SSSAllocateShares( pSSS, sv->reqNum );
			if ( res != ppe_ok )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			cPlainTextBlock ptbPrime;

			if ( ptbPrime.Allocate( keySize + 1 ) == NULL )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			ptbPrime.buffer[0] = sv->headOfPrime;
			memcpy( ptbPrime.buffer + 1, kbEncrypted, keySize );

			if ( (*pSSS)->prime.set( keySize + 1, ptbPrime.buffer ) != pw_OK )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}
		}

		pw_int x,y;

		if ( ( x.set( sizeof( sv->valX ), sv->valX ) != pw_OK ) ||
			 ( y.set( sizeof( sv->valY ), sv->valY ) != pw_OK ) )
		{
			assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}

		res = SSSAddShare( *pSSS, &x, &y );
	
		switch ( res )
		{
			case pw_OK: return ERROR_NO;
			case pw_InvalidKeyBlock:
			case pw_BufferSizeTooSmall:
				continue;
			default:
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
		}
	}

	if ( !found )
		return ERROR_INCORRECT_PASSWORD;

	return ERROR_NOT_ENOUGH_INIT_KEYS;
}

//
//	cParamsSSS::decodeKey
//		Walks collected password items from itemT and tries 
//			to implement the password to every share
//
bool cParamsSSS::decodeKey( BYTE *keyBuff, int keyBuffSize, BYTE *kbEncrypted, int kbEncryptedSize )
{
	ODSFunc;

	SSShares *s = NULL;

	cSymAlgorithm sPwdAlg;
	if ( !sPwdAlg.Init( m_algId, m_encModeId ) )
		return false; // ERROR_INVALID_ALGORITHM;

	cPlainTextBlock ptb;
	int pwdKeyBuffSize = ( sPwdAlg.m_keyLengthInBites + 7 ) / 8;
	if ( ptb.Allocate( pwdKeyBuffSize ) == NULL ) // keyBuffSize ) == NULL )
		return false; // ERROR_INTERNAL_PROBLEM;

	int decodedSize = MaxEncodedSize; // sKeyInfo::sizeWithoutKey + pwdKeyBuffSize;
	if ( decodedSize > keyBuffSize )
		return false;
	
	BYTE *pwdKeyBuff = ptb.buffer;
	BYTE *salt = kbEncrypted + primeAndSaltSize - saltSize;

	int res = ERROR_INTERNAL_PROBLEM;

	for( int i = 0; i < itemQ; i++ )
	{
		assert( m_hashAlgId != -1 );

		if ( !cParamsPBE::generateKeyFromPassword( pwdKeyBuff, pwdKeyBuffSize, &itemT[ i ].pwd, m_hashAlgId,salt, saltSize ) ||
			 !sPwdAlg.createKey( m_algId, m_encModeId, pwdKeyBuff, pwdKeyBuffSize, NULL ) )
		{
			res = ERROR_INTERNAL_PROBLEM;
			break;
		}

		if ( version == (BYTE)-1 ) // undefined version
		{
			res = decodeSingleShare<sShareV3>( version = 3, kbEncrypted, &kbEncryptedSize, decodedSize, &s, &sPwdAlg );
			if ( res == ERROR_NOT_ENOUGH_INIT_KEYS ) continue;
			if ( res == ERROR_NO ) break;

			res = decodeSingleShare<sShareV2>( version = 2, kbEncrypted, &kbEncryptedSize, decodedSize, &s, &sPwdAlg );
			if ( res == ERROR_NOT_ENOUGH_INIT_KEYS ) continue;
			if ( res == ERROR_NO ) break;
			
			res = decodeSingleShare<sShareV1>( version = 1, kbEncrypted, &kbEncryptedSize, decodedSize, &s, &sPwdAlg );
			if ( res == ERROR_NOT_ENOUGH_INIT_KEYS ) continue;

			if ( res != ERROR_NO )
				version = (BYTE)-1;

			break;
		}
		else if ( version == 1 ) res = decodeSingleShare<sShareV1>( version, kbEncrypted, &kbEncryptedSize, decodedSize, &s, &sPwdAlg );
		else if ( version == 2 ) res = decodeSingleShare<sShareV2>( version, kbEncrypted, &kbEncryptedSize, decodedSize, &s, &sPwdAlg );
		else if ( version == 3 ) res = decodeSingleShare<sShareV3>( version, kbEncrypted, &kbEncryptedSize, decodedSize, &s, &sPwdAlg );
		else				 res = ERROR_INTERNAL_PROBLEM;

		if ( res != ERROR_NOT_ENOUGH_INIT_KEYS )
			break;
	}

	if ( s == NULL )
		return false; // ERROR_INTERNAL_PROBLEM;

	if ( res == ERROR_NO )
	{
		if ( SSSRestoreKey( s, keyBuff, decodedSize ) != pw_OK )
			res = ERROR_INTERNAL_PROBLEM;
	}

	SSSFree( &s );
	
	assert( res == ERROR_NO );

	return res == ERROR_NO;
}

//static 
int cParamsSSS::getDefultVersion( DWORD hashAlgId )
{
	int digestSize = 0;
	if ( ( hae_ok == ha_getDigestSize( &digestSize, hashAlgId ) ) && 
		 ( digestSize == sizeof( ((sShareV3*)0)->digest ) )
		)
			return 3;

	return defaultSSSVersion;
}


//
//	cParamsSSS::decodeKeyWithFirstPwd
//		The function is used if key block type is undefined ( hidden or was not compatible with PBE, PKE )
//		1. Tryes first password, 
//		2. If first pwd is compatible, it asks passwords while required shares will not collected
//
bool cParamsSSS::decodeKeyWithFirstPwd( cPassword *firstPwd, BYTE *keyBuff, int keyBuffSize, BYTE *kbEncrypted, int *pKBEncryptedSize, bool checkPasswordOnly )
{
	if ( pErrorCode == NULL )
	{ assert( false ); return false; }

	SSShares *s = NULL;

	cSymAlgorithm sPwdAlg;

	if ( !sPwdAlg.Init( m_algId, m_encModeId ) )
		return false; // ERROR_INVALID_ALGORITHM;

	cPlainTextBlock ptb;
	int pwdKeyBuffSize = ( sPwdAlg.m_keyLengthInBites + 7 ) / 8;
	if ( ptb.Allocate( pwdKeyBuffSize ) == NULL ) // keyBuffSize ) == NULL )
		return false; // ERROR_INTERNAL_PROBLEM;

	int decodedSize = MaxEncodedSize; // sKeyInfo::sizeWithoutKey + pwdKeyBuffSize;
	if ( decodedSize > keyBuffSize )
		return false;
	
	BYTE *pwdKeyBuff = ptb.buffer;
	BYTE *salt = kbEncrypted + primeAndSaltSize - saltSize;

	int res = ERROR_INTERNAL_PROBLEM;

	cPassword pwd, *curPwd;
	curPwd = firstPwd;

	wchar_t text[ 256 ];
	wchar_t caption[ 256 ];
	
	LoadStringT( IH.m_hInst, IDS_SSS_FirstSharedPasswordDetected, caption, maxLen( caption ) );

	while( true )
	{
		if ( m_hashAlgId == -1 )
		{
			assert( false );
			res = ERROR_INCORRECT_HASH_ALGORITHM;
			break;
		}

		if ( !cParamsPBE::generateKeyFromPassword( pwdKeyBuff, pwdKeyBuffSize, curPwd, m_hashAlgId, salt, saltSize ) ||
			 !sPwdAlg.createKey( m_algId, m_encModeId, pwdKeyBuff, pwdKeyBuffSize, NULL ) )
		{
			res = ERROR_INTERNAL_PROBLEM;
			break;
		}

		if ( version == (BYTE)-1 ) // undefined version
		{
			res = decodeSingleShare<sShareV3>( version = 3, kbEncrypted, pKBEncryptedSize, decodedSize, &s, &sPwdAlg );
			if ( res == ERROR_NO ) break;
			if ( res == ERROR_NOT_ENOUGH_INIT_KEYS ) continue;

			res = decodeSingleShare<sShareV2>( version = 2, kbEncrypted, pKBEncryptedSize, decodedSize, &s, &sPwdAlg );
			if ( res == ERROR_NO ) break;
			if ( res == ERROR_NOT_ENOUGH_INIT_KEYS ) continue;

			res = decodeSingleShare<sShareV1>( version = 1, kbEncrypted, pKBEncryptedSize, decodedSize, &s, &sPwdAlg );
			if ( ( res != ERROR_NOT_ENOUGH_INIT_KEYS ) && ( res != ERROR_NO ) )
				version = (BYTE)-1;
		}
		else if ( version == 1 ) res = decodeSingleShare<sShareV1>( version, kbEncrypted, pKBEncryptedSize, decodedSize, &s, &sPwdAlg );
		else if ( version == 2 ) res = decodeSingleShare<sShareV2>( version, kbEncrypted, pKBEncryptedSize, decodedSize, &s, &sPwdAlg );
		else if ( version == 3 ) res = decodeSingleShare<sShareV3>( version, kbEncrypted, pKBEncryptedSize, decodedSize, &s, &sPwdAlg );
		else				 res = ERROR_INTERNAL_PROBLEM;

		if ( res == ERROR_INCORRECT_PASSWORD )
		{
			if ( curPwd == firstPwd )
			{
				*pErrorCode = res;
				break;
			}

			MessageBoxT( hWnd,
				_LoadIDS( IDS_INCORRECT_PASSWORD ),
				caption,
				MB_ICONWARNING );
		}
		else
		if ( res != ERROR_NOT_ENOUGH_INIT_KEYS )
			break;
		else
		{
			//
			//	We add all passwords to the table to check if same password is enterred before only,
			//	see checkPwdIsDifferent
			//
			addShare( 1, curPwd );
		}

		if ( checkPasswordOnly ) // do not ask additional password
		{
			res = ERROR_NO;
			break;
		}

		curPwd = &pwd; // to avoid the first pwd using.

		LoadStringT ( IH.m_hInst, IDS_AdditionalSharedPWDRequired, text, maxLen( text ) );

		while( true )
		{
			res = GetPassword( hWnd,
					text,
					caption,
					curPwd,
					NULL, // keyFileList has to be assigned 
					pFlags );

			*pErrorCode = res;

			if ( res != ERROR_NO ) break;

			if ( checkPwdIsDifferent( curPwd ) )
				break;

			LoadStringT( IH.m_hInst, IDS_SSS_SamePasswords, text, maxLen( text ) );
			LoadStringT( IH.m_hInst, IDS_SSS_SamePasswordCaption, caption, maxLen( caption ) );
		}
		
		LoadStringT( IH.m_hInst, IDS_ENTER_PASSWORD, caption, maxLen( caption ) );

		if ( res != ERROR_NO ) break;
	}

	if ( s == NULL )
		return false; // ERROR_INTERNAL_PROBLEM;

	if ( ( res == ERROR_NO ) && !checkPasswordOnly )
	{
		if ( SSSRestoreKey( s, keyBuff, decodedSize ) != pw_OK )
			res = ERROR_INTERNAL_PROBLEM;
	}

	SSSFree( &s );

	assert( res == ERROR_NO );

	return res == ERROR_NO;
}

